#ifndef LLSSOLVER_H
#define LLSSOLVER_H

/** \file LLSSolver.cpp
    \brief Linear Least Square Solver

    This class is for solving Linear Least Square (LLS) problem.
    The LLS is A * x = b, where A is of size m by n, x is n by one,
    and b is m by one

    For robustness, the singular valuse decomposition (svd) is used.
    However, svd in lapack uses Fortran convention.
    note that A is a one-dimesional array,
    which store a 2D matrix using the Fortran comvension,
    i.e., store the first column, second column, and so on.
*/

#ifdef SINGLE
#define real float
#else
#define real double
#endif

/** 
    define the maximum value of \a x and \a y

*/
#define max(x,y) ((x)>(y)?(x):(y))

/** 
    define the minimum value of \a x and \a y

*/#define min(x,y) ((x)<(y)?(x):(y))

extern "C" {

  // these subroutines are from laint.f
  // which invokes the Lapack library

#ifdef SINGLE
void  ssvd_(real *G,real *U,real *V,int *M,int *N,
	    real *W,real *WORK,int *LWORK,int *INFO);
void  ssol_(real *U,real *W,real *V,int *M,int *N,
	    real *rhs,real *aa,real *tmp);
#else
void  dsvd_(real *G,real *U,real *V,int *M,int *N,
	    real *W,real *WORK,int *LWORK,int *INFO);
void  dsol_(real *U,real *W,real *V,int *M,int *N,
	    real *rhs,real *aa,real *tmp);
#endif
}

#include <iostream>
#include <string>
#include <cmath>
#include <ctime>
#include <vector>
#include <stdlib.h>
using namespace std;

/** 
    \brief Linear Least Square Solver

   Solving a non-square system of linear equations with singular value decomposition using a library from lapack
*/
class LLSSolver {
  
 public:

  /**
     constructor
   */
  LLSSolver();

/** 
    \brief Linear Least Square Solver

    This class is for solving Linear Least Square (LLS) problem.
    The LLS is A * x = b, where A is of size m by n, x is n by one,
    and b is m by one

    For robustness, the singular valuse decomposition (svd) is used.
    However, svd in lapack uses Fortran convention.
    note that A is a one-dimesional array,
    which store a 2D matrix using the Fortran comvension,
    i.e., store the first column, second column, and so on.
*/
  void solve(int matrixSizeM, /**< m of matrix A */
	     int matrixSizeN,  /**< n of matrix A */
	     real *A,  /**< matrix from pressure and vorticity */
	     real *b,  /**< vector from pressure and vorticity */
	     real *x, /**<  coefficients*/
	     vector<real> &UFixedParticles, /**<  stores U of the LLS solver  */
	     vector<real> &SFixedParticles, /**<  stores S of the LLS solver  */
	     vector<real> &VFixedParticles, /**<  stores V of the LLS solver  */
	     int &fixedParticles /**< stores if the particles are fixed */
	     ); 

  /**
     destructor
   */
  ~LLSSolver();

 private: 
  //  void initialize(int matrixSizeM, int matrixSizeN);

  void calculateUSV(real *);

  real * allocate1dArray(int size);

  void delete1dArray(real * x);
  void releaseMemory();
 
  // matrix A has the size m x n
  // where m is the number of rows, 
  //         or the number of total pressure and vorticity points
  //         or the number of equations
  // n is the number of columns
  //         or the number of undetermined coefficients
  //            for velocity, etc.
  int m, n;
  int i, j, iter;

  int nIter;

  // use convention A = U*S*V'
  // U with size M by M
  // V with size V by V
  // S with size M by N

  // but in the lapack library, S has the dimension min(M, N)
  // S is the singular values of A
  // so, here we use the convention from lapack
  // tmp is only used for a temporary purpose
  real *U, *V, *S, *tmp;


};

#endif
