#include "NSSolver.h"


NSSolver::NSSolver(){
}

void NSSolver::solve(int nPatch,
		     real t,
		     real dt,
		     real Re,
		     int nTruncate,
		     int cellNumberPerPatch, 
		     int ** StartIndex, 
		     real *xParticleCenter,  // coordinates of particle center
		     real *yParticleCenter, 
		     real *zParticleCenter, 
		     int ** pointNumberMeshU, 
		     int ** pointNumberMeshV, 
		     int ** pointNumberMeshW, 
		     int ** pointNumberMeshP, 
		     int ** pointNumberMeshOmegaX, 
		     int ** pointNumberMeshOmegaY, 
		     int ** pointNumberMeshOmegaZ, 
		     int dim,
		     real ***xU, 
		     real ***xV, 
		     real ***xW, 
		     real ***xP, 
		     real ***xOmegaX, 
		     real ***xOmegaY, 
		     real ***xOmegaZ, 
		     real *h,
		     real ****u,
		     real ****v,
		     real ****w,
		     real ****p,
		     real ****omegaX,
		     real ****omegaY,
		     real ****omegaZ,
		     real ****uAux,
		     real ****vAux,
		     real ****wAux,
		     real ****fp, 
		     real ****exactP, 
		     int ****pUnknowns,
		     int ****uCagePoint,
		     int ****vCagePoint,
		     int ****wCagePoint,
		     int ****pCagePoint,
		     int ****omegaXCagePoint,
		     int ****omegaYCagePoint,
		     int ****omegaZCagePoint,
		     vector<real> &UFixedParticles,
		     vector<real> &SFixedParticles,
		     vector<real> &VFixedParticles,
		     int &fixedParticles,
		     int rotatingParticles,
		     real **angVel,
		     real **angVelAcc,
		     real **torqueOld,
		     real **torqueNew,
		     PetscLinearSolver &pSolver,
		     int &zeroInitialGuess){

  NumberOfMeshes = nPatch + 1;

  uIter = allocate4dRealArray(NumberOfMeshes, pointNumberMeshU);
  vIter = allocate4dRealArray(NumberOfMeshes, pointNumberMeshV);
  wIter = allocate4dRealArray(NumberOfMeshes, pointNumberMeshW);

  nCoeff = 1 + 3 * nTruncate * (nTruncate + 2);
  solvedABCoeffAllParticleOld =
    allocate2dRealArray(NumberOfMeshes, nCoeff);

  solvedABCoeffAllParticleNew =
    allocate2dRealArray(NumberOfMeshes, nCoeff);

  // initialize
  for (m=0; m<=nPatch; m++) {
    for (i=0; i<nCoeff; i++) {
      solvedABCoeffAllParticleOld[m][i] = 0.0;
    }
  }

  for (m=0; m<=nPatch; m++) {
    for (i=0; i<nCoeff; i++) {
      solvedABCoeffAllParticleNew[m][i] = 0.0;
    }
  }

  //copy to arrays for the iteration of cage points

  //use the initial field as a guess: p and vorticity

  for (m=0; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshU[m][2]; k++) {
      for (j=0; j<pointNumberMeshU[m][1]; j++) {
	for (i=0; i<pointNumberMeshU[m][0]; i++) {
	  uIter[m][i][j][k] = u[m][i][j][k];
	}
      }
    }
  }

  for (m=0; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshV[m][2]; k++) {
      for (j=0; j<pointNumberMeshV[m][1]; j++) {
	for (i=0; i<pointNumberMeshV[m][0]; i++) {
	  vIter[m][i][j][k] = v[m][i][j][k];
	}
      }
    }
  }  

  for (m=0; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshW[m][2]; k++) {
      for (j=0; j<pointNumberMeshW[m][1]; j++) {
	for (i=0; i<pointNumberMeshW[m][0]; i++) {
	  wIter[m][i][j][k] = w[m][i][j][k];
	}
      }
    }
  }  


  // cout << "step = " << (int)(t/dt) <<endl;

  iter = 0;

  do {

    relError = 0.0;

    iter++;
    cout << "    iter = " << iter<<endl;

    //from uIter and  vIter, obtain the vorticity
    vorticity vObj;
    vObj.calOmegaFromUVW(nPatch,
			pointNumberMeshOmegaX, 
			pointNumberMeshOmegaY, 
			pointNumberMeshOmegaZ, 
			h,
			uIter,
			vIter,
			wIter,
			omegaX,
			omegaY,
			omegaZ);

    clock_t tOld, tNew;		 

    tOld = clock();

    //from the guess pressure and vorticity 
    //to find the analytical velocity 
    cage cObj;
    cObj.solve(nPatch,
	       cellNumberPerPatch,
	       dim,
	       nTruncate,
	       Re,
	       xParticleCenter, 
	       yParticleCenter,
	       zParticleCenter,
	       // for p
	       xP,
	       p,
	       pointNumberMeshP,
	       pCagePoint,
	       // for u
	       xU,
	       uIter,
	       pointNumberMeshU,
	       uCagePoint,
	       // for v
	       xV,
	       vIter,
	       pointNumberMeshV,
	       vCagePoint,
	       // for w
	       xW,
	       wIter,
	       pointNumberMeshW,
	       wCagePoint,
	       // for omega
	       xOmegaX,
	       omegaX,
	       pointNumberMeshOmegaX,
	       omegaXCagePoint,
	       // for omega
	       xOmegaY,
	       omegaY,
	       pointNumberMeshOmegaY,
	       omegaYCagePoint,
	       // for omega
	       xOmegaZ,
	       omegaZ,
	       pointNumberMeshOmegaZ,
	       omegaZCagePoint,
	       solvedABCoeffAllParticleNew,
	       UFixedParticles,
	       SFixedParticles,
	       VFixedParticles,
	       fixedParticles,
	       rotatingParticles,
	       angVel,
	       angVelAcc,
	       torqueOld,
	       torqueNew);

    tNew = clock();

    //    cout << "cage part: " << (tNew - tOld) / (double) CLOCKS_PER_SEC << " secs." << endl ;

    tOld = tNew;
    
    // relative errors between old and new coefficients
    //term by term
  for (m=1; m<=nPatch; m++) {
    for (i=0; i<nCoeff; i++) {
      if ( fabs(solvedABCoeffAllParticleNew[m][i] 
		- solvedABCoeffAllParticleOld[m][i])
	   / ( fabs(solvedABCoeffAllParticleNew[m][i]) + 1.e-4 ) 
	   > relError ) {
	relError = fabs(solvedABCoeffAllParticleNew[m][i] 
			- solvedABCoeffAllParticleOld[m][i])
	  / ( fabs(solvedABCoeffAllParticleNew[m][i]) + 1.e-4 );
      }
    }
  }      
    
  /* 
    // very possible that this doesn't work
    // L2-norm of relative errors
    for (m=1; m<=nPatch; m++) {
      
      real normOld, normNew;
      real sumXiSquaredOld = 0.0, sumXiSquaredNew = 0.0;
      int  nTerms = 0;
      
      for (i=0; i<nCoeff; i++) {
	sumXiSquaredOld += pow(solvedABCoeffAllParticleOld[m][i], 2.0);
	sumXiSquaredNew += pow(solvedABCoeffAllParticleNew[m][i], 2.0);
	nTerms += 1;
      }
      
      normOld = sqrt(sumXiSquaredOld/nTerms);
      normNew = sqrt(sumXiSquaredNew/nTerms);

      if ( fabs(normNew-normOld) / normNew > relError ) {
	relError = fabs(normNew-normOld) / normNew;
      }      
    }
  */

    for (m=1; m<=nPatch; m++) {
      for (i=0; i<nCoeff; i++) {
	solvedABCoeffAllParticleOld[m][i] = solvedABCoeffAllParticleNew[m][i];
      }
    }
    
    projectionMethod pObj;
    pObj.solve(nPatch,
	       t,
	       dt,
	       Re,
	       cellNumberPerPatch, 
	       StartIndex, 
	       pointNumberMeshU, 
	       pointNumberMeshV, 
	       pointNumberMeshW, 
	       pointNumberMeshP, 
	       dim,
	       xU, 
	       xV, 
	       xW, 
	       xP, 
	       h,
	       u,
	       uIter,
	       uCagePoint,
	       v,
	       vIter,
	       vCagePoint,
	       w,
	       wIter,
	       wCagePoint,
	       p,
	       uAux,
	       vAux,
	       wAux,
	       fp, 
	       exactP, 
	       pUnknowns,
	       pCagePoint,
	       pSolver,
	       zeroInitialGuess);

    tNew = clock();

    //    cout << "projection part: " << (tNew - tOld) / (double) CLOCKS_PER_SEC << " secs." << endl ;


    cout << "              relError = " << relError << endl;

    
  } while ( (relError > 1.e-2 || iter==1) );


  for (m=1; m<=nPatch; m++) {

    // store coefficients
    filename = "results/coeffParticle" + iToString(m) + ".dat";
  
    ofParticle.open( filename.c_str(),(ios_base::out | ios_base::app) ) ;

    ofParticle<< t << "  ";
    
    for (i=0; i<nCoeff; i++) {
      ofParticle<< solvedABCoeffAllParticleNew[m][i] << "  ";
    }
    ofParticle<< endl;
    
    ofParticle.close();

    if (rotatingParticles == 1) {
      updateAngVel(m, // mth particle
		   angVel,
		   angVelAcc,
		   torqueOld,
		   torqueNew,
		   t,
		   dt,
		   Re);
    }

    // store coefficients
    filename = "results/angVelParticle" + iToString(m) + ".dat";
  
    ofParticle.open( filename.c_str(),(ios_base::out | ios_base::app) ) ;

    ofParticle<< t << "  ";
    
    for (i=0; i<3; i++) {
      ofParticle<< angVel[i][m] << "  ";
    }
    ofParticle<< endl;
    
    ofParticle.close();


  }


  //copy back

  for (m=0; m<=nPatch; m++) {
    //       cout <<"============u==========="<<endl;
    for (k=0; k<pointNumberMeshU[m][2]; k++) {
      for (j=0; j<pointNumberMeshU[m][1]; j++) {
	for (i=0; i<pointNumberMeshU[m][0]; i++) {
	  u[m][i][j][k] = uIter[m][i][j][k];
	  //		cout << u[m][i][j] << " ";
	}
	//      cout << endl;
      }
    }
  }



  for (m=0; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshV[m][2]; k++) {
      for (j=0; j<pointNumberMeshV[m][1]; j++) {
	for (i=0; i<pointNumberMeshV[m][0]; i++) {
	  v[m][i][j][k] = vIter[m][i][j][k];
	}
      }
    }
  }

  for (m=0; m<=nPatch; m++) {
    //       cout <<"============u==========="<<endl;
    for (k=0; k<pointNumberMeshW[m][2]; k++) {
      for (j=0; j<pointNumberMeshW[m][1]; j++) {
	for (i=0; i<pointNumberMeshW[m][0]; i++) {
	  w[m][i][j][k] = wIter[m][i][j][k];
	  //		cout << u[m][i][j] << " ";
	}
	//      cout << endl;
      }
    }
  }



  delete4dArray(uIter, NumberOfMeshes, pointNumberMeshU);
  delete4dArray(vIter, NumberOfMeshes, pointNumberMeshV);
  delete4dArray(wIter, NumberOfMeshes, pointNumberMeshW);

  delete2dArray(solvedABCoeffAllParticleOld, NumberOfMeshes);
  delete2dArray(solvedABCoeffAllParticleNew, NumberOfMeshes);

}

//update angular velocity
void NSSolver::updateAngVel(int particle, // mth particle
			    real **angVel,
			    real **angVelAcc,
			    real **torqueOld,
			    real **torqueNew,
			    real t,
			    real dt,
			    real Re){
  
  /*  
  if ( t > 0.5) { //turn on the acceleration

    torqueNew[0][particle] = 30.0/7.0 * solvedABCoeffAllParticleNew[particle][8]/ pow(Re, 2.0);
    torqueNew[1][particle] = 30.0/7.0 * solvedABCoeffAllParticleNew[particle][9] / pow(Re, 2.0);
    torqueNew[2][particle] = 30.0/7.0 * solvedABCoeffAllParticleNew[particle][3] / pow(Re, 2.0);
    
    angVelAcc[0][particle] = (torqueOld[0][particle] + torqueNew[0][particle]) / 2.0;
    angVelAcc[1][particle] = (torqueOld[1][particle] + torqueNew[1][particle]) / 2.0;
    angVelAcc[2][particle] = (torqueOld[2][particle] + torqueNew[2][particle]) / 2.0;
    
  } else { // for some transitional stage
    
    torqueNew[0][particle] = 15.0 * solvedABCoeffAllParticleNew[particle][8]/ pow(Re, 2.0);
    torqueNew[1][particle] = 15.0 * solvedABCoeffAllParticleNew[particle][9] / pow(Re, 2.0);
    torqueNew[2][particle] = 15.0 * solvedABCoeffAllParticleNew[particle][3] / pow(Re, 2.0);
    
    angVelAcc[0][particle] = 0.0;
    angVelAcc[1][particle] = 0.0;
    angVelAcc[2][particle] = 0.0;
  }

  // trapzoidal rule
  angVel[0][particle] += dt * (torqueOld[0][particle] + torqueNew[0][particle]) / 2.0;
  angVel[1][particle] += dt * (torqueOld[1][particle] + torqueNew[1][particle]) / 2.0;
  angVel[2][particle] += dt * (torqueOld[2][particle] + torqueNew[2][particle]) / 2.0;
  */

  real TaylorNumber = 70.0;

  torqueNew[0][particle] = 0.0;
  torqueNew[1][particle] = 0.0;
  torqueNew[2][particle] = 0.0;
  
  angVelAcc[0][particle] = 0.0;
  angVelAcc[1][particle] = 0.0;
  angVelAcc[2][particle] = 0.0;  



  // set Omega_X
  /*   
  real t_end = -0.4;
  if ( t < t_end ) {
    angVel[0][particle] = TaylorNumber * t / t_end ;
    //  } else if ( t > 6.0 ) {
    //angVel[0][particle] = 0.0;
  } else {
    angVel[0][particle] = TaylorNumber;
  }
  angVel[1][particle] = 0.0;
  angVel[2][particle] = 0.0;
  */

  // set Omega_Y
      
  angVel[0][particle] = 0.0;
  angVel[2][particle] = 0.0;
  real t_end = -0.4;
  if ( t < t_end ) {
    angVel[1][particle] = TaylorNumber * t / t_end ;
    //  } else if ( t > 6.0 ) {
    //angVel[1][particle] = 0.0;
  } else {
    angVel[1][particle] = TaylorNumber;
  }
  
  
  // set Omega_Z
  /*
  angVel[0][particle] = 0.0;
  angVel[1][particle] = 0.0;
  real t_end = 0.4;
  if ( t < t_end ) {
    //    angVel[2][particle] = TaylorNumber * tanh( 1.0 / 0.5 * t);
    angVel[2][particle] = TaylorNumber * t / t_end ;
  } else {
    angVel[2][particle] = TaylorNumber;
  }
  */

  torqueOld[0][particle] = 0.0;
  torqueOld[1][particle] = 0.0;
  torqueOld[2][particle] = 0.0;

}

string NSSolver::iToString(int k){
  // Converts an integer to a string
  // Requires <sstream> include file

  stringstream s;
  s <<  k;
  return s.str();
}


real ** NSSolver::allocate2dRealArray(int nMeshes, int size){
  real **x = new real *[nMeshes];
  for (m=0; m<nMeshes; m++) {
    x[m] = new real [size];
  }
  return x;
}

void NSSolver::delete2dArray(real **x, int nMeshes){
  for (m=0; m<nMeshes; m++) {
    delete [] x[m];
  }
  delete [] x;
}

real **** NSSolver::allocate4dRealArray(int nMeshes, int **pointNumberMesh){
  real ****x = new real ***[nMeshes];
  for (m=0; m<nMeshes; m++) {
    x[m] = new real **[pointNumberMesh[m][0]];
    for (i=0; i<pointNumberMesh[m][0]; i++) {
      x[m][i] = new real*[pointNumberMesh[m][1]];
      for (j=0; j<pointNumberMesh[m][1]; j++) {
	x[m][i][j] = new real[pointNumberMesh[m][2]];
      }
    }
  }
  return x;
}

void NSSolver::delete4dArray(real ****x, int nMeshes, int **pointNumberMesh){
  for (m=0; m<nMeshes; m++) {
    for (i=0; i<pointNumberMesh[m][0]; i++) {
      for (j=0; j<pointNumberMesh[m][1]; j++) {
	delete [] x[m][i][j];
      }
      delete [] x[m][i];
    }
    delete [] x[m];
  }
  delete [] x;
}



NSSolver::~NSSolver(){
}
