#ifndef NSSOLVER_H
#define NSSOLVER_H

/** \file NSSolver.cpp
    \brief Navier-Stokes equation Solver 

    This class is the acutal Navier-Stokes equation Solver. It calls the classes cage for the inner velocity boundary conditions associated with the particle no-slip boundary conditions and projectionMethod to integate the Navier-Stokes equation.

*/


#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cmath>
#include <vector>
#include <cstdlib> // for rand() and srand()
#include <ctime>   // for time()
#include "projectionMethod.h"
#include "cage.h"
#include "vorticity.h"
using namespace std;

#ifdef SINGLE
#define real float
#else
#define real double
#endif

/** 
    \brief Navier-Stokes equation Solver 

    This class is the acutal Navier-Stokes equation Solver. It calls the classes cage for the inner velocity boundary conditions associated with the particle no-slip boundary conditions and projectionMethod to integate the Navier-Stokes equation.

*/
class NSSolver {

 public:

  /**
     constructor
   */
  NSSolver();

/** 
    \brief Navier-Stokes equation Solver 

    This class is the acutal Navier-Stokes equation Solver. It calls the classes cage for the inner velocity boundary conditions associated with the particle no-slip boundary conditions and projectionMethod to integate the Navier-Stokes equation.

   At the time step n, the field of u, v, and p are given. this class is used to calculate the fields in the next time step n+1

*/
  void solve(int nPatch,/**< number of patches */
	     real t,/**<  time */
	     real dt,/**<  time increment */
	     real Re,/**<  Reynolds number */
	     int nTruncate,/**<  the order for truncated spectral analytical solution*/
	     int cellNumberPerPatch, /**<  cell numbers per patch surrounding the sphere*/
	     int ** StartIndex, /**< used to put the refined mesh */
	     real *xParticleCenter,  /**< x coordinate of particle center */
	     real *yParticleCenter, /**< y coordinate of particle center */
	     real *zParticleCenter, /**< z coordinate of particle center */
	     int ** pointNumberMeshU, /**< mesh points for u */
	     int ** pointNumberMeshV, /**< mesh points for v */
	     int ** pointNumberMeshW, /**< mesh points for w */
	     int ** pointNumberMeshP, /**< mesh points for p */
	     int ** pointNumberMeshOmegaX, /**< mesh points for omegaX */
	     int ** pointNumberMeshOmegaY,  /**< mesh points for omegaY */
	     int ** pointNumberMeshOmegaZ,  /**< mesh points for omegaY */
	     int dim,/**<  dimension */
	     real ***xU, /**< coordinates of u  */
	     real ***xV, /**< coordinates of v  */
	     real ***xW, /**< coordinates of w  */
	     real ***xP, /**< coordinates of p  */
	     real ***xOmegaX, /**< coordinates of omegaX  */
	     real ***xOmegaY, /**< coordinates of omegaX  */
	     real ***xOmegaZ, /**< coordinates of omegaX  */
	     real *h,/**< space step  */
	     real ****u,/**< velocity u */
	     real ****v,/**< velocity v */
	     real ****w,/**< velocity w */
	     real ****p,/**<  pressure */
	     real ****omegaX,/**< vorticity omegaX */
	     real ****omegaY,/**< vorticity omegaY */
	     real ****omegaZ,/**< vorticity omegaZ */
	     real ****uAux,/**< predicted u */
	     real ****vAux,/**< predicted v */
	     real ****wAux,/**< predicted w */
	     real ****fp, /**< extra force */
	     real ****exactP, /**< exact pressure */
	     int ****pUnknowns,/**< mesh points for p */
	     int ****uCagePoint,/**< cage points for u */
	     int ****vCagePoint,/**< cage points for v */
	     int ****wCagePoint,/**< cage points for w */
	     int ****pCagePoint, /**< cage points for p */
	     int ****omegaXCagePoint,/**< cage points for omegaX */
	     int ****omegaYCagePoint,/**< cage points for omegaY */
	     int ****omegaZCagePoint,/**< cage points for omegaY */
	     vector<real> &UFixedParticles,/**< stores U of the LLS solver */
	     vector<real> &SFixedParticles,/**< stores S of the LLS solver */
	     vector<real> &VFixedParticles,/**< stores V of the LLS solver */
	     int &fixedParticles,/**< stores if the particles are fixed */
	     int rotatingParticles,/**< stores if the particles are rotating */
	     real **angVel,/**< angular velocity of particles */
	     real **angVelAcc,/**< angular velocity acceleration of particles */
	     real **torqueOld,/**< torque value in previous iteration of particles */
	     real **torqueNew,/**< torque value in current iteration of particles */
	     PetscLinearSolver &pSolver,/**< PETSC linear solver */
	     int &zeroInitialGuess/**< if use zero as the initial guess for pressure */
	     );

  /**
     destructor
   */
  ~NSSolver();

 private: 

  //update angular velocity
  void updateAngVel(int particle, // mth particle
		    real **angVel,
		    real **angVelAcc,
		    real **torqueOld,
		    real **torqueNew,
		    real t,
		    real dt,
		    real Re);
  
  string iToString(int k);
  real ** allocate2dRealArray(int nMeshes, int size);
  void delete2dArray(real **x, int nMeshes);
  real **** allocate4dRealArray(int nMeshes, int **pointNumberMesh);
  void delete4dArray(real ****x, int nMeshes, int **pointNumberMesh);

  int m, i, j, k, iter;

  int NumberOfMeshes;

  // note that u is needed to keep
  // thus, the following is needed for iteration
  real ****uIter; // solution corresponding to pUnknowns
  real ****vIter; // solution corresponding to pUnknowns
  real ****wIter; // solution corresponding to pUnknowns

  int nCoeff;
  real **solvedABCoeffAllParticleOld; 
  real **solvedABCoeffAllParticleNew; 
  real relError;

  string filename;
  ofstream ofParticle ;

};

#endif
