#include "PetscLinearSolver.h"

//   Concepts: KSP^solving a system of linear equations
//   AMatrix x = rhs
//   Processors: 1

#undef __FUNCT__
#define __FUNCT__ ""

PetscLinearSolver::PetscLinearSolver(){
}

void PetscLinearSolver::LinearSolver(int matrixSize,real **valueAMatrix,int **columnAMatrix, int totalColumn, real *rhs, real *Sol, PCType preCond, KSPType KrylovM, real tol, int &zeroInitialGuess) {

  int result;
  result = initialize(matrixSize,preCond,KrylovM, tol, zeroInitialGuess); // set parameters
  result = setSparseAMatrix(valueAMatrix, columnAMatrix, totalColumn, zeroInitialGuess); //set A matrix
  result = setRHS(rhs);  // set right hand side (RHS)
  result = solve(zeroInitialGuess); // solve the soution
  getSol(Sol); // output the solution
}


int PetscLinearSolver::initialize(int matrixSize,PCType preCond, KSPType KrylovM, real tol, int &zeroInitialGuess){

  if ( zeroInitialGuess == 1 ) {
  n = matrixSize;
  preConditioner = preCond;
  KrylovMethod = KrylovM;
  KSPTolerance = tol;


  PetscMPIInt    size;
  ierr = MPI_Comm_size(PETSC_COMM_WORLD,&size);CHKERRQ(ierr);
  if (size != 1) SETERRQ(1,"This is a uniprocessor example only!");
  ierr = PetscOptionsGetInt(PETSC_NULL,"-n",&n,PETSC_NULL);CHKERRQ(ierr);

  ierr = VecCreate(PETSC_COMM_WORLD,&x);CHKERRQ(ierr);
  ierr = PetscObjectSetName((PetscObject) x, "Solution");CHKERRQ(ierr);
  ierr = VecSetSizes(x,PETSC_DECIDE,n);CHKERRQ(ierr);
  ierr = VecSetFromOptions(x);CHKERRQ(ierr);
  }
  return 0;
}

int PetscLinearSolver::setRHS(real *rhs){

  int i;
  PetscScalar   rhsValue[1];
  PetscInt      rhsIndex[1];

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  //         Compute the matrix and right-hand-side vector that define
  //       the linear system, Ax = b.
  //   - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  //     Create vectors.  Note that we form 1 vector from scratch and
  //     then duplicate as needed.

  ierr = VecDuplicate(x,&b);CHKERRQ(ierr);


  for (i=0; i<n; i++) {
    rhsValue[0] = (PetscScalar)rhs[i];
    rhsIndex[0] = i;
    ierr = VecSetValues(b, 1,rhsIndex, rhsValue, INSERT_VALUES);CHKERRQ(ierr);  
  }
  VecAssemblyBegin(b);
  VecAssemblyEnd(b);

  return 0;
}


int PetscLinearSolver::setAMatrix(real **AMatrix){

  // A matrix in the form of 2d array

  PetscInt       i, j, col[1];
  PetscScalar    value[1];
  //     Create matrix.  When using MatCreate(), the matrix format can
  //     be specified at runtime.

  //     Performance tuning note:  For problems of substantial size,
  //     preallocation of matrix memory is crucial for attaining good 
  //     performance. See the matrix chapter of the users manual for details.


  ierr = MatCreate(PETSC_COMM_WORLD,&A);CHKERRQ(ierr);
  ierr = MatSetSizes(A,PETSC_DECIDE,PETSC_DECIDE,n,n);CHKERRQ(ierr);
  ierr = MatSetFromOptions(A);CHKERRQ(ierr);

  //     Assemble matrix
  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      if (fabs( AMatrix[i][j] ) > KSPTolerance) {
	value[0] = AMatrix[i][j];
	col[0] = j;
	ierr = MatSetValues(A,1,&i,1,col,value,INSERT_VALUES);CHKERRQ(ierr);
      }
    }
  }
  ierr = MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  ierr = MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);

  //   ierr = MatView(A,PETSC_VIEWER_STDOUT_WORLD);CHKERRQ(ierr);
  //  ierr = MatView(A,PETSC_VIEWER_DRAW_WORLD);CHKERRQ(ierr);
  
  return 0;
}


int PetscLinearSolver::setSparseAMatrix(real **AMatrix, int **columnAMatrix, int totalColumn, int &zeroInitialGuess){

  // A matrix in the form of 2d array

  if ( zeroInitialGuess == 1 ) {

  PetscInt       i, j, col[1];
  PetscScalar    value[1];
  //     Create matrix.  When using MatCreate(), the matrix format can
  //     be specified at runtime.

  //     Performance tuning note:  For problems of substantial size,
  //     preallocation of matrix memory is crucial for attaining good 
  //     performance. See the matrix chapter of the users manual for details.

  ierr = MatCreateSeqAIJ(PETSC_COMM_SELF,n, n,totalColumn,PETSC_NULL, &A);CHKERRQ(ierr);

  // inefficient due to:
  // dynamic memory allocation and copy from old storagy to the new one
  //  ierr = MatCreate(PETSC_COMM_WORLD,&A);CHKERRQ(ierr);
  //  ierr = MatSetSizes(A,n,n,n,n);CHKERRQ(ierr);
  //  ierr = MatSetFromOptions(A);CHKERRQ(ierr);

  //     Assemble matrix
  for (i=0; i<n; i++) {
    for (j=0; j<totalColumn; j++) {
     // Is it possible to remove the if statement
      // using certain data structure for efficiency?
      // Linked list is a candidate
      // need to modify the pressure class later
	value[0] = AMatrix[i][j];
	col[0] = columnAMatrix[i][j];
	ierr = MatSetValues(A,1,&i,1,col,value,INSERT_VALUES);CHKERRQ(ierr);
    }
  }
  ierr = MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  ierr = MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  }
  //   ierr = MatView(A,PETSC_VIEWER_STDOUT_WORLD);CHKERRQ(ierr);
  //  ierr = MatView(A,PETSC_VIEWER_DRAW_WORLD);CHKERRQ(ierr);

  return 0;
}



int PetscLinearSolver::solve(int &zeroInitialGuess)
{

  //                Create the linear solver and set various options

  //     Create linear solver context

  if ( zeroInitialGuess == 1 ) {
    ierr = KSPCreate(PETSC_COMM_WORLD,&ksp);CHKERRQ(ierr);

  //     Set operators. Here the matrix that defines the linear system
  //     also serves as the preconditioning matrix.

  ierr = KSPSetOperators(ksp,A,A,DIFFERENT_NONZERO_PATTERN);CHKERRQ(ierr);

  


  //     Set linear solver defaults for this problem (optional).
  //     - By extracting the KSP and PC contexts from the KSP context,
  //       we can then directly call any KSP and PC routines to set
  //       various options.
  //     - The following four statements are optional; all of these
  //       parameters could alternatively be specified at runtime via
  //       KSPSetFromOptions();

  ierr = KSPGetPC(ksp,&pc);CHKERRQ(ierr);
  ierr = PCSetType(pc,preConditioner);CHKERRQ(ierr);
  ierr = KSPSetTolerances(ksp,KSPTolerance,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT);CHKERRQ(ierr);

  ierr = KSPSetType(ksp,KrylovMethod);CHKERRQ(ierr);
  }
  //    Set runtime options, e.g.,
  //        -ksp_type <type> -pc_type <type> -ksp_monitor -ksp_rtol <rtol>
  //    These options will override those specified above as long as
  //    KSPSetFromOptions() is called _after_ any other customization
  //    routines.

  //  ierr = KSPSetFromOptions(ksp);CHKERRQ(ierr);


  if ( zeroInitialGuess == 1 ) {
    ierr = KSPSetInitialGuessNonzero(ksp, PETSC_TRUE); CHKERRQ(ierr);
    zeroInitialGuess = 0;
  }
 
  // Solve the linear system
  ierr = KSPSolve(ksp,b,x);CHKERRQ(ierr); 

  ierr = VecDestroy(b);CHKERRQ(ierr); 

  /* 
     View solver info; we could instead use the option -ksp_view to
     print this info to the screen at the conclusion of KSPSolve().
  */
  //  ierr = KSPView(ksp,PETSC_VIEWER_STDOUT_WORLD);CHKERRQ(ierr);


  return 0;
}


void PetscLinearSolver::getSol(real *Sol){

  int i;
  PetscScalar *array;
  VecGetArray(x, &array);
  for (i=0;i<n;i++) {
    Sol[i] = (real)*(array+i);
  }
  VecRestoreArray(x, &array);

}


void PetscLinearSolver::vecPrint(Vec v, int n){
  //  VecView(x, PETSC_VIEWER_STDOUT_SELF);

  int i;
  PetscScalar *array;
  VecGetArray(v, &array);
  for (i=0;i<n;i++)
    PetscPrintf(PETSC_COMM_WORLD, " %g ", (float)*(array+i));
  printf("\n");
  VecRestoreArray(v, &array);
}

int PetscLinearSolver::finalize( ){
  //  ierr = KSPGetIterationNumber(ksp,&its);CHKERRQ(ierr);
  //  ierr = PetscPrintf(PETSC_COMM_WORLD,"Iterations %D\n",its);CHKERRQ(ierr);

  //     Free work space.  All PETSc objects should be destroyed when they
  //     are no longer needed.
  ierr = VecDestroy(x);CHKERRQ(ierr);
  ierr = MatDestroy(A);CHKERRQ(ierr);
  ierr = KSPDestroy(ksp);CHKERRQ(ierr);

  return 0;
}

PetscLinearSolver::~PetscLinearSolver(){

}
