#ifndef PETSCLINEARSOLVER_H
#define PETSCLINEARSOLVER_H

/** \file PetscLinearSolver.cpp
    \brief Linear system solver

    This class use PETSC to solve the linear system A x = b, which results from the pressure Poisson equation. A is the matrix, x is the soultion of pressure, and b is the source terms. Note that A is represented by a sparse matrix for efficiency. The performace of the solver is mainly determined by this part, because the Poisson equation is the most expensive part to solve.

*/

#include "petscksp.h"
#include <iostream>
#include <string>
using namespace std;

#ifdef SINGLE
#define real float
#else
#define real double
#endif


/** 
    \brief Linear system solver

    This class use PETSC to solve the linear system A x = b, which results from the pressure Poisson equation. A is the matrix, x is the soultion of pressure, and b is the source terms. Note that A is represented by a sparse matrix for efficiency. The performace of the solver is mainly determined by this part, because the Poisson equation is the most expensive part to solve.

    Include "petscksp.h" so that we can use KSP solvers.  Note that this file
    automatically includes:

    petsc.h       - base PETSc routines,    petscvec.h - vectors ,
    petscsys.h    - system routines ,      petscmat.h - matrices,
    petscis.h     - index sets ,           petscksp.h - Krylov subspace methods,
    petscviewer.h - viewers ,              petscpc.h  - preconditioners,

*/
class PetscLinearSolver {
  
 public:
  /**
     constructor
   */
  PetscLinearSolver();

/** 
    \brief Linear system solver

    This class use PETSC to solve the linear system A x = b, which results from the pressure Poisson equation. A is the matrix, x is the soultion of pressure, and b is the source terms. Note that A is represented by a sparse matrix for efficiency. The performace of the solver is mainly determined by this part, because the Poisson equation is the most expensive part to solve.

    Concepts: KSP^solving a system of linear equations: AMatrix x = rhs

    Processors: 1

*/
  void LinearSolver(int matrixSize, /**<  matrix size */
		    real **valueAMatrix, /**<  matrix elements */
		    int **columnAMatrix, /**<  matrix index */
		    int totalColumn, /**<  total number of columns */
		    real *rhs ,/**<  mright hand side of the matrix */
		    real *Sol,  /**<  solution of the linear system */
		    PCType preCond, /**<  preconditioner for PETSC */
		    KSPType KrylovM, /**<  Krylov method */
		    real tol, /**<  tolerance */
		    int &zeroInitialGuess/**< if use zero as the initial guess for pressure */
		    );

  /**
     clean up the memory assigned for PETSC
   */
  int finalize( );

  /**
     destructor
   */
  ~PetscLinearSolver();

 private: 
  int initialize(int matrixSize,PCType preCond, KSPType KrylovM, real tol, int &zeroInitialGuess);
  int setRHS(real *rhs);
  int setAMatrix(real **AMatrix);
  int setSparseAMatrix(real **AMatrix, int **columnAMatrix, int totalColumn, int &zeroInitialGuess);
  int solve(int &zeroInitialGuess);
  void vecPrint(Vec x, int n);
  void getSol(real *Sol);
  
  PetscInt       n, its;
  PCType         preConditioner;
  KSPType        KrylovMethod;

  Vec            x, b;      // approx solution, RHS
  Mat            A;         // linear system matrix

  KSP            ksp;       // linear solver context
  PC             pc;        // preconditioner context
  PetscErrorCode ierr;
  real           KSPTolerance;
};

#endif
