#include "analyticalSol.h"

//   construct analytical solution in one cage for pressure, vorticity,
//   velocities at those cage points

// write everything in the form, using pressure as an example,
// p = pWeight dotproduct ABCoeff
// where 
// pWeight = [pWeightA0, pWeightB0, 
//            pWeightA1, pWeightATilde1,pWeightB1, pWeightBTilde1,
//            pWeightAK, pWeightATildeK,pWeightBK, pWeightBTildeK]
// and
// ABCoeff      = [A0, B0, 
//            A1, ATilde1,B1, BTilde1,
//            AK, ATildeK,BK, BTildeK]
//            where B0 denotes p0 * Re^2
//
// similarly,
// vorticity = omegaWeight dotproduct ABCoeff
// uR = uRWeight dotproduct ABCoeff
// uTheta = uThetaWeight dotproduct ABCoeff
// u = uWeight dotproduct ABCoeff
// v = vWeight dotproduct ABCoeff

// ABCoeff is to be determined from weightMatrix ABCoeff = b

// note that r is the normalized distance from the origin 

// Formula are in spherical coordinate:
// p, u, v, w, omegaX, omegaY, and omegaZ

analyticalSol::analyticalSol(){
}

void analyticalSol::solve(vector<real> &p,
			  const vector<real> &rP,
			  const vector<real> &thetaP,
			  const vector<real> &phiP,
			  vector<real> &omegaX,
			  const vector<real> &rOmegaX,
			  const vector<real> &thetaOmegaX,
			  const vector<real> &phiOmegaX,
			  vector<real> &omegaY,
			  const vector<real> &rOmegaY,
			  const vector<real> &thetaOmegaY,
			  const vector<real> &phiOmegaY,
			  vector<real> &omegaZ,
			  const vector<real> &rOmegaZ,
			  const vector<real> &thetaOmegaZ,
			  const vector<real> &phiOmegaZ,
			  vector<real> &u,
			  const vector<real> &rU,
			  const vector<real> &thetaU,
			  const vector<real> &phiU,
			  vector<real> &v,
			  const vector<real> &rV,
			  const vector<real> &thetaV,
			  const vector<real> &phiV,
			  vector<real> &w,
			  const vector<real> &rW,
			  const vector<real> &thetaW,
			  const vector<real> &phiW,
			  int nTruncate,
			  real ReynoldsNumber,
			  vector<real> &solvedABCoeffOneParticle,
			  vector<real> &UFixedParticles,
			  vector<real> &SFixedParticles,
			  vector<real> &VFixedParticles,
			  int &fixedParticles,
			  int rotatingParticles,
			  int particle, // mth particle
			  real **angVel,
			  real **angVelAcc,
			  real **torqueOld,
			  real **torqueNew){

  // all input and output fields of pressure, velocity, and vorticity
  // are from the finite-difference scheme
  // i.e., in the original inertial frame
  
  nP = p.size();

  nOmegaX = omegaX.size();
  nOmegaY = omegaY.size();
  nOmegaZ = omegaZ.size();

  nEquations = nP + nOmegaX + nOmegaY + nOmegaZ;

  Nc = nTruncate;  //degree
  nCoeff = 1 + 3 * Nc * (Nc + 2);

  Re = ReynoldsNumber;

  allocateMemory();
  solvedABCoeffOneParticle.assign(nCoeff, 0.0);

  // for rotating particles
  if (rotatingParticles == 1) {
    modifiedFields(p,
		   rP,
		   thetaP,
		   phiP,
		   omegaX,
		   rOmegaX,
		   thetaOmegaX,
		   phiOmegaX,
		   omegaY,
		   rOmegaY,
		   thetaOmegaY,
		   phiOmegaY,
		   omegaZ,
		   rOmegaZ,
		   thetaOmegaZ,
		   phiOmegaZ,
		   particle, // mth particle
		   angVel,
		   angVelAcc);
  }

  //solve coeffcients from pressure and vorticity points
  solveABCoeff(p, rP, thetaP,phiP,
	       omegaX, rOmegaX, thetaOmegaX, phiOmegaX,
	       omegaY, rOmegaY, thetaOmegaY, phiOmegaY,
	       omegaZ, rOmegaZ, thetaOmegaZ, phiOmegaZ,
	       UFixedParticles,
	       SFixedParticles,
	       VFixedParticles,
	       fixedParticles);

  for (i=0; i<nCoeff;i++){
    solvedABCoeffOneParticle[i] = ABCoeff[i];
    //    cout <<"coefficients:" <<endl;
    //    cout << "i = " <<i <<": " << ABCoeff[i]<<endl;
  }
 

  nU =  u.size();
  nV =  v.size();
  nW =  w.size();

  // allocate memory
  velocityWeight.assign(nCoeff, 0.0);

  //  update velocity
  //assign analytical velocity
  updateVelocity(u, rU, thetaU, phiU, 
		 v, rV, thetaV, phiV,
		 w, rW, thetaW, phiW);

  // for rotating particles
  if (rotatingParticles == 1) {
    modifiedVelocity(u,
		     rU,
		     thetaU,
		     phiU,
		     v,
		     rV,
		     thetaV,
		     phiV,
		     w,
		     rW,
		     thetaW,
		     phiW,
		     particle, // mth particle
		     angVel,
		     angVelAcc);
  }

  releaseMemory();   

}

// modify fields of pressure and vorticity due to rotation
void analyticalSol::modifiedFields(vector<real> &p,
				   const vector<real> &rP,
				   const vector<real> &thetaP,
				   const vector<real> &phiP,
				   vector<real> &omegaX,
				   const vector<real> &rOmegaX,
				   const vector<real> &thetaOmegaX,
				   const vector<real> &phiOmegaX,
				   vector<real> &omegaY,
				   const vector<real> &rOmegaY,
				   const vector<real> &thetaOmegaY,
				   const vector<real> &phiOmegaY,
				   vector<real> &omegaZ,
				   const vector<real> &rOmegaZ,
				   const vector<real> &thetaOmegaZ,
				   const vector<real> &phiOmegaZ,
				   int particle, // mth particle
				   real **angVel,
				   real **angVelAcc){
  
  // modify pressure
  for (i=0; i<nP;i++) {

    real x, y, z;
    x = rP[i] * cos(phiP[i]) * sin(thetaP[i]);
    y = rP[i] * sin(phiP[i]) * sin(thetaP[i]);
    z = rP[i] * cos(thetaP[i]);

    p[i] -= ( pow( angVel[1][particle] * z - angVel[2][particle] * y, 2.0)
	      + pow( angVel[0][particle] * z - angVel[2][particle] * x, 2.0)
	      + pow( angVel[0][particle] * y - angVel[1][particle] * x, 2.0) 
	      ) / 2.0;      
  }

  // modify omegaX
  for (i=0; i<nOmegaX;i++){

    real x, y, z;
    x = rOmegaX[i] * cos(phiOmegaX[i]) * sin(thetaOmegaX[i]);
    y = rOmegaX[i] * sin(phiOmegaX[i]) * sin(thetaOmegaX[i]);
    z = rOmegaX[i] * cos(thetaOmegaX[i]);

    omegaX[i] -= 2.0 * angVel[0][particle]
      + ( angVelAcc[0][particle] * (4.0 * pow(rOmegaX[i], 2.0) + 1.0/pow(rOmegaX[i], 3.0) )
	  - ( angVelAcc[0][particle] * x
	      + angVelAcc[1][particle] * y
	      + angVelAcc[2][particle] * z )
	  * x * (2.0 + 3.0 / pow(rOmegaX[i], 5.0)) ) * Re / 10.0;     
  }

  // modify omegaY
  for (i=0; i<nOmegaY;i++){

    real x, y, z;
    x = rOmegaY[i] * cos(phiOmegaY[i]) * sin(thetaOmegaY[i]);
    y = rOmegaY[i] * sin(phiOmegaY[i]) * sin(thetaOmegaY[i]);
    z = rOmegaY[i] * cos(thetaOmegaY[i]);

    omegaY[i] -= 2.0 * angVel[1][particle]
      + ( angVelAcc[1][particle] * (4.0 * pow(rOmegaY[i], 2.0) + 1.0/pow(rOmegaY[i], 3.0) )
	  - ( angVelAcc[0][particle] * x
	      + angVelAcc[1][particle] * y
	      + angVelAcc[2][particle] * z )
	  * y * (2.0 + 3.0 / pow(rOmegaY[i], 5.0)) ) * Re / 10.0;    
  }

  // modify omegaZ
  for (i=0; i<nOmegaZ;i++){

    real x, y, z;
    x = rOmegaZ[i] * cos(phiOmegaZ[i]) * sin(thetaOmegaZ[i]);
    y = rOmegaZ[i] * sin(phiOmegaZ[i]) * sin(thetaOmegaZ[i]);
    z = rOmegaZ[i] * cos(thetaOmegaZ[i]);

    omegaZ[i] -= 2.0 * angVel[2][particle]
      + ( angVelAcc[2][particle] * (4.0 * pow(rOmegaZ[i], 2.0) + 1.0/pow(rOmegaZ[i], 3.0) )
	  - ( angVelAcc[0][particle] * x
	      + angVelAcc[1][particle] * y
	      + angVelAcc[2][particle] * z )
	  * z * (2.0 + 3.0 / pow(rOmegaZ[i], 5.0)) ) * Re / 10.0 ;    
  }

}


// modify fields of velocity due to rotation
void analyticalSol::modifiedVelocity(vector<real> &u,
				     const vector<real> &rU,
				     const vector<real> &thetaU,
				     const vector<real> &phiU,
				     vector<real> &v,
				     const vector<real> &rV,
				     const vector<real> &thetaV,
				     const vector<real> &phiV,
				     vector<real> &w,
				     const vector<real> &rW,
				     const vector<real> &thetaW,
				     const vector<real> &phiW,
				     int particle, // mth particle
				     real **angVel,
				     real **angVelAcc){

  // modify U
  for (i=0; i<nU;i++){

    real x, y, z;
    x = rU[i] * cos(phiU[i]) * sin(thetaU[i]);
    y = rU[i] * sin(phiU[i]) * sin(thetaU[i]);
    z = rU[i] * cos(thetaU[i]);

    u[i] += angVel[1][particle] * z - angVel[2][particle] * y
      + ( angVelAcc[1][particle] * z - angVelAcc[2][particle] * y )
      * (pow(rU[i], 2.0) - 1.0 / pow(rU[i], 3.0)) * Re / 10.0 ;    
  }

  // modify V
  for (i=0; i<nV;i++){

    real x, y, z;
    x = rV[i] * cos(phiV[i]) * sin(thetaV[i]);
    y = rV[i] * sin(phiV[i]) * sin(thetaV[i]);
    z = rV[i] * cos(thetaV[i]);

    v[i] -= angVel[0][particle] * z - angVel[2][particle] * x
      + ( angVelAcc[0][particle] * z - angVelAcc[2][particle] * x )
      * (pow(rV[i], 2.0) - 1.0 / pow(rV[i], 3.0)) * Re / 10.0 ;    
  }

  // modify W
  for (i=0; i<nW;i++){

    real x, y, z;
    x = rW[i] * cos(phiW[i]) * sin(thetaW[i]);
    y = rW[i] * sin(phiW[i]) * sin(thetaW[i]);
    z = rW[i] * cos(thetaW[i]);

    w[i] += angVel[0][particle] * y - angVel[1][particle] * x
      + ( angVelAcc[0][particle] * y - angVelAcc[1][particle] * x )
      * (pow(rW[i], 2.0) - 1.0 / pow(rW[i], 3.0)) * Re / 10.0 ;    
  }


}


/////////////////////////////////////////////////
/////////////////////////////////////////////////
// solve AB coefficients
/////////////////////////////////////////////////
/////////////////////////////////////////////////

// solving coefficients of A, B, and their tilde counterparts
void analyticalSol::solveABCoeff(const vector<real> &p,
				 const vector<real> &rP,
				 const vector<real> &thetaP,
				 const vector<real> &phiP,
				 const vector<real> &omegaX,
				 const vector<real> &rOmegaX,
				 const vector<real> &thetaOmegaX,
				 const vector<real> &phiOmegaX,
				 const vector<real> &omegaY,
				 const vector<real> &rOmegaY,
				 const vector<real> &thetaOmegaY,
				 const vector<real> &phiOmegaY,
				 const vector<real> &omegaZ,
				 const vector<real> &rOmegaZ,
				 const vector<real> &thetaOmegaZ,
				 const vector<real> &phiOmegaZ,
				 vector<real> &UFixedParticles,
				 vector<real> &SFixedParticles,
				 vector<real> &VFixedParticles,
				 int &fixedParticles){

  // to solve the non-square linear system
  // weightMatrix ABCoeff = b

  // allocate memory for three arrays
  // note that the SVD needs arrays as arguments

  setB(p, omegaX, omegaY, omegaZ);

  setWeightMatrix(rP, thetaP, phiP,
		  rOmegaX, thetaOmegaX, phiOmegaX,
		  rOmegaY, thetaOmegaY, phiOmegaY,
		  rOmegaZ, thetaOmegaZ, phiOmegaZ);


  //  use LLSSolver to solve ABCoeff;
  // note: m > n
  // nEquations = m = matrixsizeM
  // nCoeff = n = matrixsizeN
  // note that the SVD needs arrays as arguments
  LLSSolver solver;
  solver.solve(nEquations, nCoeff, weightMatrix, b, ABCoeff,
	       UFixedParticles,
	       SFixedParticles,
	       VFixedParticles,
	       fixedParticles);
 
}


void analyticalSol::setB(const vector<real> &p,
			 const vector<real> &omegaX,
			 const vector<real> &omegaY,
			 const vector<real> &omegaZ){

  // note that the factor Re^2 for the nondimensional form
  for (i=0; i<nP;i++){
    b[i] = p[i] * Re * Re; //need to check
  }
  
  // note that the factor Re for the nondimensional form
  for (i=0; i<nOmegaX;i++){
    b[i + nP] = omegaX[i] * Re; //need to check
  }

  // note that the factor Re for the nondimensional form
  for (i=0; i<nOmegaY;i++){
    b[i + nP + nOmegaX] = omegaY[i] * Re; //need to check
  }

  // note that the factor Re for the nondimensional form
  for (i=0; i<nOmegaZ;i++){
    b[i + nP + nOmegaX + nOmegaY] = omegaZ[i] * Re; //need to check
  }
  
}

void analyticalSol::setWeightMatrix(const vector<real> &rP,
				    const vector<real> &thetaP,
				    const vector<real> &phiP,
				    const vector<real> &rOmegaX,
				    const vector<real> &thetaOmegaX,
				    const vector<real> &phiOmegaX,
				    const vector<real> &rOmegaY,
				    const vector<real> &thetaOmegaY,
				    const vector<real> &phiOmegaY,
				    const vector<real> &rOmegaZ,
				    const vector<real> &thetaOmegaZ,
				    const vector<real> &phiOmegaZ){
  //from p
  setWeightMatrixFromP(rP, thetaP, phiP);
  //from omegaX
  setWeightMatrixFromOmega(rOmegaX, thetaOmegaX, phiOmegaX, 
			   nP, nP + nOmegaX, 1);
  //from omegaY
  setWeightMatrixFromOmega(rOmegaY, thetaOmegaY, phiOmegaY,
			   nP + nOmegaX, nP + nOmegaX + nOmegaY, 2);
  //for omegaZ
  setWeightMatrixFromOmega(rOmegaZ, thetaOmegaZ, phiOmegaZ,
			   nP + nOmegaX + nOmegaY, nEquations, 3);

}

void analyticalSol::setWeightMatrixFromP(const vector<real> &rP,
					 const vector<real> &thetaP,
					 const vector<real> &phiP){
  
  for (i=0; i<nP;i++){
    // normalized r, theta, and phi spherical coords.
    
    r = rP[i];
    theta = thetaP[i];
    phi = phiP[i];

    cosTheta = cos(theta);

    ldp = Nc + 1;
#ifdef SINGLE
    salegpsd_(&Nc,&Nc,&cosTheta,pl,ps,pd,&ldp);
#else
    alegpsd_(&Nc,&Nc,&cosTheta,pl,ps,pd,&ldp);
#endif
    ldp = Nc + 2;

    // n=0
    // A_0^0
    j = 0;
    weightMatrix[i + j*nEquations] = 1./sqrt(pi); // doesn't matter?  don't know why
    j++;
    
    for(n=1; n<=Nc;n++){

      rToN = pow(r, n);
      real c1 = rToN - n*(2.0*n-1.0)/2.0/(n+1.0)/rToN/r;
      real c2 = -n*(4.0*n*n-1.0)/(n+1.0)/rToN/r;

      for ( m = 0; m <= n; m++ ) {

	if( m == 0 ) { 
	  //A_n^0
	  weightMatrix[i + j*nEquations] = c1*pl[n];
	  j++;
	  //C_n^0
	  weightMatrix[i + j*nEquations] = c2*pl[n];
	  j++;
	  //E_n^0
	  weightMatrix[i + j*nEquations] = 0.0;
	  j++;
	} else { 
	  //A_n^m and B_n^m
	  weightMatrix[i + j*nEquations]  = c1*pl[n+ldp*m]*cos(m*phi);
	  j++;
	  weightMatrix[i + j*nEquations]  = c1*pl[n+ldp*m]*sin(m*phi);
	  j++;
	  
	  //C_n^m and D_n^m
	  weightMatrix[i + j*nEquations]  = c2*pl[n+ldp*m]*cos(m*phi);
	  j++;
	  weightMatrix[i + j*nEquations]  = c2*pl[n+ldp*m]*sin(m*phi);
	  j++;
	  
	  //E_n^m and F_n^m
	  weightMatrix[i + j*nEquations]  = 0.0;
	  j++;
	  weightMatrix[i + j*nEquations]  = 0.0;
	  j++;
	} //for if statement
      
      } //for m
    
    } //for n
    
  } //for i

} //for this function


void analyticalSol::setWeightMatrixFromOmega(const vector<real> &rOmega,
					     const vector<real> &thetaOmega,
					     const vector<real> &phiOmega,
					     int nStart,
					     int nEnd,
					     int vorticityXYZ){
  // omegaX: vorticityXYZ = 1
  // omegaY: vorticityXYZ = 2
  // omegaZ: vorticityXYZ = 3

  for (i=nStart; i<nEnd;i++){
    
    // normalized r, theta polar coords.
    // n is the order of the harmonic functions 
    
    r = rOmega[i-nStart];
    theta = thetaOmega[i-nStart];
    phi = phiOmega[i-nStart];
    
    cosTheta = cos(theta);
    
    ldp = Nc + 1;
#ifdef SINGLE
    salegpsd_(&Nc,&Nc,&cosTheta,pl,ps,pd,&ldp);
#else
    alegpsd_(&Nc,&Nc,&cosTheta,pl,ps,pd,&ldp);
#endif
    ldp = Nc + 2;
    
    // n=0
    // A_0^0
    j = 0;
    weightMatrix[i + j*nEquations] = 0.0;
    j++;
    
    // pn
    for( n = 1; n<= Nc; n++ ) {

      rToN = pow(r,n);
      real c1 = n*(n+1.0)*(rToN/r - 1.0/r/r/rToN);
      real c2 =-rToN/(n+1.0) - (2.0*n-1.0)/2.0/(n+1.0)/rToN/r;
      real c3 =-(4*n*n-1.0)/r/(n+1.0)/rToN;
      real c4 = (n+1.0)*rToN/r + n/r/r/rToN;

      for ( m = 0; m <= n; m++ ) {

	if( m == 0 ) { 
	  //A_n^0
	  omegaR = 0.0;
	  omegaTheta = 0.0;
	  omegaPhi = -c2*pd[n];
	  weightMatrix[i + j*nEquations] 
	    = sphericalToCartesian(omegaR, omegaTheta, omegaPhi, theta, phi, vorticityXYZ);
	  j++;
	  //C_n^0
	  omegaR = 0.0;
	  omegaTheta = 0.0;
	  omegaPhi = -c3*pd[n];
	  weightMatrix[i + j*nEquations] 
	    = sphericalToCartesian(omegaR, omegaTheta, omegaPhi, theta, phi, vorticityXYZ);
	  j++;
	  //E_n^0
	  omegaR = c1*pl[n];
	  omegaTheta = c4*pd[n];
	  omegaPhi = 0.0;
	  weightMatrix[i + j*nEquations] 
	    = sphericalToCartesian(omegaR, omegaTheta, omegaPhi, theta, phi, vorticityXYZ);
	  j++;
	}
	else { 
	  //A_n^m and B_n^m
	  omegaR = 0.0;
	  omegaTheta = -m*c2*ps[n+ldp*m]*sin(m*phi);
	  omegaPhi = -c2*pd[n+ldp*m]*cos(m*phi); 
	  weightMatrix[i + j*nEquations] 
	    = sphericalToCartesian(omegaR, omegaTheta, omegaPhi, theta, phi, vorticityXYZ);
	  j++;
	  omegaR = 0.0;
	  omegaTheta = m*c2*ps[n+ldp*m]*cos(m*phi);
	  omegaPhi = -c2*pd[n+ldp*m]*sin(m*phi); 
	  weightMatrix[i + j*nEquations] 
	    = sphericalToCartesian(omegaR, omegaTheta, omegaPhi, theta, phi, vorticityXYZ);
	  j++;
	  //C_n^m and D_n^m
	  omegaR = 0.0;
	  omegaTheta = -m*c3*ps[n+ldp*m]*sin(m*phi);
	  omegaPhi = -c3*pd[n+ldp*m]*cos(m*phi);
	  weightMatrix[i + j*nEquations] 
	    = sphericalToCartesian(omegaR, omegaTheta, omegaPhi, theta, phi, vorticityXYZ);
	  j++;
	  omegaR = 0.0;
	  omegaTheta = m*c3*ps[n+ldp*m]*cos(m*phi);
	  omegaPhi = -c3*pd[n+ldp*m]*sin(m*phi);
	  weightMatrix[i + j*nEquations] 
	    = sphericalToCartesian(omegaR, omegaTheta, omegaPhi, theta, phi, vorticityXYZ);
	  j++;
	  //E_n^m and F_n^m
	  omegaR = c1*pl[n+ldp*m]*cos(m*phi);
	  omegaTheta = c4*pd[n+ldp*m]*cos(m*phi);
	  omegaPhi = -m*c4*ps[n+ldp*m]*sin(m*phi);
	  weightMatrix[i + j*nEquations] 
	    = sphericalToCartesian(omegaR, omegaTheta, omegaPhi, theta, phi, vorticityXYZ);
	  j++;
	  omegaR = c1*pl[n+ldp*m]*sin(m*phi);
	  omegaTheta = c4*pd[n+ldp*m]*sin(m*phi);
	  omegaPhi = m*c4*ps[n+ldp*m]*cos(m*phi);
	  weightMatrix[i + j*nEquations] 
	    = sphericalToCartesian(omegaR, omegaTheta, omegaPhi, theta, phi, vorticityXYZ);
	  j++;
	} // for if
      } // for m
    } //for n
    
  } //for i
  
} //for this function




/////////////////////////////////////////////////
/////////////////////////////////////////////////
// update velocity
/////////////////////////////////////////////////
/////////////////////////////////////////////////

void analyticalSol::updateVelocity(vector<real> &u,
				   const vector<real> &rU,
				   const vector<real> &thetaU,
				   const vector<real> &phiU,
				   vector<real> &v,
				   const vector<real> &rV,
				   const vector<real> &thetaV,
				   const vector<real> &phiV,
				   vector<real> &w,
				   const vector<real> &rW,
				   const vector<real> &thetaW,
				   const vector<real> &phiW){
  // note that r denotes the normalized distance
  
  for (int j=0; j<nU;j++){
    setVelocityWeight(rU[j], thetaU[j], phiU[j], 1);
    
    u[j] = 0.0;
    for (i=0; i<nCoeff;i++){
      u[j] += velocityWeight[i] * ABCoeff[i] / Re; // Re from nondimensional form
    }
  }
  
  for (int j=0; j<nV;j++){
    setVelocityWeight(rV[j], thetaV[j], phiV[j], 2);
    
    v[j] = 0.0;
    for (i=0; i<nCoeff;i++){
      v[j] += velocityWeight[i] * ABCoeff[i] / Re; // Re from nondimensional form
    }
  }

  for (int j=0; j<nV;j++){
    setVelocityWeight(rW[j], thetaW[j], phiW[j], 3);
    
    w[j] = 0.0;
    for (i=0; i<nCoeff;i++){
      w[j] += velocityWeight[i] * ABCoeff[i] / Re; // Re from nondimensional form
    }
  }

}


void analyticalSol::setVelocityWeight(real rVelocity, 
				      real thetaVelocity,
				      real phiVelocity,
				      int velocityXYZ){

  // for a point
  // note that r denotes the normalized distance
  // theta and phi are the angle   

  // u: velocityXYZ = 1
  // v: velocityXYZ = 2
  // w: velocityXYZ = 3

  r = rVelocity;
  theta = thetaVelocity;
  phi = phiVelocity;
  
  cosTheta = cos(theta);
  
  ldp = Nc + 1;
#ifdef SINGLE
  salegpsd_(&Nc,&Nc,&cosTheta,pl,ps,pd,&ldp);
#else
  alegpsd_(&Nc,&Nc,&cosTheta,pl,ps,pd,&ldp);
#endif
  ldp = Nc + 2;
  
  // n=0
  // A_0^0
  j = 0;
  velocityWeight[j] = 0.0;
  j++;
  
  // pn
  for( n = 1; n<= Nc; n++ ) {
    
    rToN = pow(r,n);
    real c1 =n/2.0/(2.0*n+3.0)*rToN*r
      -n/4.0/rToN+n/4.0*(2.0*n+1.0)/(2.0*n+3.0)/rToN/r/r ;
    real c2 =n*rToN/r
      -0.5*n*(2.0*n+1.0)/rToN+0.5*n*(2.0*n-1.0)/rToN/r/r;
    real c3 =(n+3.0)/2.0/(2.0*n+3.0)/(n+1.0)*rToN*r
      +(n-2.0)/4.0/(n+1.0)/rToN 
      -n/4.0*(2.0*n+1.0)/(2.0*n+3.0)/(n+1.0)/rToN/r/r;
    real c4 =rToN/r+(n-2.0)/2.0/(n+1.0)*(2.0*n+1.0)/rToN
      -0.5*n*(2.0*n-1.0)/(n+1.0)/rToN/r/r ;
    real c5 =rToN - 1./rToN/r;
    
    for ( m = 0; m <= n; m++ ) {
      
      if( m == 0 ) { 
	//A_n^0
	uR = c1*pl[n];
	uTheta = c3*pd[n];
	uPhi = 0.0;
	velocityWeight[j]  
	  = sphericalToCartesian(uR, uTheta, uPhi, theta, phi, velocityXYZ);
	j++;
	//C_n^0
	uR = c2*pl[n];
	uTheta = c4*pd[n];
	uPhi = 0.0;
	velocityWeight[j]  
	  = sphericalToCartesian(uR, uTheta, uPhi, theta, phi, velocityXYZ);
	j++;
	//E_n^0
	uR = 0.0;
	uTheta = 0.0;
	uPhi = -c5*pd[n];
	velocityWeight[j] 
	  = sphericalToCartesian(uR, uTheta, uPhi, theta, phi, velocityXYZ);
	j++;
      }
      else { 
	//A_n^m and B_n^m
	uR = c1*pl[n+ldp*m]*cos(m*phi);
	uTheta = c3*pd[n+ldp*m]*cos(m*phi);
	uPhi = -m*c3*ps[n+ldp*m]*sin(m*phi); 
	velocityWeight[j] 
	  = sphericalToCartesian(uR, uTheta, uPhi, theta, phi, velocityXYZ);
	j++;
	uR = c1*pl[n+ldp*m]*sin(m*phi);
	uTheta = c3*pd[n+ldp*m]*sin(m*phi);
	uPhi = m*c3*ps[n+ldp*m]*cos(m*phi); 
	velocityWeight[j] 
	  = sphericalToCartesian(uR, uTheta, uPhi, theta, phi, velocityXYZ);
	j++;
	//C_n^m and D_n^m
	uR = c2*pl[n+ldp*m]*cos(m*phi);
	uTheta = c4*pd[n+ldp*m]*cos(m*phi);
	uPhi = -m*c4*ps[n+ldp*m]*sin(m*phi);
	velocityWeight[j] 
	  = sphericalToCartesian(uR, uTheta, uPhi, theta, phi, velocityXYZ);
	j++;
	uR = c2*pl[n+ldp*m]*sin(m*phi);
	uTheta =c4*pd[n+ldp*m]*sin(m*phi);
	uPhi = m*c4*ps[n+ldp*m]*cos(m*phi);
	velocityWeight[j] 
	  = sphericalToCartesian(uR, uTheta, uPhi, theta, phi, velocityXYZ);
	j++;
	//E_n^m and F_n^m
	uR = 0.0;
	uTheta =-m*c5*ps[n+ldp*m]*sin(m*phi);
	uPhi =-c5*pd[n+ldp*m]*cos(m*phi);
	velocityWeight[j] 
	  = sphericalToCartesian(uR, uTheta, uPhi, theta, phi, velocityXYZ);
	j++;
	uR = 0.0;
	uTheta = m*c5*ps[n+ldp*m]*cos(m*phi);
	uPhi = -c5*pd[n+ldp*m]*sin(m*phi);
	velocityWeight[j] 
	  = sphericalToCartesian(uR, uTheta, uPhi, theta, phi, velocityXYZ);
	j++;
      } // for if
    } // for m
  } //for n
  
  
} //for this function


real analyticalSol::sphericalToCartesian(real vectorR,
					 real vectorTheta,
					 real vectorPhi,
					 real theta,
					 real phi,
					 int vectorXYZ){

  // tranformation from r, theta, and phi components of vector
  // to x, y, and z components

  switch ( vectorXYZ ){
  case 1: // for vectorX
    return sin(theta) * cos(phi) * vectorR
      + cos(theta) * cos(phi) * vectorTheta
      - sin(phi) * vectorPhi;
    break;
  case 2: // for vectorY
    return sin(theta) * sin(phi) * vectorR
      + cos(theta) * sin(phi) * vectorTheta
      + cos(phi) * vectorPhi;
    break;
  case 3: // for vectorZ
    return cos(theta) * vectorR - sin(theta) * vectorTheta;
    break;
  }
  return 0.0;
}

real * analyticalSol::allocate1dArray(int size){
  real *x = new real[size];
  return x;
}

void analyticalSol::delete1dArray(real * x){
  delete [] x;
}

void analyticalSol::allocateMemory(){

  weightMatrix = allocate1dArray(nEquations * nCoeff); 
  b = allocate1dArray(nEquations);
  ABCoeff = allocate1dArray(nCoeff);
  pl = allocate1dArray( (Nc+2)*(Nc+2) );
  ps = allocate1dArray( (Nc+2)*(Nc+2) );
  pd = allocate1dArray( (Nc+2)*(Nc+2) );

}

void analyticalSol::releaseMemory(){

  delete1dArray(weightMatrix); 
  delete1dArray(b); 
  delete1dArray(ABCoeff); 
  delete1dArray(pl); 
  delete1dArray(ps); 
  delete1dArray(pd); 

}


analyticalSol::~analyticalSol(){
}
