#ifndef ANALYTICALSOL_H
#define ANALYTICALSOL_H

/** \file analyticalSol.cpp
    \brief Physalis analytical solutions for the Stokes equation

    This class receives the finite-difference solutions of pressure and vorticity, solves the analytical solutions of velocity in the cage points, and returns these analytical solutions as the boundary conditions of the finite-difference scheme of the flow field.
*/

#ifdef SINGLE
#define real float
#else
#define real double
#endif

extern "C" {
#ifdef SINGLE
void salegpsd_(int *nMax,int *mMax,real *cosTheta,real *pl,real *ps,real *pd,int *ldp);
#else
void alegpsd_(int *nMax,int *mMax,real *cosTheta,real *pl,real *ps,real *pd,int *ldp);
#endif
} 



#include <iostream>
#include <string>
#include <cmath>
#include <vector>
#include "LLSSolver.h"
#include "constant.h"

using namespace std;

/**
   \brief Class analytical solutions for the Physalis cage points

   This class receives the finite-difference solutions of pressure and vorticity, solves the analytical solutions of velocity in the cage points, and returns these analytical solutions as the boundary conditions of the finite-difference scheme of the flow field.

*/

class analyticalSol {
  
 public:

  /**
     constructor
   */

  analyticalSol();

  /**
     Receives finite-difference solutions, obtains the analytical velocity for the Stokes equation, and returns these analytical solution to the finite-difference scheme

     construct analytical solution in one cage for pressure, vorticity, velocities at those cage points

    write everything in the form, using pressure as an example,
    p = pWeight dotproduct ABCoeff
    where 
    pWeight = [pWeightA0, pWeightB0, 
           pWeightA1, pWeightATilde1,pWeightB1, pWeightBTilde1,
           pWeightAK, pWeightATildeK,pWeightBK, pWeightBTildeK]
     and
     ABCoeff      = [A0, B0, 
            A1, ATilde1,B1, BTilde1,
            AK, ATildeK,BK, BTildeK]
            where B0 denotes p0 * Re^2

    similarly,
    vorticity = omegaWeight dotproduct ABCoeff, 
    uR = uRWeight dotproduct ABCoeff ,
    uTheta = uThetaWeight dotproduct ABCoeff, 
    u = uWeight dotproduct ABCoeff, 
    v = vWeight dotproduct ABCoeff, 

    ABCoeff is to be determined from weightMatrix ABCoeff = b

    note that r is the normalized distance from the origin 

    Formula are in spherical coordinate:
    p, u, v, w, omegaX, omegaY, and omegaZ.

    all input and output fields of pressure, velocity, and vorticity
    are from the finite-difference scheme
    i.e., in the original inertial frame
   */

  void solve(vector<real> &p,  /**< pressure */
	     const vector<real> &rP,   /**< r value of pressure points */
	     const vector<real> &thetaP, /**< theta value of pressure points */
	     const vector<real> &phiP,   /**< phi value of pressure points */
	     vector<real> &omegaX,       /**< omegaX */
	     const vector<real> &rOmegaX, /**< r value of omegaX points */
	     const vector<real> &thetaOmegaX, /**< theta value of omegaX points */
	     const vector<real> &phiOmegaX, /**< phi value of omegaX points */
	     vector<real> &omegaY, /**< omegaY */
	     const vector<real> &rOmegaY, /**< r value of omegaY points */
	     const vector<real> &thetaOmegaY, /**< theta value of omegaY points */
	     const vector<real> &phiOmegaY, /**< phi value of omegaY points */
	     vector<real> &omegaZ, /**< omegaZ */
	     const vector<real> &rOmegaZ, /**< r value of omegaZ points */
	     const vector<real> &thetaOmegaZ, /**< theta value of omegaZ points */
	     const vector<real> &phiOmegaZ, /**< phi value of omegaZ points */
	     vector<real> &u, /**< u */
	     const vector<real> &rU, /**< r value of u points */
	     const vector<real> &thetaU, /**< theta value of u points */
	     const vector<real> &phiU,/**< phi value of u points */
	     vector<real> &v, /**< v */
	     const vector<real> &rV, /**< r value of v points */
	     const vector<real> &thetaV, /**< theta value of v points */
	     const vector<real> &phiV,/**< phi value of v points */
	     vector<real> &w, /**< w */
	     const vector<real> &rW, /**< r value of w points */
	     const vector<real> &thetaW, /**< theta value of w points */
	     const vector<real> &phiW, /**< phi value of w points */
	     int nTruncate,  /**< Order of the truncated spertral solutions */
	     real ReynoldsNumber,  /**< Reynolds number */
	     vector<real> &solvedABCoeffOneParticle, /**< Coefficiences of A, B, and C */
	     vector<real> &UFixedParticles, /**< stores U of the LLS solver */
	     vector<real> &SFixedParticles, /**< stores S of the LLS solver */
	     vector<real> &VFixedParticles, /**< stores V of the LLS solver */
	     int &fixedParticles, /**< stores if the particles are fixed */
	     int rotatingParticles, /**< stores if the particles are rotating */
	     int particle, /**< m-th particles */
	     real **angVel, /**< angular velocity of particles */
	     real **angVelAcc, /**< angular velocity acceleration of particles */
	     real **torqueOld, /**< torque value in previous iteration of particles */
	     real **torqueNew /**< torque value in current iteration of particles */
	     );

  /**
     destructor
   */

  ~analyticalSol();

 private: 

// modify fields of pressure and vorticity
  void modifiedFields(vector<real> &p,
		      const vector<real> &rP,
		      const vector<real> &thetaP,
		      const vector<real> &phiP,
		      vector<real> &omegaX,
		      const vector<real> &rOmegaX,
		      const vector<real> &thetaOmegaX,
		      const vector<real> &phiOmegaX,
		      vector<real> &omegaY,
		      const vector<real> &rOmegaY,
		      const vector<real> &thetaOmegaY,
		      const vector<real> &phiOmegaY,
		      vector<real> &omegaZ,
		      const vector<real> &rOmegaZ,
		      const vector<real> &thetaOmegaZ,
		      const vector<real> &phiOmegaZ,
		      int particle, // mth particle
		      real **angVel,
		      real **angVelAcc);

// modify fields of velocity due to rotation
  void modifiedVelocity(vector<real> &u,
			const vector<real> &rU,
			const vector<real> &thetaU,
			const vector<real> &phiU,
			vector<real> &v,
			const vector<real> &rV,
			const vector<real> &thetaV,
			const vector<real> &phiV,
			vector<real> &w,
			const vector<real> &rW,
			const vector<real> &thetaW,
			const vector<real> &phiW,
			int particle, // mth particle
			real **angVel,
			real **angVelAcc);

  // solving coefficients of A, B, and their tilde counterparts
  void solveABCoeff(const vector<real> &p,
		    const vector<real> &rP,
		    const vector<real> &thetaP,
		    const vector<real> &phiP,
		    const vector<real> &omegaX,
		    const vector<real> &rOmegaX,
		    const vector<real> &thetaOmegaX,
		    const vector<real> &phiOmegaX,
		    const vector<real> &omegaY,
		    const vector<real> &rOmegaY,
		    const vector<real> &thetaOmegaY,
		    const vector<real> &phiOmegaY,
		    const vector<real> &omegaZ,
		    const vector<real> &rOmegaZ,
		    const vector<real> &thetaOmegaZ,
		    const vector<real> &phiOmegaZ,
		    vector<real> &UFixedParticles,
		    vector<real> &SFixedParticles,
		    vector<real> &VFixedParticles,
		    int &fixedParticles);

  void setB(const vector<real> &p,
	    const vector<real> &omegaX,
	    const vector<real> &omegaY,
	    const vector<real> &omegaZ);

  void setWeightMatrix(const vector<real> &rP,
		       const vector<real> &thetaP,
		       const vector<real> &phiP,
		       const vector<real> &rOmegaX,
		       const vector<real> &thetaOmegaX,
		       const vector<real> &phiOmegaX,
		       const vector<real> &rOmegaY,
		       const vector<real> &thetaOmegaY,
		       const vector<real> &phiOmegaY,
		       const vector<real> &rOmegaZ,
		       const vector<real> &thetaOmegaZ,
		       const vector<real> &phiOmegaZ);

  void setWeightMatrixFromP(const vector<real> &rP,
			    const vector<real> &thetaP,
			    const vector<real> &phiP);

  void setWeightMatrixFromOmega(const vector<real> &rOmega,
				const vector<real> &thetaOmega,
				const vector<real> &phiOmega,
				int nStart,
				int nEnd,
				int vorticityXYZ);


  // update velocity
  void updateVelocity(vector<real> &u,
		      const vector<real> &rU,
		      const vector<real> &thetaU,
		      const vector<real> &phiU,
		      vector<real> &v,
		      const vector<real> &rV,
		      const vector<real> &thetaV,
		      const vector<real> &phiV,
		      vector<real> &w,
		      const vector<real> &rW,
		      const vector<real> &thetaW,
		      const vector<real> &phiW);

  void setVelocityWeight(real rVelocity, 
			 real thetaVelocity,
			 real phiVelocity,
			 int velocityXYZ);
  
  
  real sphericalToCartesian(real vectorR,
			    real vectorTheta,
			    real vectorPhi,
			    real theta,
			    real phi,
			    int vectorXYZ);
  
  real * allocate1dArray(int size);
  void delete1dArray(real * x);

  void allocateMemory();
  void releaseMemory(); 

  int i, j, k; // dummy index

  int n;  // degree
  int m;  // order

  real rToN;
 
  // number of pressure points and vortivity points
  int nP;
  int nOmegaX;
  int nOmegaY;
  int nOmegaZ;
  int nEquations; 

  int Nc;
  int nCoeff;

  real Re;

  int nU;
  int nV;
  int nW;

  real r, theta, phi;

  real cosTheta;
  real omegaR, omegaTheta, omegaPhi;

  real * weightMatrix;
  real * b;
  real * ABCoeff;

  vector<real> velocityWeight;
  real uR, uTheta, uPhi;

  // for associated Legendre polynomials
  real *pl;
  real *ps;
  real *pd;
  int ldp;


};

#endif
