#ifndef CAGE_H
#define CAGE_H

/** \file cage.cpp
    \brief Calculate coordinates of cage points and call the class analyticalSol to calculate the velocity boundary conditions at cage points

    This class first calculate the coordinates of pressure, vorticity, and velocity at cage points. Then use their coordinates and values as the input to the class analyticalSol to calculate the velocity components at cage points. Last, use these valocity as the boundary conditions of the finite-difference scheme for the outer flows. Note that this file is not related with m=0, the base mesh

*/


#ifdef SINGLE
#define real float
#else
#define real double
#endif


#include <iostream>
#include <string>
#include <cmath>
#include <vector>
#include "analyticalSol.h"
using namespace std;

/** 
    \brief Calculate coordinates of cage points and call the class analyticalSol to calculate the velocity boundary conditions at cage points

    This class first calculate the coordinates of pressure, vorticity, and velocity at cage points. Then use their coordinates and values as the input to the class analyticalSol to calculate the velocity components at cage points. Last, use these valocity as the boundary conditions of the finite-difference scheme for the outer flows.

*/

class cage {
  
 public:
  /**
     constructor
   */
  cage();

/** 
    \brief Calculate coordinates of cage points and the velocity at cage points.

    Use the analytical solutions of the Stokes flows to calculate the the velocity at cage points, which acts as the inner boundary conditions for the finite-difference scheme.

*/  
  void solve(int numberOfCage, /**< number of cages */
	     int cellNumberPerPatch, /**<  cell numbers per patch surrounding the sphere*/
	     int dimension, /**<  dimension */
	     int nTruncate, /**<  the order for truncated spectral analytical solution*/
	     real ReynoldsNumber, /**<  Reynolds Number */
	     real *xParticleCenter,  /**< x coordinate of particle center */
	     real *yParticleCenter, /**< y coordinate of particle center  */
	     real *zParticleCenter, /**< z coordinate of particle center  */
	     // for p 
	     real ***xP, /**< coordinates of pressure */
	     real ****p, /**< pressure */
	     int **pointNumberMeshP, /**< mesh points for pressure */
	     int  ****pCagePoint, /**< cage points for pressure */
	     // for u
	     real ***xU, /**< coordinates of u  */
	     real ****u, /**< u */
	     int **pointNumberMeshU, /**< mesh points for u */
	     int  ****uCagePoint, /**< cage points for u */
	     // for v
	     real ***xV, /**<  coordinates of v */
	     real ****v, /**< v */
	     int **pointNumberMeshV, /**< mesh points for v */
	     int  ****vCagePoint, /**< cage points for v */
	     // for w
	     real ***xW, /**<  coordinates of w */
	     real ****w, /**< w */
	     int **pointNumberMeshW, /**< mesh points for w */
	     int  ****wCagePoint, /**< cage points for w */
	     // for omegaX
	     real ***xOmegaX, /**<  coordinates of omega_x */
	     real ****omegaX, /**< omega_x */
	     int **pointNumberMeshOmegaX, /**< mesh points for omega_x */
	     int  ****omegaXCagePoint, /**< cage points for omega_x */
	     // for omegaY 
	     real ***xOmegaY, /**< coordinates of omega_y */
	     real ****omegaY, /**< omega_y */
	     int **pointNumberMeshOmegaY, /**< mesh points for omega_y */
	     int  ****omegaYCagePoint, /**< cage points for omega_y */
	     // for omegaZ
	     real ***xOmegaZ, /**< coordinates of omega_z */
	     real ****omegaZ, /**< omega_z */
	     int **pointNumberMeshOmegaZ, /**< mesh points for omega_z */
	     int  ****omegaZCagePoint, /**< cage points for omega_z */
	     real **solvedABCoeffAllParticleNew,  /**< the new values of A, B, and C */
	     vector<real> &UFixedParticles,  /**< store U for fixed partilce */
	     vector<real> &SFixedParticles, /**< store S for fixed partilce */
	     vector<real> &VFixedParticles, /**< store V for fixed partilce */
	     int &fixedParticles, /**< if the partilce is fixed */
	     int rotatingParticles, /**< if the particle rotating */
	     real **angVel, /**< angular velocity */
	     real **angVelAcc, /**< angular velocity acceleration */
	     real **torqueOld, /**< old torque */
	     real **torqueNew /**< new torque */
	     );

  /**
     destructor
   */
  ~cage();
  
 private: 

  void setIndexVariableCage(int ** pointNumberMeshVariable,
			    int **** indexVariableCage,
			    int **** variableCagePoint);

  void setNVariableCage(int &nVariableCage,
			int ** pointNumberMeshVariable,
			int **** variableCagePoint);

  void setVector(int nVariableCage,
		 int **** indexVariableCage,
		 int * indexIVariableCage,
		 int * indexJVariableCage,
		 int * indexKVariableCage,
		 vector<real> &variableCage,
		 vector<real> &xVariableCage,
		 vector<real> &yVariableCage,   // in the cage class
		 vector<real> &zVariableCage,   // in the cage class
		 vector<real> &rVariableCage,
		 vector<real> &thetaVariableCage,
		 vector<real> &phiVariableCage,
		 //outer class 
		 real **** variable,
		 real *** xVariable,
		 int ** pointNumberMeshVariable,
		 int **** variableCagePoint,
		 real * xParticleCenter,
		 real * yParticleCenter,
		 real * zParticleCenter);

  int * allocate1dArray(int size);
  void delete1dArray(int * x);

  int **** allocate4dIntArrayCage(int nMeshes, int **pointNumberMesh);
  void delete4dArrayCage(int ****x, int nMeshes, int **pointNumberMesh);

  int i, j, k, m; // dummy index
 
  // number of pressure points and vortivity points
  int nCage;
  int nMeshes;
  int nCellPerPatch;
  int dim;

  int **** indexPCage;
  int * indexIPCage;
  int * indexJPCage;
  int * indexKPCage;
  int nPCage;
  vector<real> pCage;
  vector<real> xPCage;
  vector<real> yPCage;
  vector<real> zPCage;
  vector<real> rPCage;
  vector<real> thetaPCage;
  vector<real> phiPCage;

  int **** indexOmegaXCage;
  int * indexIOmegaXCage;
  int * indexJOmegaXCage;
  int * indexKOmegaXCage;
  int nOmegaXCage;
  vector<real> omegaXCage;
  vector<real> xOmegaXCage;
  vector<real> yOmegaXCage;
  vector<real> zOmegaXCage;
  vector<real> rOmegaXCage;
  vector<real> thetaOmegaXCage;
  vector<real> phiOmegaXCage;

  int **** indexOmegaYCage;
  int * indexIOmegaYCage;
  int * indexJOmegaYCage;
  int * indexKOmegaYCage;
  int nOmegaYCage;
  vector<real> omegaYCage;
  vector<real> xOmegaYCage;
  vector<real> yOmegaYCage;
  vector<real> zOmegaYCage;
  vector<real> rOmegaYCage;
  vector<real> thetaOmegaYCage;
  vector<real> phiOmegaYCage;

  int **** indexOmegaZCage;
  int * indexIOmegaZCage;
  int * indexJOmegaZCage;
  int * indexKOmegaZCage;
  int nOmegaZCage;
  vector<real> omegaZCage;
  vector<real> xOmegaZCage;
  vector<real> yOmegaZCage;
  vector<real> zOmegaZCage;
  vector<real> rOmegaZCage;
  vector<real> thetaOmegaZCage;
  vector<real> phiOmegaZCage;

  int **** indexUCage;
  int * indexIUCage;
  int * indexJUCage;
  int * indexKUCage;
  int nUCage;
  vector<real> uCage;
  vector<real> xUCage;
  vector<real> yUCage;
  vector<real> zUCage;
  vector<real> rUCage;
  vector<real> thetaUCage;
  vector<real> phiUCage;

  int **** indexVCage;
  int * indexIVCage;
  int * indexJVCage;
  int * indexKVCage;
  int nVCage;
  vector<real> vCage;
  vector<real> xVCage;
  vector<real> yVCage;
  vector<real> zVCage;
  vector<real> rVCage;
  vector<real> thetaVCage;
  vector<real> phiVCage;

  int **** indexWCage;
  int * indexIWCage;
  int * indexJWCage;
  int * indexKWCage;
  int nWCage;
  vector<real> wCage;
  vector<real> xWCage;
  vector<real> yWCage;
  vector<real> zWCage;
  vector<real> rWCage;
  vector<real> thetaWCage;
  vector<real> phiWCage;

  int Nc;
  real Re;

  vector<real> solvedABCoeffOneParticle;




};

#endif
