#include "flowSolver.h"

// i and j are consistent with the physical domain
// i.e., same i means same x coordinate
//       same j means same y coordinate
// discretization stencil is:
//              (i,j+1)
//     (i-1,j)  (i, j )   (i+1,j)
//              (i,j-1)

//////////////////////////////////////////////
//
// fields of primitive varibles in the original two-dimensional domain.
// thus, use matrix form.
//
//////////////////////////////////////////////


flowSolver::flowSolver(){
}

void flowSolver::solve(){

  //read input file for initialization
  // use the form read(stringOfName, value)
  inputData iObj;
  iObj.read("dim", dim); 
  iObj.read("t", t); 
  iObj.read("dt", dt); 
  iObj.read("nPatch", nPatch); 
  iObj.read("Re", Re); // Reynolds number based on diameter
  iObj.read("Nc", Nc); 
  iObj.read("nCoarseCellCompDomain", nCoarseCellCompDomain); 
  iObj.read("cellNumberPerPatch", cellNumberPerPatch); 
  iObj.read("nCellPerParticle", nCellPerParticle); 
  iObj.read("restart", restart); 
  iObj.read("fixedParticles", fixedParticles); 
  iObj.read("rotatingParticles", rotatingParticles); 

  angVel = allocate2dRealArray(3, nPatch + 1);
  angVelAcc = allocate2dRealArray(3, nPatch + 1);
  torqueOld = allocate2dRealArray(3, nPatch + 1);
  torqueNew = allocate2dRealArray(3, nPatch + 1);
  
  // initialization
  for (m=1; m<=nPatch; m++) {
    for (i=0; i<dim; i++){
      angVel[i][m] = 0.0;
      angVelAcc[i][m] = 0.0;
      torqueOld[i][m] = 0.0;
      torqueOld[i][m] = 0.0;
    }
  }

  //  int factor = (int)pow(2.0, 0.0);

  // ---------------------- PETSc starts ---------------------
  // PETSc solving A x = b
  int zeroInitialGuess = 1;
  PetscLinearSolver pSolver;


  filename = "results/nCoeff.dat";
  
  ofParticle.open( filename.c_str(),(ios_base::out | ios_base::app) ) ;

  ofParticle<< 1 + 3 * Nc * (Nc + 2) <<endl;

  ofParticle.close();


  // set factor such that there are factor^3 particles in the domain
  //  int factor = 1;
  
  // nCoarseCellCompDomain:
  // number of coarse cells
  // the computational domain use (>=12):

  NumberOfMeshes = nPatch + 1;
 
  // number of coarse cells a patch uses
  // cellNumberPerPatch

  //particle
  //number in fine mesh
  nCellPerParticle = 2 * nCellPerParticle; // need less than 2*cellNumberPerPatch


  nCell = allocate2dArray(NumberOfMeshes, dim);
  // x -- 0; y -- 1; z -- 2
  nCell[0][0]= nCoarseCellCompDomain / 2 + nCellPerParticle / 4 + 4; 
  nCell[0][1]= 60; // y direction
  nCell[0][2]= nCoarseCellCompDomain; // z direction

  for (m=1; m<=nPatch; m++) {
    //in both x and y directions:
    // each coarse cell separates to two fine cell
    // including two ghost points
    nCell[m][0] = 2 * (cellNumberPerPatch + 2); // add two more at two ends 
    nCell[m][1] = 2 * (cellNumberPerPatch + 2); // add two more at two ends
    nCell[m][2] = 2 * (cellNumberPerPatch + 2); // add two more at two ends
  }

  // 
  //set up StartIndex for all particles

  //start index based on pressure nodes for each patch 
  StartIndex = allocate2dArray(NumberOfMeshes, dim);
  StartIndex[0][0] = 0; // the computational mesh
  StartIndex[0][1] = 0;
  StartIndex[0][2] = 0;


  // first patch:   x direction
  StartIndex[1][0] = nCell[0][0] - nCoarseCellCompDomain / 2 - nCellPerParticle / 4 - 1 ; 
  //  y direction
  StartIndex[1][1] = (nCell[0][1] - cellNumberPerPatch)/2;
  //  z direction
  StartIndex[1][2] = (nCell[0][2] - cellNumberPerPatch)/2;
    
  /*
  // first patch:   x direction
  StartIndex[1][0] = nCoarseCellCompDomain/2 - cellNumberPerPatch/2; 
  //  y direction
  StartIndex[1][1] = nCoarseCellCompDomain/2 - cellNumberPerPatch/2;
  //  z direction
  StartIndex[1][2] = nCoarseCellCompDomain/2 - cellNumberPerPatch/2;
  */


  /*  
  for (k=0; k<factor; k++) {
    for (j=0; j<factor; j++) {
      for (i=0; i<factor; i++) {

	// the index for the first patch is 1
	int indexPatch = i * factor * factor + j * factor + k + 1;

	// first patch:   x direction
	StartIndex[indexPatch][0] = StartIndex[1][0] + i * nCoarseCellCompDomain;
	//  y firection
	StartIndex[indexPatch][1] = StartIndex[1][1] + j * nCoarseCellCompDomain;
	//  z firection
	StartIndex[indexPatch][2] = StartIndex[1][2] + k * nCoarseCellCompDomain;

      }
    }
  }	
 
  */
 
  //
  // set the mesh size of u, v, and p  
  //

  pointNumberMeshU = allocate2dArray(NumberOfMeshes, dim);
  pointNumberMeshU[0][0] = nCell[0][0] + 1; // for x direction
  pointNumberMeshU[0][1] = nCell[0][1] + 2; // for y direction
  pointNumberMeshU[0][2] = nCell[0][2] + 2; // for z direction
  for (m=1; m<=nPatch; m++) {
    //in both x and y directions:
    // each coarse cell separates to two fine cell
    // including two ghost points
    pointNumberMeshU[m][0] = nCell[m][0] + 1; 
    pointNumberMeshU[m][1] = nCell[m][1]; 
    pointNumberMeshU[m][2] = nCell[m][2]; 
  }

  pointNumberMeshV = allocate2dArray(NumberOfMeshes, dim);
  pointNumberMeshV[0][0] = nCell[0][0] + 2; // for x direction
  pointNumberMeshV[0][1] = nCell[0][1] + 1; // for y direction
  pointNumberMeshV[0][2] = nCell[0][2] + 2; // for z direction
  for (m=1; m<=nPatch; m++) {
    //in both x and y directions:
    // each coarse cell separates to two fine cell
    // including two ghost points
    pointNumberMeshV[m][0] = nCell[m][0]; 
    pointNumberMeshV[m][1] = nCell[m][1] + 1; 
    pointNumberMeshV[m][2] = nCell[m][2]; 
  }

  pointNumberMeshW = allocate2dArray(NumberOfMeshes, dim);
  pointNumberMeshW[0][0] = nCell[0][0] + 2; // for x direction
  pointNumberMeshW[0][1] = nCell[0][1] + 2; // for y direction
  pointNumberMeshW[0][2] = nCell[0][2] + 1; // for z direction
  for (m=1; m<=nPatch; m++) {
    //in both x and y directions:
    // each coarse cell separates to two fine cell
    // including two ghost points
    pointNumberMeshW[m][0] = nCell[m][0]; 
    pointNumberMeshW[m][1] = nCell[m][1]; 
    pointNumberMeshW[m][2] = nCell[m][2] + 1; 
  }

  pointNumberMeshP = allocate2dArray(NumberOfMeshes, dim);
  for (m=0; m<=nPatch; m++) {
    //in both x and y directions:
    // each coarse cell separates to two fine cell
    // including two ghost points
    pointNumberMeshP[m][0] = nCell[m][0];  // for x direction
    pointNumberMeshP[m][1] = nCell[m][1];  // for y direction
    pointNumberMeshP[m][2] = nCell[m][2];  // for z direction
  }

  pointNumberMeshOmegaX = allocate2dArray(NumberOfMeshes, dim);
  pointNumberMeshOmegaX[0][0] = nCell[0][0] + 2; // for x direction
  pointNumberMeshOmegaX[0][1] = nCell[0][1] + 1; // for y direction
  pointNumberMeshOmegaX[0][2] = nCell[0][1] + 1; // for y direction
  for (m=1; m<=nPatch; m++) {
    //in both x and y directions:
    // each coarse cell separates to two fine cell
    // including two ghost points
    pointNumberMeshOmegaX[m][0] = nCell[m][0]; 
    pointNumberMeshOmegaX[m][1] = nCell[m][1] - 1; 
    pointNumberMeshOmegaX[m][2] = nCell[m][2] - 1; 
  }

  pointNumberMeshOmegaY = allocate2dArray(NumberOfMeshes, dim);
  pointNumberMeshOmegaY[0][0] = nCell[0][0] + 1; // for x direction
  pointNumberMeshOmegaY[0][1] = nCell[0][1] + 2; // for y direction
  pointNumberMeshOmegaY[0][2] = nCell[0][1] + 1; // for y direction
  for (m=1; m<=nPatch; m++) {
    //in both x and y directions:
    // each coarse cell separates to two fine cell
    // including two ghost points
    pointNumberMeshOmegaY[m][0] = nCell[m][0] - 1; 
    pointNumberMeshOmegaY[m][1] = nCell[m][1]; 
    pointNumberMeshOmegaY[m][2] = nCell[m][2] - 1; 
  }

  pointNumberMeshOmegaZ = allocate2dArray(NumberOfMeshes, dim);
  pointNumberMeshOmegaZ[0][0] = nCell[0][0] + 1; // for x direction
  pointNumberMeshOmegaZ[0][1] = nCell[0][1] + 1; // for y direction
  pointNumberMeshOmegaZ[0][2] = nCell[0][1] + 2; // for y direction
  for (m=1; m<=nPatch; m++) {
    //in both x and y directions:
    // each coarse cell separates to two fine cell
    // including two ghost points
    pointNumberMeshOmegaZ[m][0] = nCell[m][0] - 1; 
    pointNumberMeshOmegaZ[m][1] = nCell[m][1] - 1; 
    pointNumberMeshOmegaZ[m][2] = nCell[m][2]; 
  }
  
  h = allocate1dRealArray(NumberOfMeshes);
  hFine= 2.0/nCellPerParticle;  // the nondimensional diameter is 2.0
  hCoarse= 2.0 * hFine;  // for the coarse mesh
  h[0] = hCoarse;
  for (m=1; m<=nPatch; m++) {
    // each coarse cell separates to two fine cell
    h[m] = hFine;   // for fine mesh
  }

  // all in form of variable[m][i][j]
  // m: index of meshes of computational domain and patchs
  // i for x direction
  //j for y direction
  xP = allocate3dRealArray(NumberOfMeshes, dim, pointNumberMeshP);
  fp = allocate4dRealArray(NumberOfMeshes, pointNumberMeshP);
  exactP = allocate4dRealArray(NumberOfMeshes, pointNumberMeshP);
  p = allocate4dRealArray(NumberOfMeshes, pointNumberMeshP);
  pUnknowns = allocate4dIntArray(NumberOfMeshes, pointNumberMeshP);

  // velocity u
  xU = allocate3dRealArray(NumberOfMeshes, dim, pointNumberMeshU);
  exactU = allocate4dRealArray(NumberOfMeshes, pointNumberMeshU);
  u = allocate4dRealArray(NumberOfMeshes, pointNumberMeshU);
  uAux = allocate4dRealArray(NumberOfMeshes, pointNumberMeshU);
  uUnknowns = allocate4dIntArray(NumberOfMeshes, pointNumberMeshU);

  // velocity v
  xV = allocate3dRealArray(NumberOfMeshes, dim, pointNumberMeshV);
  exactV = allocate4dRealArray(NumberOfMeshes, pointNumberMeshV);
  v = allocate4dRealArray(NumberOfMeshes, pointNumberMeshV);
  vAux = allocate4dRealArray(NumberOfMeshes, pointNumberMeshV);
  vUnknowns = allocate4dIntArray(NumberOfMeshes, pointNumberMeshV);

  // velocity w
  xW = allocate3dRealArray(NumberOfMeshes, dim, pointNumberMeshW);
  exactW = allocate4dRealArray(NumberOfMeshes, pointNumberMeshW);
  w = allocate4dRealArray(NumberOfMeshes, pointNumberMeshW);
  wAux = allocate4dRealArray(NumberOfMeshes, pointNumberMeshW);
  wUnknowns = allocate4dIntArray(NumberOfMeshes, pointNumberMeshW);

  // vorticity omegaX
  xOmegaX = allocate3dRealArray(NumberOfMeshes, dim, pointNumberMeshOmegaX);
  omegaX = allocate4dRealArray(NumberOfMeshes, pointNumberMeshOmegaX);

  // vorticity omegaY
  xOmegaY = allocate3dRealArray(NumberOfMeshes, dim, pointNumberMeshOmegaY);
  omegaY = allocate4dRealArray(NumberOfMeshes, pointNumberMeshOmegaY);

  // vorticity omegaZ
  xOmegaZ = allocate3dRealArray(NumberOfMeshes, dim, pointNumberMeshOmegaZ);
  omegaZ = allocate4dRealArray(NumberOfMeshes, pointNumberMeshOmegaZ);
 
  //wavenumebr in x and y directions
  nx = 1;
  ny = 1;
  nz = 1;

  setUnknowns();

  setCoord();

  cout << "x range of the sphere:"<<endl;
  cout << xU[1][0][cellNumberPerPatch - nCellPerParticle/2 + 2] <<endl;


  //// particles
  xParticleCenter = allocate1dRealArray(NumberOfMeshes);
  yParticleCenter = allocate1dRealArray(NumberOfMeshes);
  zParticleCenter = allocate1dRealArray(NumberOfMeshes);

  for (m=0; m<=nPatch; m++) {
    i = pointNumberMeshU[m][0] - 1;
    xParticleCenter[m] = ( xU[m][0][0] + xU[m][0][i] ) / 2.0 ; 
 
    j = pointNumberMeshV[m][1] - 1;
    yParticleCenter[m] = ( xV[m][1][0] + xV[m][1][j] ) / 2.0 ;  

    k = pointNumberMeshW[m][2] - 1;
    zParticleCenter[m] = ( xW[m][2][0] + xW[m][2][k] ) / 2.0 ;  

  }


  pInsideParticle = allocate4dIntArrayCage(NumberOfMeshes, pointNumberMeshP);
  for (m=1; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshP[m][2]; k++) {
      for (j=0; j<pointNumberMeshP[m][1]; j++) {
	for (i=0; i<pointNumberMeshP[m][0]; i++) {
	  pInsideParticle[m][i][j][k] = 0; // initialize
	  pUnknowns[m][i][j][k] = 1;
	  if ( pow( xP[m][0][i] - xParticleCenter[m], 2.0 )
	       + pow( xP[m][1][j] - yParticleCenter[m], 2.0 ) 
	       + pow( xP[m][2][k] - zParticleCenter[m], 2.0 ) 
	       < pow( nCellPerParticle * h[m] / 2.0, 2.0) ){
	    pInsideParticle[m][i][j][k] = 1; // test if inside the particle
	    pUnknowns[m][i][j][k] = 0; //don't include inside points
	  }
	}
      }
    }  
  }

  pCagePoint = allocate4dIntArrayCage(NumberOfMeshes, pointNumberMeshP);
  uCagePoint = allocate4dIntArrayCage(NumberOfMeshes, pointNumberMeshU);
  vCagePoint = allocate4dIntArrayCage(NumberOfMeshes, pointNumberMeshV);
  wCagePoint = allocate4dIntArrayCage(NumberOfMeshes, pointNumberMeshW);
  omegaXCagePoint = allocate4dIntArrayCage(NumberOfMeshes, pointNumberMeshOmegaX);
  omegaYCagePoint = allocate4dIntArrayCage(NumberOfMeshes, pointNumberMeshOmegaY);
  omegaZCagePoint = allocate4dIntArrayCage(NumberOfMeshes, pointNumberMeshOmegaZ);



  for (m=1; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshP[m][2]; k++) { // interior points
      for (j=0; j<pointNumberMeshP[m][1]; j++) { // interior points
	for (i=0; i<pointNumberMeshP[m][0]; i++) {
	  pCagePoint[m][i][j][k] = 0; // initialize
	}
      }
    }
  }

  for (m=1; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshU[m][2]; k++) { // interior points
      for (j=0; j<pointNumberMeshU[m][1]; j++) { // interior points
	for (i=0; i<pointNumberMeshU[m][0]; i++) {
	  uCagePoint[m][i][j][k] = 0; // initialize
	}
      }
    }
  }

  for (m=1; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshV[m][2]; k++) { // interior points
      for (j=0; j<pointNumberMeshV[m][1]; j++) { // interior points
	for (i=0; i<pointNumberMeshV[m][0]; i++) {
	  vCagePoint[m][i][j][k] = 0; // initialize
	}
      }
    }
  }


  for (m=1; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshW[m][2]; k++) { // interior points
      for (j=0; j<pointNumberMeshW[m][1]; j++) { // interior points
	for (i=0; i<pointNumberMeshW[m][0]; i++) {
	  wCagePoint[m][i][j][k] = 0; // initialize
	}
      }
    }
  }

  for (m=1; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshOmegaX[m][2]; k++) { // interior points
      for (j=0; j<pointNumberMeshOmegaX[m][1]; j++) { // interior points
	for (i=0; i<pointNumberMeshOmegaX[m][0]; i++) {
	  omegaXCagePoint[m][i][j][k] = 0; // initialize
	}
      }
    }
  }

  for (m=1; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshOmegaY[m][2]; k++) { // interior points
      for (j=0; j<pointNumberMeshOmegaY[m][1]; j++) { // interior points
	for (i=0; i<pointNumberMeshOmegaY[m][0]; i++) {
	  omegaYCagePoint[m][i][j][k] = 0; // initialize
	}
      }
    }
  }

  for (m=1; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshOmegaZ[m][2]; k++) { // interior points
      for (j=0; j<pointNumberMeshOmegaZ[m][1]; j++) { // interior points
	for (i=0; i<pointNumberMeshOmegaZ[m][0]; i++) {
	  omegaZCagePoint[m][i][j][k] = 0; // initialize
	}
      }
    }
  }
  
  for (m=1; m<=nPatch; m++) {
    for (k=1; k<pointNumberMeshP[m][2]-1; k++) { // interior points
      for (j=1; j<pointNumberMeshP[m][1]-1; j++) { // interior points
	for (i=1; i<pointNumberMeshP[m][0]-1; i++) {
	  if ( pInsideParticle[m][i][j][k] == 0  // consider points outside particles
	       && ( ( pow( xP[m][0][i+1] - xParticleCenter[m], 2.0 )
		      + pow( xP[m][1][j] - yParticleCenter[m], 2.0 ) 
		      + pow( xP[m][2][k] - zParticleCenter[m], 2.0 ) 
		      < pow( nCellPerParticle * h[m] / 2.0, 2.0) ) 
		    || ( pow( xP[m][0][i-1] - xParticleCenter[m], 2.0 )
			 + pow( xP[m][1][j] - yParticleCenter[m], 2.0 ) 
			 + pow( xP[m][2][k] - zParticleCenter[m], 2.0 ) 
			 < pow( nCellPerParticle * h[m] / 2.0, 2.0) ) 
		    || ( pow( xP[m][0][i] - xParticleCenter[m], 2.0 )
			 + pow( xP[m][1][j+1] - yParticleCenter[m], 2.0 ) 
			 + pow( xP[m][2][k] - zParticleCenter[m], 2.0 ) 
			 < pow( nCellPerParticle * h[m] / 2.0, 2.0) ) 
		    || ( pow( xP[m][0][i] - xParticleCenter[m], 2.0 )
			 + pow( xP[m][1][j-1] - yParticleCenter[m], 2.0 ) 
			 + pow( xP[m][2][k] - zParticleCenter[m], 2.0 ) 
			 < pow( nCellPerParticle * h[m] / 2.0, 2.0) ) 
		    || ( pow( xP[m][0][i] - xParticleCenter[m], 2.0 )
			 + pow( xP[m][1][j] - yParticleCenter[m], 2.0 ) 
			 + pow( xP[m][2][k+1] - zParticleCenter[m], 2.0 ) 
			 < pow( nCellPerParticle * h[m] / 2.0, 2.0) ) 
		    || ( pow( xP[m][0][i] - xParticleCenter[m], 2.0 )
			 + pow( xP[m][1][j] - yParticleCenter[m], 2.0 ) 
			 + pow( xP[m][2][k-1] - zParticleCenter[m], 2.0 ) 
			 < pow( nCellPerParticle * h[m] / 2.0, 2.0) ) 
		    ) 
	       ) {
	    pCagePoint[m][i][j][k] = 1; // test if inside the particle
	  }
	  //include outside points and also part of inside points
	  //	pUnknowns[m][i][j] += pCagePoint[m][i][j]; 
	}
      }
    }
  } 


  for (m=1; m<=nPatch; m++) {
    for (k=1; k<pointNumberMeshP[m][2]-1; k++) { // interior points
      for (j=1; j<pointNumberMeshP[m][1]-1; j++) { // interior points
	for (i=1; i<pointNumberMeshP[m][0]-1; i++) {
	  
	  if ( pCagePoint[m][i][j][k] == 1 ) {
	    
	    //test east face of a cell
	    if ( pCagePoint[m][i-1][j][k] == 0   //not a pressure point
		 && fabs( xU[m][0][i-1] - xParticleCenter[m] ) 
		 < fabs( xU[m][0][i] - xParticleCenter[m] )  // close to center
		 ) {
	      
	      uUnknowns[m][i][j][k] = 1;
	      uCagePoint[m][i][j][k] = 1;
	      
	      if ( pCagePoint[m][i][j-1][k] == 1
		   || pCagePoint[m][i-1][j-1][k] == 1 ) {
		vUnknowns[m][i-1][j][k] = 1;
		vCagePoint[m][i-1][j][k] = 1;
	      }
	      
	      if ( pCagePoint[m][i][j+1][k] == 1
		   || pCagePoint[m][i-1][j+1][k] == 1 ) {
		vUnknowns[m][i-1][j+1][k] = 1;
		vCagePoint[m][i-1][j+1][k] = 1;
	      }

	      if ( pCagePoint[m][i][j][k-1] == 1
		   || pCagePoint[m][i-1][j][k-1] == 1 ) {
		wUnknowns[m][i-1][j][k] = 1;
		wCagePoint[m][i-1][j][k] = 1;
	      }
	      
	      if ( pCagePoint[m][i][j][k+1] == 1
		   || pCagePoint[m][i-1][j][k+1] == 1 ) {
		wUnknowns[m][i-1][j][k+1] = 1;
		wCagePoint[m][i-1][j][k+1] = 1;
	      }

	            
	      if (  pow( xOmegaY[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i-1][j][k-1] = 1;
	      }
	      
	      if (  pow( xOmegaY[m][0][i-1] - xParticleCenter[m], 2.0 )
 		  + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
 		  + pow( xOmegaY[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i-1][j][k] = 1;
	      }
	      
	      if (  pow( xOmegaZ[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i-1][j-1][k] = 1;
	      }
	      
	      if (  pow( xOmegaZ[m][0][i-1] - xParticleCenter[m], 2.0 )
 		  + pow( xOmegaZ[m][1][j] - yParticleCenter[m], 2.0 ) 
 		  + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i-1][j][k] = 1;
	      }
	      
	      
	      /*
	      if (  pow( xOmegaY[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i][j][k-1] = 1;
	      }
	      
	      if (  pow( xOmegaY[m][0][i] - xParticleCenter[m], 2.0 )
 		  + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
 		  + pow( xOmegaY[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i][j][k] = 1;
	      }
	      
	      if (  pow( xOmegaZ[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i][j-1][k] = 1;
	      }
	      
	      if (  pow( xOmegaZ[m][0][i] - xParticleCenter[m], 2.0 )
 		  + pow( xOmegaZ[m][1][j] - yParticleCenter[m], 2.0 ) 
 		  + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i][j][k] = 1;
	      }
	      */	      

	    }
	    
	    //test west face of a cell
	    if ( pCagePoint[m][i+1][j][k] == 0   //not a pressure point
		 && fabs( xU[m][0][i+1] - xParticleCenter[m] ) 
		 < fabs( xU[m][0][i] - xParticleCenter[m] ) // close to center
		 ) {

	      uUnknowns[m][i+1][j][k] = 1;
	      uCagePoint[m][i+1][j][k] = 1;
	      
	      if ( pCagePoint[m][i][j-1][k] == 1
		   || pCagePoint[m][i+1][j-1][k] == 1 ) {
		vUnknowns[m][i+1][j][k] = 1;
		vCagePoint[m][i+1][j][k] = 1;
	      }
	      
	      if ( pCagePoint[m][i][j+1][k] == 1
		   || pCagePoint[m][i+1][j+1][k] == 1 ) {
		vUnknowns[m][i+1][j+1][k] = 1;
		vCagePoint[m][i+1][j+1][k] = 1;
	      }

	      if ( pCagePoint[m][i][j][k-1] == 1
		   || pCagePoint[m][i+1][j][k-1] == 1 ) {
		wUnknowns[m][i+1][j][k] = 1;
		wCagePoint[m][i+1][j][k] = 1;
	      }
	      
	      if ( pCagePoint[m][i][j][k+1] == 1
		   || pCagePoint[m][i+1][j][k+1] == 1 ) {
		wUnknowns[m][i+1][j][k+1] = 1;
		wCagePoint[m][i+1][j][k+1] = 1;
	      }

	      	      
	      if (  pow( xOmegaY[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i][j][k-1] = 1;
	      }
	      
	      if (  pow( xOmegaY[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i][j][k] = 1;
	      }

	      if (  pow( xOmegaZ[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i][j-1][k] = 1;
	      }
	      
	      if (  pow( xOmegaZ[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i][j][k] = 1;
	      }
	      

	      /*
	      if (  pow( xOmegaY[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i-1][j][k-1] = 1;
	      }
	      
	      if (  pow( xOmegaY[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i-1][j][k] = 1;
	      }

	      if (  pow( xOmegaZ[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i-1][j-1][k] = 1;
	      }
	      
	      if (  pow( xOmegaZ[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i-1][j][k] = 1;
	      }
	      */
	      
	    }
	    

	    //test south face of a cell
	    if ( pCagePoint[m][i][j-1][k] == 0   //not a pressure point
		 && fabs( xU[m][1][j-1] - yParticleCenter[m] ) 
		 < fabs( xU[m][1][j] - yParticleCenter[m] ) ) {
	      
	      vUnknowns[m][i][j][k] = 1;
	      vCagePoint[m][i][j][k] = 1;
	      
	      if ( pCagePoint[m][i-1][j][k] == 1
		   || pCagePoint[m][i-1][j-1][k] == 1 ) {
		uUnknowns[m][i][j-1][k] = 1;
		uCagePoint[m][i][j-1][k] = 1;
	      }
	      
	      if ( pCagePoint[m][i+1][j][k] == 1
		   || pCagePoint[m][i+1][j-1][k] == 1 ) {
		uUnknowns[m][i+1][j-1][k] = 1;
		uCagePoint[m][i+1][j-1][k] = 1;
	      }

	      if ( pCagePoint[m][i][j][k-1] == 1
		   || pCagePoint[m][i][j-1][k-1] == 1 ) {
		wUnknowns[m][i][j-1][k] = 1;
		wCagePoint[m][i][j-1][k] = 1;
	      }
	      
	      if ( pCagePoint[m][i][j][k+1] == 1
		   || pCagePoint[m][i][j-1][k+1] == 1 ) {
		wUnknowns[m][i][j-1][k+1] = 1;
		wCagePoint[m][i][j-1][k+1] = 1;
	      }

	      	            
	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j-1][k-1] = 1;
	      }
	      
	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j-1][k] = 1;
	      }

	      if (  pow( xOmegaZ[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i-1][j-1][k] = 1;
	      }
	      
	      if (  pow( xOmegaZ[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i][j-1][k] = 1;
	      }
	      

	      /*
	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j][k-1] = 1;
	      }
	      
	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j][k] = 1;
	      }

	      if (  pow( xOmegaZ[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i-1][j][k] = 1;
	      }
	      
	      if (  pow( xOmegaZ[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i][j][k] = 1;
	      }
	      */
	      
	    }
	    

	    //test north face of a cell
	    if ( pCagePoint[m][i][j+1][k] == 0   //not a pressure point
		 && fabs( xU[m][1][j+1] - yParticleCenter[m] ) 
		 < fabs( xU[m][1][j] - yParticleCenter[m] ) ) {
	      
	      vUnknowns[m][i][j+1][k] = 1;
	      vCagePoint[m][i][j+1][k] = 1;
	      
	      if ( pCagePoint[m][i-1][j][k] == 1
		   || pCagePoint[m][i-1][j+1][k] == 1 ) {
		uUnknowns[m][i][j+1][k] = 1;
		uCagePoint[m][i][j+1][k] = 1;
	      }
	      
	      if ( pCagePoint[m][i+1][j][k] == 1
		   || pCagePoint[m][i+1][j+1][k] == 1 ) {
		uUnknowns[m][i+1][j+1][k] = 1;
		uCagePoint[m][i+1][j+1][k] = 1;
	      }

	      if ( pCagePoint[m][i][j][k-1] == 1
		   || pCagePoint[m][i][j+1][k-1] == 1 ) {
		wUnknowns[m][i][j+1][k] = 1;
		wCagePoint[m][i][j+1][k] = 1;
	      }
	      
	      if ( pCagePoint[m][i][j][k+1] == 1
		   || pCagePoint[m][i][j+1][k+1] == 1 ) {
		wUnknowns[m][i][j+1][k+1] = 1;
		wCagePoint[m][i][j+1][k+1] = 1;
	      }

	      
	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j][k-1] = 1;
	      }
	      
	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j][k] = 1;
	      }

	      if (  pow( xOmegaZ[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i-1][j][k] = 1;
	      }
	      
	      if (  pow( xOmegaZ[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i][j][k] = 1;
	      }
	      

	      /*
	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j-1][k-1] = 1;
	      }
	      
	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j-1][k] = 1;
	      }

	      if (  pow( xOmegaZ[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i-1][j-1][k] = 1;
	      }
	      
	      if (  pow( xOmegaZ[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaZ[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaZCagePoint[m][i][j-1][k] = 1;
	      }
	      */	      
	      
	    }
	    
    
	    //test bottom face of a cell
	    if ( pCagePoint[m][i][j][k-1] == 0   //not a pressure point
		 && fabs( xU[m][2][k-1] - zParticleCenter[m] ) 
		 < fabs( xU[m][2][k] - zParticleCenter[m] ) ) {
	      
	      wUnknowns[m][i][j][k] = 1;
	      wCagePoint[m][i][j][k] = 1;
	      
	      if ( pCagePoint[m][i][j-1][k] == 1
		   || pCagePoint[m][i][j-1][k-1] == 1 ) {
		vUnknowns[m][i][j][k-1] = 1;
		vCagePoint[m][i][j][k-1] = 1;
	      }
	      
	      if ( pCagePoint[m][i][j+1][k] == 1
		   || pCagePoint[m][i][j+1][k-1] == 1 ) {
		vUnknowns[m][i][j+1][k-1] = 1;
		vCagePoint[m][i][j+1][k-1] = 1;
	      }

	      if ( pCagePoint[m][i-1][j][k] == 1
		   || pCagePoint[m][i-1][j][k-1] == 1 ) {
		uUnknowns[m][i][j][k-1] = 1;
		uCagePoint[m][i][j][k-1] = 1;
	      }
	      
	      if ( pCagePoint[m][i+1][j][k] == 1
		   || pCagePoint[m][i+1][j][k-1] == 1 ) {
		uUnknowns[m][i+1][j][k-1] = 1;
		uCagePoint[m][i+1][j][k-1] = 1;
	      }

	      	            
	      if (  pow( xOmegaY[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i-1][j][k-1] = 1;
	      }

      
	      if (  pow( xOmegaY[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i][j][k-1] = 1;
	      }

	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j-1][k-1] = 1;
	      }
	      
	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j][k-1] = 1;
	      }
	      
	      
	      /*
	      if (  pow( xOmegaY[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i-1][j][k] = 1;
	      }

      
	      if (  pow( xOmegaY[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i][j][k] = 1;
	      }

	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j-1][k] = 1;
	      }
	      
	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j][k] = 1;
	      }
	      */	      

	    }
	    

	    //test top face of a cell
	    if ( pCagePoint[m][i][j][k+1] == 0   //not a pressure point
		 && fabs( xU[m][2][k+1] - zParticleCenter[m] ) 
		 < fabs( xU[m][2][k] - zParticleCenter[m] ) ) {
	      
	      wUnknowns[m][i][j][k+1] = 1;
	      wCagePoint[m][i][j][k+1] = 1;
	      
	      if ( pCagePoint[m][i][j-1][k] == 1
		   || pCagePoint[m][i][j-1][k+1] == 1 ) {
		vUnknowns[m][i][j][k+1] = 1;
		vCagePoint[m][i][j][k+1] = 1;
	      }
	      
	      if ( pCagePoint[m][i][j+1][k] == 1
		   || pCagePoint[m][i][j+1][k+1] == 1 ) {
		vUnknowns[m][i][j+1][k+1] = 1;
		vCagePoint[m][i][j+1][k+1] = 1;
	      }

	      if ( pCagePoint[m][i-1][j][k] == 1
		   || pCagePoint[m][i-1][j][k+1] == 1 ) {
		uUnknowns[m][i][j][k+1] = 1;
		uCagePoint[m][i][j][k+1] = 1;
	      }
	      
	      if ( pCagePoint[m][i+1][j][k] == 1
		   || pCagePoint[m][i+1][j][k+1] == 1 ) {
		uUnknowns[m][i+1][j][k+1] = 1;
		uCagePoint[m][i+1][j][k+1] = 1;
	      }

	      
	      if (  pow( xOmegaY[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i-1][j][k] = 1;
	      }
	      
	      if (  pow( xOmegaY[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i][j][k] = 1;
	      }

	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j-1][k] = 1;
	      }
	      
	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j][k] = 1;
	      }
	      
	      
	      /*
	      if (  pow( xOmegaY[m][0][i-1] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i-1][j][k-1] = 1;
	      }
	      
	      if (  pow( xOmegaY[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaY[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaYCagePoint[m][i][j][k-1] = 1;
	      }

	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j-1] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j-1][k-1] = 1;
	      }
	      
	      if (  pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		    + pow( xOmegaX[m][1][j] - yParticleCenter[m], 2.0 ) 
		    + pow( xOmegaX[m][2][k-1] - zParticleCenter[m], 2.0 ) 
		    > pow( nCellPerParticle * h[m] / 2.0, 2.0) - h[m] / 100.0 ){
		omegaXCagePoint[m][i][j][k-1] = 1;
	      }
	      */
	      
	    }
	    
	    
	    
	  } // first if
	  
	} // i
      }  // j
    }  // k

  } // m


  for (m=1; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshP[m][2]; k++) { // interior points
      for (j=0; j<pointNumberMeshP[m][1]; j++) { // interior points
	for (i=0; i<pointNumberMeshP[m][0]; i++) {
	  
	  //include those points surrending pressure points
	  if ( pUnknowns[m][i][j][k] == 1 ) {
	    uUnknowns[m][i][j][k] = 1;
	    uUnknowns[m][i+1][j][k] = 1;
	    vUnknowns[m][i][j][k] = 1;
	    vUnknowns[m][i][j+1][k] = 1;
	    wUnknowns[m][i][j][k] = 1;
	    wUnknowns[m][i][j][k+1] = 1;
	  }
	  
	}
      }
    }  
  }

  /*
  
  for (m=1; m<NumberOfMeshes; m++) {
    cout << "--------omegaCage-------" <<endl;
    for (j=0; j<pointNumberMeshOmega[m][1]; j++) {
      for (i=0; i<pointNumberMeshOmega[m][0]; i++) {
 	cout << omegaCagePoint[m][i][j] << "  ";
      }
      cout <<endl;
    }
    cout << "-----------------" <<endl;
  }
  
    
  
  for (m=1; m<NumberOfMeshes; m++) {
    cout << "--------ucage-------" <<endl;
    for (j=0; j<pointNumberMeshU[m][1]; j++) {
      for (i=0; i<pointNumberMeshU[m][0]; i++) {
	cout << uCagePoint[m][i][j] << "  ";
      }
      cout <<endl;
    }
    cout << "-----------------" <<endl;
  }

  

  for (m=1; m<NumberOfMeshes; m++) {
    cout << "--------vcage-------" <<endl;
    for (j=0; j<pointNumberMeshV[m][1]; j++) {
      for (i=0; i<pointNumberMeshV[m][0]; i++) {
	cout << vCagePoint[m][i][j] << "  ";
      }
      cout <<endl;
    }
    cout << "-----------------" <<endl;
  }

  for (m=1; m<NumberOfMeshes; m++) {
    cout << "--------pcage-------" <<endl;
    for (j=0; j<pointNumberMeshP[m][1]; j++) {
      for (i=0; i<pointNumberMeshP[m][0]; i++) {
	cout << pCagePoint[m][i][j] << "  ";
      }
      cout <<endl;
    }
    cout << "-----------------" <<endl;
  }

  

  for (m=1; m<NumberOfMeshes; m++) {
    cout << "--------uUnknowns-------" <<endl;
    for (j=0; j<pointNumberMeshU[m][1]; j++) {
      for (i=0; i<pointNumberMeshU[m][0]; i++) {
	cout << uUnknowns[m][i][j] << "  ";
      }
      cout <<endl;
    }
    cout << "-----------------" <<endl;
  }

  for (m=1; m<NumberOfMeshes; m++) {
    cout << "--------vUnknowns-------" <<endl;
    for (j=0; j<pointNumberMeshV[m][1]; j++) {
      for (i=0; i<pointNumberMeshV[m][0]; i++) {
	cout << vUnknowns[m][i][j] << "  ";
      }
      cout <<endl;
    }
    cout << "-----------------" <<endl;
  }

  for (m=1; m<NumberOfMeshes; m++) {
    cout << "--------pUnknowns-------" <<endl;
    for (j=0; j<pointNumberMeshP[m][1]; j++) {
      for (i=0; i<pointNumberMeshP[m][0]; i++) {
	cout << pUnknowns[m][i][j] << "  ";
      }
      cout <<endl;
    }
    cout << "-----------------" <<endl;
  }
*/
  //// particles

  ifstream  fin ;
  int tStepInit;

  if (restart == 0 ) {

  tStepInit = 1; // time step

  filename = "results/tStepInitOld.dat"; 
  ofParticle.open( filename.c_str() ) ;
  ofParticle<< tStepInit <<endl;
  ofParticle.close();

  
  // set up initial conditions for all field variables
  // u, v, pressure, and vorticity
  initCond initObj;
  initObj.initialize(NumberOfMeshes,
		     pointNumberMeshU, 
		     xU, 
		     pointNumberMeshV, 
		     xV, 
		     pointNumberMeshW, 
		     xW, 
		     pointNumberMeshP, 
		     xP, 
		     u, 
		     v,
		     w,
		     p);
  } else {

    fin.open("results/tStepInitOld.dat") ;   
    fin >> tStepInit;
    fin.close() ;

    fin.open("results/restartFieldOld.dat") ; 
    /*  
    for (m=0; m<NumberOfMeshes; m++) {
      for (k=0; k<pointNumberMeshU[m][2]; k++) {
	for (j=0; j<pointNumberMeshU[m][1]; j++) {
	  for (i=0; i<pointNumberMeshU[m][0]; i++) {
	    fin >> u[m][i][j][k];
	  }
	}
      }
    }
    
    for (m=0; m<NumberOfMeshes; m++) {
      for (k=0; k<pointNumberMeshV[m][2]; k++) {
	for (j=0; j<pointNumberMeshV[m][1]; j++) {
	  for (i=0; i<pointNumberMeshV[m][0]; i++) {
	    fin >> v[m][i][j][k];
	  }
	}
      }
    }
    
    for (m=0; m<NumberOfMeshes; m++) {
      for (k=0; k<pointNumberMeshW[m][2]; k++) {
	for (j=0; j<pointNumberMeshW[m][1]; j++) {
	  for (i=0; i<pointNumberMeshW[m][0]; i++) {
	    fin >> w[m][i][j][k];
	  }
	}
      }
    }
    
    for (m=0; m<NumberOfMeshes; m++) {
      for (k=0; k<pointNumberMeshP[m][2]; k++) {
	for (j=0; j<pointNumberMeshP[m][1]; j++) {
	  for (i=0; i<pointNumberMeshP[m][0]; i++) {
	    fin >> p[m][i][j][k];
	  }
	}
      }
    }
    */
    fin.close() ;
    
  }


  if (restart == 0) {  
    initStage iSObj;
    iSObj.solve(nPatch,
		t,
		dt,
		Re,
		Nc,
		cellNumberPerPatch, 
		StartIndex, 
		xParticleCenter,  // coordinates of particle center
		yParticleCenter, 
		zParticleCenter, 
		pointNumberMeshU, 
		pointNumberMeshV, 
		pointNumberMeshW, 
		pointNumberMeshP, 
		dim,
		xU, 
		xV, 
		xW, 
		xP, 
		h,
		u,
		uUnknowns,
		v,
		vUnknowns,
		w,
		wUnknowns,
		p,
		uAux,
		vAux,
		wAux,
		fp, 
		exactP, 
		pUnknowns,
		uCagePoint,
		vCagePoint,
		wCagePoint,
		pCagePoint,
		nCell,
		pSolver,
		zeroInitialGuess);
  }             

	      
  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshU[m][2]; k++) {
      for (j=0; j<pointNumberMeshU[m][1]; j++) {
	for (i=0; i<pointNumberMeshU[m][0]; i++) {
	  u[m][i][j][k] = u[m][i][j][k] * uUnknowns[m][i][j][k];
	}
      }
    }
  }

  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshV[m][2]; k++) {
      for (j=0; j<pointNumberMeshV[m][1]; j++) {
	for (i=0; i<pointNumberMeshV[m][0]; i++) {
	  v[m][i][j][k] = v[m][i][j][k] * vUnknowns[m][i][j][k];
	}
      }
    }
  }

  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshW[m][2]; k++) {
      for (j=0; j<pointNumberMeshW[m][1]; j++) {
	for (i=0; i<pointNumberMeshW[m][0]; i++) {
	  w[m][i][j][k] = w[m][i][j][k] * wUnknowns[m][i][j][k];
	}
      }
    }
  }


  filename = "results/restartFieldOld.dat";
  
  ofParticle.open( filename.c_str() ) ;
  /*
  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshU[m][2]; k++) {
      for (j=0; j<pointNumberMeshU[m][1]; j++) {
	for (i=0; i<pointNumberMeshU[m][0]; i++) {
	  ofParticle<< u[m][i][j][k] <<endl;
	}
      }
    }
  }
  
  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshV[m][2]; k++) {
      for (j=0; j<pointNumberMeshV[m][1]; j++) {
	for (i=0; i<pointNumberMeshV[m][0]; i++) {
	  ofParticle<< v[m][i][j][k] <<endl;
	}
      }
    }
  }

  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshW[m][2]; k++) {
      for (j=0; j<pointNumberMeshW[m][1]; j++) {
	for (i=0; i<pointNumberMeshW[m][0]; i++) {
	  ofParticle<< w[m][i][j][k] <<endl;
	}
      }
    }
  }

  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshP[m][2]; k++) {
      for (j=0; j<pointNumberMeshP[m][1]; j++) {
	for (i=0; i<pointNumberMeshP[m][0]; i++) {
	  ofParticle<< p[m][i][j][k] <<endl;
	}
      }
    }
  }
  */
  ofParticle.close();

 
  outputData oObj;
  oObj.initialize(tStepInit,
		  NumberOfMeshes,
		  cellNumberPerPatch,
		  StartIndex, 
		  pointNumberMeshU, 
		  pointNumberMeshV, 
		  pointNumberMeshW, 
		  pointNumberMeshP, 
		  xU, 
		  xV, 
		  xW, 
		  xP, 
		  u, 
		  v,
		  w,
		  p);
  
  // to calculate U, S, V for 
  // fixed particles
  // with same geometry
  //fixedParticles = 0;

  /*
  for (m=1; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshOmegaX[m][2]; k++) { // interior points
      for (j=0; j<pointNumberMeshOmegaX[m][1]; j++) { // interior points
	for (i=0; i<pointNumberMeshOmegaX[m][0]; i++) {

	  if (  sqrt( pow( xOmegaX[m][0][i] - xParticleCenter[m], 2.0 )
		      + pow( xOmegaX[m][1][j] - yParticleCenter[m], 2.0 ) 
		      + pow( xOmegaX[m][2][k] - zParticleCenter[m], 2.0 ) )
		> 1.0 + 1.01 * h[m] ){
	    omegaXCagePoint[m][i][j][k] = 0; // initialize
	  }
	      
	}
      }
    }
  }

  for (m=1; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshOmegaY[m][2]; k++) { // interior points
      for (j=0; j<pointNumberMeshOmegaY[m][1]; j++) { // interior points
	for (i=0; i<pointNumberMeshOmegaY[m][0]; i++) {

	  if (  sqrt( pow( xOmegaY[m][0][i] - xParticleCenter[m], 2.0 )
		      + pow( xOmegaY[m][1][j] - yParticleCenter[m], 2.0 ) 
		      + pow( xOmegaY[m][2][k] - zParticleCenter[m], 2.0 ) )
		> 1.0 + 1.01 * h[m] ){
	    omegaYCagePoint[m][i][j][k] = 0; // initialize
	  }

	}
      }
    }
  }

  for (m=1; m<=nPatch; m++) {
    for (k=0; k<pointNumberMeshOmegaZ[m][2]; k++) { // interior points
      for (j=0; j<pointNumberMeshOmegaZ[m][1]; j++) { // interior points
	for (i=0; i<pointNumberMeshOmegaZ[m][0]; i++) {

	  if (  sqrt( pow( xOmegaZ[m][0][i] - xParticleCenter[m], 2.0 )
		      + pow( xOmegaZ[m][1][j] - yParticleCenter[m], 2.0 ) 
		      + pow( xOmegaZ[m][2][k] - zParticleCenter[m], 2.0 ) )
		> 1.0 + 1.01 * h[m] ){
	    omegaZCagePoint[m][i][j][k] = 0; // initialize
	  }

	}
      }
    }
  }
  */

for (tStep=1+tStepInit; tStep<=30000; tStep++) {

  cout <<endl;
  cout <<"tStep = " << tStep <<endl;

  exactSol exactObj;
  exactObj.initialize(NumberOfMeshes,
		      t+dt,
		      pointNumberMeshU, 
		      xU, 
		      pointNumberMeshV, 
		      xV, 
		      pointNumberMeshW, 
		      xW, 
		      pointNumberMeshP, 
		      xP, 
		      exactU, 
		      exactV,
		      exactW,
		      exactP);

  setDt sObj;
  sObj.initialize(nPatch,
		  dt,
		  Re,
		  pointNumberMeshU, 
		  pointNumberMeshV, 
		  pointNumberMeshW, 
		  h,
		  u,
		  uUnknowns,
		  v,
		  vUnknowns,
		  w,
		  wUnknowns);

  NSSolver NSObj;
  NSObj.solve(nPatch,
	      t,
	      dt,
	      Re,
	      Nc,
	      cellNumberPerPatch, 
	      StartIndex, 
	      xParticleCenter,  // coordinates of particle center
	      yParticleCenter, 
	      zParticleCenter, 
	      pointNumberMeshU, 
	      pointNumberMeshV, 
	      pointNumberMeshW, 
	      pointNumberMeshP, 
	      pointNumberMeshOmegaX, 
	      pointNumberMeshOmegaY, 
	      pointNumberMeshOmegaZ, 
	      dim,
	      xU, 
	      xV, 
	      xW, 
	      xP, 
	      xOmegaX, 
	      xOmegaY, 
	      xOmegaZ, 
	      h,
	      u,
	      v,
	      w,
	      p,
	      omegaX,
	      omegaY,
	      omegaZ,
	      uAux,
	      vAux,
	      wAux,
	      fp, 
	      exactP, 
	      pUnknowns,
	      uCagePoint,
	      vCagePoint,
	      wCagePoint,
	      pCagePoint,
	      omegaXCagePoint,
	      omegaYCagePoint,
	      omegaZCagePoint,
	      UFixedParticles,
	      SFixedParticles,
	      VFixedParticles,
	      fixedParticles,
	      rotatingParticles,
	      angVel,
	      angVelAcc,
	      torqueOld,
	      torqueNew,
	      pSolver,
	      zeroInitialGuess);

  t = t + dt;

  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshU[m][2]; k++) {
      for (j=0; j<pointNumberMeshU[m][1]; j++) {
	for (i=0; i<pointNumberMeshU[m][0]; i++) {
	  u[m][i][j][k] = u[m][i][j][k] * uUnknowns[m][i][j][k];
	}
      }
    }
  }

  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshV[m][2]; k++) {
      for (j=0; j<pointNumberMeshV[m][1]; j++) {
	for (i=0; i<pointNumberMeshV[m][0]; i++) {
	  v[m][i][j][k] = v[m][i][j][k] * vUnknowns[m][i][j][k];
	}
      }
    }
  }

  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshW[m][2]; k++) {
      for (j=0; j<pointNumberMeshW[m][1]; j++) {
	for (i=0; i<pointNumberMeshW[m][0]; i++) {
	  w[m][i][j][k] = w[m][i][j][k] * wUnknowns[m][i][j][k];
	}
      }
    }
  }

  if (tStep % 3000 == 0 ) {

  filename = "results/restartFieldNew.dat";
  
  ofParticle.open( filename.c_str() ) ;
  /*
  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshU[m][2]; k++) {
      for (j=0; j<pointNumberMeshU[m][1]; j++) {
	for (i=0; i<pointNumberMeshU[m][0]; i++) {
	  ofParticle<< u[m][i][j][k] <<endl;
	}
      }
    }
  }

  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshV[m][2]; k++) {
      for (j=0; j<pointNumberMeshV[m][1]; j++) {
	for (i=0; i<pointNumberMeshV[m][0]; i++) {
	  ofParticle<< v[m][i][j][k] <<endl;
	}
      }
    }
  }

  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshW[m][2]; k++) {
      for (j=0; j<pointNumberMeshW[m][1]; j++) {
	for (i=0; i<pointNumberMeshW[m][0]; i++) {
	  ofParticle<< w[m][i][j][k] <<endl;
	}
      }
    }
  }

  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshP[m][2]; k++) {
      for (j=0; j<pointNumberMeshP[m][1]; j++) {
	for (i=0; i<pointNumberMeshP[m][0]; i++) {
	  ofParticle<< p[m][i][j][k] <<endl;
	}
      }
    }
  }
  */
  ofParticle.close();

  filename = "results/tStepInitNew.dat"; 
  ofParticle.open( filename.c_str() ) ;
  ofParticle<< tStep <<endl;
  ofParticle.close();

  int temp1, temp2;
  temp1 = system("cp results/restartFieldNew.dat results/restartFieldOld.dat");

  temp2 = system("cp results/tStepInitNew.dat results/tStepInitOld.dat");
 
  temp1 = temp2 +1;
  }
 
  if (tStep%1000 == 0 ) {
    
  outputData oObj;
  oObj.initialize(tStep/1000,
		  NumberOfMeshes,
		  cellNumberPerPatch,
		  StartIndex, 
		  pointNumberMeshU, 
		  pointNumberMeshV, 
		  pointNumberMeshW, 
		  pointNumberMeshP, 
		  xU, 
		  xV, 
		  xW, 
		  xP, 
		  u, 
		  v,
		  w,
		  p);
  }  
  

  /*
  real maxError = 0.0;
  //    if (tStep%10 == 0 ) {
    //    cout << "tStep = " << tStep << "  :";
    for (m=0; m<NumberOfMeshes; m++) {
      for (k=0; k<pointNumberMeshW[m][2]; k++) {
	for (j=0; j<pointNumberMeshW[m][1]; j++) {
	  for (i=0; i<pointNumberMeshW[m][0]; i++) {
	    //	  real temp = fabs( (exactU[m][i][j] -u[m][i][j]) )
	    //	    *uUnknowns[m][i][j] * exp(2.0*t);

	    real temp = fabs( (exactW[m][i][j][k] -w[m][i][j][k]) )
	      *wUnknowns[m][i][j][k] * exp(3.0*t)/2.0;

	    //	  real temp = fabs( exactP[m][i][j] )
	    //	    *pUnknowns[m][i][j] * exp(4.0*t) *2.0;
	    if (temp > maxError)  maxError = temp;
	    //	    	    cout <<i<<" "<<j<<" "<<k<<" " <<temp << endl;
	  }
	  //	  cout << endl;
	}
	//	cout << endl;
      }
      //      cout << "--------------------"<<endl;
    }
    //  cout <<"max error is " <<maxError << endl;
    //             cout <<tStep*dt <<" " << maxError << endl;
             cout <<tStep << endl;
	     //    }
  */


 }
 

  fin.close() ;

 pSolver.finalize();
	
  releaseMemory();

}


void flowSolver::setUnknowns(){
  //set exact solutions and forces for coarse and fine meshes
  // is unknowns that are solved in the linear system?
  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshP[m][2]; k++) {
      for (j=0; j<pointNumberMeshP[m][1]; j++) {
	for (i=0; i<pointNumberMeshP[m][0]; i++) {
	  pUnknowns[m][i][j][k]= 1;
	}
      }
    }
  }

  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshU[m][2]; k++) {
      for (j=0; j<pointNumberMeshU[m][1]; j++) {
	for (i=0; i<pointNumberMeshU[m][0]; i++) {
	  uUnknowns[m][i][j][k]= 1;
	}
      }
    }
  }
  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshV[m][2]; k++) {
      for (j=0; j<pointNumberMeshV[m][1]; j++) {
	for (i=0; i<pointNumberMeshV[m][0]; i++) {
	  vUnknowns[m][i][j][k]= 1;
	}
      }
    }
  }
  for (m=0; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshW[m][2]; k++) {
      for (j=0; j<pointNumberMeshW[m][1]; j++) {
	for (i=0; i<pointNumberMeshW[m][0]; i++) {
	  wUnknowns[m][i][j][k]= 1;
	}
      }
    }
  }


  //coarse mesh:  some points inside fine mesh
  for (m=1; m<=nPatch; m++) {
    for (k=StartIndex[m][2]+1; k<=StartIndex[m][2]+cellNumberPerPatch-2; k++){
      for (j=StartIndex[m][1]+1; j<=StartIndex[m][1]+cellNumberPerPatch-2; j++){
	for (i=StartIndex[m][0]+1; i<=StartIndex[m][0]+cellNumberPerPatch-2; i++){
	  pUnknowns[0][i][j][k] = 0;
	}    
      }  
    }
  }
  for (m=1; m<=nPatch; m++) {
    for (k=StartIndex[m][2]+2; k<=StartIndex[m][2]+cellNumberPerPatch-1; k++){
      for (j=StartIndex[m][1]+2; j<=StartIndex[m][1]+cellNumberPerPatch-1; j++){
	for (i=StartIndex[m][0]+2; i<=StartIndex[m][0]+cellNumberPerPatch-2; i++){
	  uUnknowns[0][i][j][k] = 0;
	}    
      }  
    }
  }
  for (m=1; m<=nPatch; m++) {
    for (k=StartIndex[m][2]+2; k<=StartIndex[m][2]+cellNumberPerPatch-1; k++){
      for (j=StartIndex[m][1]+2; j<=StartIndex[m][1]+cellNumberPerPatch-2; j++){
	for (i=StartIndex[m][0]+2; i<=StartIndex[m][0]+cellNumberPerPatch-1; i++){
	  vUnknowns[0][i][j][k] = 0;
	}    
      }  
    }
  }
  for (m=1; m<=nPatch; m++) {
    for (k=StartIndex[m][2]+2; k<=StartIndex[m][2]+cellNumberPerPatch-2; k++){
      for (j=StartIndex[m][1]+2; j<=StartIndex[m][1]+cellNumberPerPatch-1; j++){
	for (i=StartIndex[m][0]+2; i<=StartIndex[m][0]+cellNumberPerPatch-1; i++){
	  wUnknowns[0][i][j][k] = 0;
	}    
      }  
    }
  }


  //fine mesh:  initialize to zero
  for (m=1; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshP[m][2]; k++) {
      for (j=0; j<pointNumberMeshP[m][1]; j++) {
	for (i=0; i<pointNumberMeshP[m][0]; i++) {
	  pUnknowns[m][i][j][k]= 0;
	}
      }
    }
  }

  for (m=1; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshU[m][2]; k++) {
      for (j=0; j<pointNumberMeshU[m][1]; j++) {
	for (i=0; i<pointNumberMeshU[m][0]; i++) {
	  uUnknowns[m][i][j][k]= 0;
	}
      }
    }
  }

  for (m=1; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshV[m][2]; k++) {
      for (j=0; j<pointNumberMeshV[m][1]; j++) {
	for (i=0; i<pointNumberMeshV[m][0]; i++) {
	  vUnknowns[m][i][j][k]= 0;
	}
      }
    }
  }

  for (m=1; m<NumberOfMeshes; m++) {
    for (k=0; k<pointNumberMeshW[m][2]; k++) {
      for (j=0; j<pointNumberMeshW[m][1]; j++) {
	for (i=0; i<pointNumberMeshW[m][0]; i++) {
	  wUnknowns[m][i][j][k]= 0;
	}
      }
    }
  }

  
}



  //set coordinates for coarse mesh
  // i.e. i for x direction and j for y direction
  // NOTE: sweep in horizontal direction and then move up
  // NOTE: the order of i and j for loops
void flowSolver::setCoord(){
  
  //
  //coordinates for velocity u
  //
  for (i=0; i<pointNumberMeshU[0][0]; i++) {
    xU[0][0][i] = i*h[0];
  }
  for (j=0; j<pointNumberMeshU[0][1]; j++) {
    xU[0][1][j] = -hCoarse/2.0 + j*h[0];
  }
  for (k=0; k<pointNumberMeshU[0][2]; k++) {
    xU[0][2][k] = -hCoarse/2.0 + k*h[0];
  }

  //moving (-hCoarse, -hCoarse)
  //set for patchs in fine mesh
  for (m=1; m<NumberOfMeshes; m++) {
    for (i=0; i<pointNumberMeshU[m][0]; i++) {
      int iStart = StartIndex[m][0];
      xU[m][0][i] = xU[0][0][iStart] - hCoarse + i*h[m];
    }
    for (j=0; j<pointNumberMeshU[m][1]; j++) {
      int jStart = StartIndex[m][1];
      xU[m][1][j] = xU[0][1][jStart] - hCoarse/4.0 + j*h[m];
    }
    for (k=0; k<pointNumberMeshU[m][2]; k++) {
      int kStart = StartIndex[m][2];
      xU[m][2][k] = xU[0][2][kStart] - hCoarse/4.0 + k*h[m];
    }
    
  }
  
  //
  //coordinates for velocity v
  //
  for (i=0; i<pointNumberMeshV[0][0]; i++) {
    xV[0][0][i] = -hCoarse/2.0 + i*h[0];
  }
  for (j=0; j<pointNumberMeshV[0][1]; j++) {
    xV[0][1][j] = j*h[0];
  }
  for (k=0; k<pointNumberMeshV[0][2]; k++) {
    xV[0][2][k] = -hCoarse/2.0 + k*h[0];
  }


  //moving (-hCoarse, -hCoarse)
  //set for patchs in fine mesh
  for (m=1; m<NumberOfMeshes; m++) {
    for (i=0; i<pointNumberMeshV[m][0]; i++) {
      int iStart = StartIndex[m][0];
      xV[m][0][i] = xV[0][0][iStart] - hCoarse/4.0 + i*h[m];
    }
    for (j=0; j<pointNumberMeshV[m][1]; j++) {
      int jStart = StartIndex[m][1];
      xV[m][1][j] = xV[0][1][jStart] - hCoarse + j*h[m];
    }
    for (k=0; k<pointNumberMeshV[m][2]; k++) {
      int kStart = StartIndex[m][2];
      xV[m][2][k] = xV[0][2][kStart] - hCoarse/4.0 + k*h[m];
    }
  }

  //
  //coordinates for velocity w
  //
  for (i=0; i<pointNumberMeshW[0][0]; i++) {
    xW[0][0][i] = -hCoarse/2.0 + i*h[0];
  }
  for (j=0; j<pointNumberMeshW[0][1]; j++) {
    xW[0][1][j] = -hCoarse/2.0 + j*h[0];
  }
  for (k=0; k<pointNumberMeshW[0][2]; k++) {
    xW[0][2][k] = k*h[0];
  }

  //moving (-hCoarse, -hCoarse)
  //set for patchs in fine mesh
  for (m=1; m<NumberOfMeshes; m++) {

    for (i=0; i<pointNumberMeshW[m][0]; i++) {
      int iStart = StartIndex[m][0];
      xW[m][0][i] = xW[0][0][iStart] - hCoarse/4.0 + i*h[m];
    }
    for (j=0; j<pointNumberMeshW[m][1]; j++) {
      int jStart = StartIndex[m][1];
      xW[m][1][j] = xW[0][1][jStart] - hCoarse/4.0 + j*h[m];
    }	
    for (k=0; k<pointNumberMeshW[m][2]; k++) {
      int kStart = StartIndex[m][2];
      xW[m][2][k] = xW[0][2][kStart] - hCoarse + k*h[m];
    }
  }

  //
  //coordinates for pressure p
  //
  for (i=0; i<pointNumberMeshP[0][0]; i++) {
    xP[0][0][i] = hCoarse/2.0 + i*h[0];
  }
  for (j=0; j<pointNumberMeshP[0][1]; j++) {
    xP[0][1][j] = hCoarse/2.0 + j*h[0];
  }
  for (k=0; k<pointNumberMeshP[0][2]; k++) {
    xP[0][2][k] = hCoarse/2.0 + k*h[0];
  }

  //moving (-hCoarse, -hCoarse)
  //set for patchs in fine mesh
  for (m=1; m<NumberOfMeshes; m++) {
    for (i=0; i<pointNumberMeshP[m][0]; i++) {
      int iStart = StartIndex[m][0];
      xP[m][0][i] = xP[0][0][iStart] 
	- 5.0 * hCoarse/4.0 + i*h[m];
    }
    for (j=0; j<pointNumberMeshP[m][1]; j++) {
      int jStart = StartIndex[m][1];
      xP[m][1][j] = xP[0][1][jStart] 
	- 5.0 * hCoarse/4.0 + j*h[m];
    }
    for (k=0; k<pointNumberMeshP[m][2]; k++) {
      int kStart = StartIndex[m][2];
      xP[m][2][k] = xP[0][2][kStart] 
	- 5.0 * hCoarse/4.0 + k*h[m];
    }

  }

  //
  //coordinates for vorticity
  //
  for (i=0; i<pointNumberMeshOmegaX[0][0]; i++) {
    xOmegaX[0][0][i] = xW[0][0][0] + i*h[m];
  }
  for (j=0; j<pointNumberMeshOmegaX[0][1]; j++) {
    xOmegaX[0][1][j] = j*h[0];
  }
  for (k=0; k<pointNumberMeshOmegaX[0][2]; k++) {
    xOmegaX[0][2][k] = k*h[0];
  }
 
  //set for patchs in fine mesh
  for (m=1; m<NumberOfMeshes; m++) {
    for (i=0; i<pointNumberMeshOmegaX[m][0]; i++) {
      xOmegaX[m][0][i] = xW[m][0][0] + i*h[m];
    }
    for (j=0; j<pointNumberMeshOmegaX[m][1]; j++) {
      xOmegaX[m][1][j] = xV[m][1][0] + 1.0*hCoarse/2.0 + j*h[m];
    }
    for (k=0; k<pointNumberMeshOmegaX[m][2]; k++) {
      xOmegaX[m][2][k] = xW[m][2][0] + 1.0*hCoarse/2.0 + k*h[m];
    }
  }

  for (i=0; i<pointNumberMeshOmegaY[0][0]; i++) {
    xOmegaY[0][0][i] = i*h[0];
  }
  for (j=0; j<pointNumberMeshOmegaY[0][1]; j++) {
    xOmegaY[0][1][j] = xU[0][1][0] + j*h[m];
  }
  for (k=0; k<pointNumberMeshOmegaY[0][2]; k++) {
    xOmegaY[0][2][k] = k*h[0];
  }


  //set for patchs in fine mesh
  for (m=1; m<NumberOfMeshes; m++) {
    for (i=0; i<pointNumberMeshOmegaY[m][0]; i++) {
      xOmegaY[m][0][i] = xU[m][0][0] + 1.0*hCoarse/2.0 + i*h[m];
    }
    for (j=0; j<pointNumberMeshOmegaY[m][1]; j++) {
      xOmegaY[m][1][j] = xU[m][1][0] + j*h[m];
    }
    for (k=0; k<pointNumberMeshOmegaY[m][2]; k++) {
      xOmegaY[m][2][k] = xW[m][2][0] + 1.0*hCoarse/2.0 + k*h[m];
    }
  }
  

  for (i=0; i<pointNumberMeshOmegaZ[0][0]; i++) {
    xOmegaZ[0][0][i] = i*h[0];
  }
  for (j=0; j<pointNumberMeshOmegaZ[0][1]; j++) {
    xOmegaZ[0][1][j] = j*h[0];
  }
  for (k=0; k<pointNumberMeshOmegaZ[0][2]; k++) {
    xOmegaZ[0][2][k] = xV[0][2][0] + k*h[m];
  }
 
  //set for patchs in fine mesh
  for (m=1; m<NumberOfMeshes; m++) {
    for (i=0; i<pointNumberMeshOmegaZ[m][0]; i++) {
      xOmegaZ[m][0][i] = xU[m][0][0] + 1.0*hCoarse/2.0 + i*h[m];
    }
    for (j=0; j<pointNumberMeshOmegaZ[m][1]; j++) {
      xOmegaZ[m][1][j] = xV[m][1][0] + 1.0*hCoarse/2.0 + j*h[m];
    }
    for (k=0; k<pointNumberMeshOmegaZ[m][2]; k++) {
      xOmegaZ[m][2][k] = xV[m][2][0] + k*h[m];
    }
  }

}



int * flowSolver::allocate1dIntArray(int size){
// pointer can be used only after initialized

  int *x = new int[size];
  return x; 
}
void flowSolver::delete1dArray(int *x){
  delete [] x;
}

real * flowSolver::allocate1dRealArray(int size){
  real *x = new real[size];
  return x;
}
void flowSolver::delete1dArray(real *x){
  delete [] x;
}


int ** flowSolver::allocate2dArray(int nMeshes, int dimension){
  int **x = new int *[nMeshes];
  for (m=0; m<nMeshes; m++) {
    x[m] = new int[dimension];
  }   
  return x;
}
void flowSolver::delete2dArray(int **x, int nMeshes){
  for (m=0; m<nMeshes; m++) {
    delete [] x[m];
  }   
  delete [] x;
}

// allocate inflow conditions
real ** flowSolver::allocate2dRealArray(int ny, int nz){
  real **x = new real *[ny];
  for (m=0; m<ny; m++) {
    x[m] = new real[nz];
  }   
  return x;
}
void flowSolver::delete2dRealArray(real **x, int ny){
  for (m=0; m<ny; m++) {
    delete [] x[m];
  }   
  delete [] x;
}


// allocate coordinates
real *** flowSolver::allocate3dRealArray(int nMeshes, int dim, int **pointNumberMesh){
  real ***x = new real **[nMeshes];
  for (m=0; m<nMeshes; m++) {
    x[m] = new real *[dim];
    for (i=0; i<dim; i++) {
      x[m][i] = new real[pointNumberMesh[m][i]];
    }
  }
  return x;
}

void flowSolver::delete3dArray(real ***x, int nMeshes, int dim){
  for (m=0; m<nMeshes; m++) {
    for (i=0; i<dim; i++) {
      delete []  x[m][i];
    }
   delete [] x[m];
  }
  delete [] x;
}

real **** flowSolver::allocate4dRealArray(int nMeshes, int **pointNumberMesh){
  real ****x = new real ***[nMeshes];
  for (m=0; m<nMeshes; m++) {
    x[m] = new real **[pointNumberMesh[m][0]];
    for (i=0; i<pointNumberMesh[m][0]; i++) {
      x[m][i] = new real*[pointNumberMesh[m][1]];
      for (j=0; j<pointNumberMesh[m][1]; j++) {
	x[m][i][j] = new real[pointNumberMesh[m][2]];
      }
    }
  }
  return x;
}

void flowSolver::delete4dArray(real ****x, int nMeshes, int **pointNumberMesh){
  for (m=0; m<nMeshes; m++) {
    for (i=0; i<pointNumberMesh[m][0]; i++) {
      for (j=0; j<pointNumberMesh[m][1]; j++) {
	delete [] x[m][i][j];
      }
      delete [] x[m][i];
    }
    delete [] x[m];
  }
  delete [] x;
}

int **** flowSolver::allocate4dIntArray(int nMeshes, int **pointNumberMesh){

  int ****x = new int ***[nMeshes];
  for (m=0; m<nMeshes; m++) {
    x[m] = new int **[pointNumberMesh[m][0]];
    for (i=0; i<pointNumberMesh[m][0]; i++) {
      x[m][i] = new int*[pointNumberMesh[m][1]];
      for (j=0; j<pointNumberMesh[m][1]; j++) {
	x[m][i][j] = new int[pointNumberMesh[m][2]];
      }
    }
  }
  return x;
}
void flowSolver::delete4dArray(int ****x, int nMeshes, int **pointNumberMesh){
  for (m=0; m<nMeshes; m++) {
    for (i=0; i<pointNumberMesh[m][0]; i++) {
      for (j=0; j<pointNumberMesh[m][1]; j++) {
	delete [] x[m][i][j];
      }
      delete [] x[m][i];
    }
    delete [] x[m];
  }
  delete [] x;
}

real **** flowSolver::allocate4dRealArrayCage(int nMeshes, int **pointNumberMesh){
  real ****x = new real ***[nMeshes];
  for (m=1; m<nMeshes; m++) {
    x[m] = new real **[pointNumberMesh[m][0]];
    for (i=0; i<pointNumberMesh[m][0]; i++) {
      x[m][i] = new real*[pointNumberMesh[m][1]];
      for (j=0; j<pointNumberMesh[m][1]; j++) {
	x[m][i][j] = new real[pointNumberMesh[m][2]];
      }
    }
  }
  return x;
}

void flowSolver::delete4dArrayCage(real ****x, int nMeshes, int **pointNumberMesh){
  for (m=1; m<nMeshes; m++) {
    for (i=0; i<pointNumberMesh[m][0]; i++) {
      for (j=0; j<pointNumberMesh[m][1]; j++) {
	delete [] x[m][i][j];
      }
      delete [] x[m][i];
    }
    delete [] x[m];
  }
  delete [] x;
}

int **** flowSolver::allocate4dIntArrayCage(int nMeshes, int **pointNumberMesh){

  int ****x = new int ***[nMeshes];
  for (m=1; m<nMeshes; m++) {
    x[m] = new int **[pointNumberMesh[m][0]];
    for (i=0; i<pointNumberMesh[m][0]; i++) {
      x[m][i] = new int*[pointNumberMesh[m][1]];
      for (j=0; j<pointNumberMesh[m][1]; j++) {
	x[m][i][j] = new int[pointNumberMesh[m][2]];
      }
    }
  }
  return x;
}
void flowSolver::delete4dArrayCage(int ****x, int nMeshes, int **pointNumberMesh){
  for (m=1; m<nMeshes; m++) {
    for (i=0; i<pointNumberMesh[m][0]; i++) {
      for (j=0; j<pointNumberMesh[m][1]; j++) {
	delete [] x[m][i][j];
      }
      delete [] x[m][i];
    }
    delete [] x[m];
  }
  delete [] x;
}



void flowSolver::releaseMemory(){

  delete1dArray(h);

  delete2dArray(StartIndex,NumberOfMeshes );
  delete2dArray(nCell,NumberOfMeshes );

  delete3dArray(xP, NumberOfMeshes, dim);
  delete4dArray(fp, NumberOfMeshes, pointNumberMeshP);
  delete4dArray(exactP, NumberOfMeshes, pointNumberMeshP);
  delete4dArray(pUnknowns, NumberOfMeshes, pointNumberMeshP);
  delete4dArray(p, NumberOfMeshes, pointNumberMeshP);

  delete3dArray(xU, NumberOfMeshes, dim);
  delete4dArray(exactU, NumberOfMeshes, pointNumberMeshU);
  delete4dArray(uUnknowns, NumberOfMeshes, pointNumberMeshU);
  delete4dArray(u, NumberOfMeshes, pointNumberMeshU);
  delete4dArray(uAux, NumberOfMeshes, pointNumberMeshU);

  delete3dArray(xV, NumberOfMeshes, dim);
  delete4dArray(exactV, NumberOfMeshes, pointNumberMeshV);
  delete4dArray(vUnknowns, NumberOfMeshes, pointNumberMeshV);
  delete4dArray(v, NumberOfMeshes, pointNumberMeshV);
  delete4dArray(vAux, NumberOfMeshes, pointNumberMeshV);

  delete3dArray(xW, NumberOfMeshes, dim);
  delete4dArray(exactW, NumberOfMeshes, pointNumberMeshW);
  delete4dArray(wUnknowns, NumberOfMeshes, pointNumberMeshW);
  delete4dArray(w, NumberOfMeshes, pointNumberMeshW);
  delete4dArray(wAux, NumberOfMeshes, pointNumberMeshW);


  //particles
  delete1dArray(xParticleCenter);
  delete1dArray(yParticleCenter);
  delete1dArray(zParticleCenter);
  delete4dArrayCage(pInsideParticle, NumberOfMeshes, pointNumberMeshP);
  delete4dArrayCage(pCagePoint, NumberOfMeshes, pointNumberMeshP);
  delete4dArrayCage(uCagePoint, NumberOfMeshes, pointNumberMeshU);
  delete4dArrayCage(vCagePoint, NumberOfMeshes, pointNumberMeshV);
  delete4dArrayCage(wCagePoint, NumberOfMeshes, pointNumberMeshW);

  delete3dArray(xOmegaX, NumberOfMeshes, dim);
  delete4dArrayCage(omegaXCagePoint, NumberOfMeshes, pointNumberMeshOmegaX);
  delete4dArray(omegaX, NumberOfMeshes, pointNumberMeshOmegaX);

  delete3dArray(xOmegaY, NumberOfMeshes, dim);
  delete4dArrayCage(omegaYCagePoint, NumberOfMeshes, pointNumberMeshOmegaY);
  delete4dArray(omegaY, NumberOfMeshes, pointNumberMeshOmegaY);
 
  delete3dArray(xOmegaZ, NumberOfMeshes, dim);
  delete4dArrayCage(omegaZCagePoint, NumberOfMeshes, pointNumberMeshOmegaZ);
  delete4dArray(omegaZ, NumberOfMeshes, pointNumberMeshOmegaZ);

  //NOTE: pointNumberMeshP must be released last
  //because it is needed to realesed 3d arrayes
  delete2dArray(pointNumberMeshP,NumberOfMeshes );
  delete2dArray(pointNumberMeshU,NumberOfMeshes );
  delete2dArray(pointNumberMeshV,NumberOfMeshes );
  delete2dArray(pointNumberMeshW,NumberOfMeshes );
  delete2dArray(pointNumberMeshOmegaX,NumberOfMeshes );
  delete2dArray(pointNumberMeshOmegaY,NumberOfMeshes );
  delete2dArray(pointNumberMeshOmegaZ,NumberOfMeshes );

}

string flowSolver::iToString(int k){
  // Converts an integer to a string
  // Requires <sstream> include file

  stringstream s;
  s <<  k;
  return s.str();
}

flowSolver::~flowSolver(){
}
