#include "pressure.h"


pressure::pressure(){
}

void pressure::initialize(int numberOfPatches, 
			  int cellNumberPerPatch, 
			  int ** StartIndex, 
			  int ** pointNumberMeshP, 
			  int dimension,
			  real ***xP, 
			  real *h,
			  real ****fp, 
			  real ****exactPressure, 
			  int ****pUnknowns,
			  real ****p,
			  int ****pCagePoint,
			  PetscLinearSolver &pSolver,
			  int &zeroInitialGuess)
{

  nPatch = numberOfPatches;
  nMeshes = numberOfPatches + 1;
  nCellPerPatch = cellNumberPerPatch;
  dim = dimension;
  tol = 1.0e-9; 

  // total number of unknowns in the coarse and fine meshes
  n = 0;
  for (m=0; m<=nPatch; m++) {
    for (k=0; k<=pointNumberMeshP[m][2]-1; k++){
      for (j=0; j<=pointNumberMeshP[m][1]-1; j++){
	for (i=0; i<=pointNumberMeshP[m][0]-1; i++){
	  n += pUnknowns[m][i][j][k]; 
	}
      }
    }
  }

  //    using 7 points for poisson equation
  nStencil = 9; //need 9 points for boundary points
 
  index = allocate4dIntArray(nMeshes, pointNumberMeshP);
 
  int counter = -1; // make the first point with index 0, consistent to c++ convention
  for (m=0; m<nMeshes; m++) {
    for (k=0; k<pointNumberMeshP[m][2]; k++){
      for (j=0; j<pointNumberMeshP[m][1]; j++){
	for (i=0; i<pointNumberMeshP[m][0]; i++){
	  index[m][i][j][k] = counter + pUnknowns[m][i][j][k];
	  counter = index[m][i][j][k];
	}    
      }
    }  
  }

  indexM = allocate1dIntArray(n);
  indexI = allocate1dIntArray(n);
  indexJ = allocate1dIntArray(n);
  indexK = allocate1dIntArray(n);
  // solving linear system
  b = allocate1dRealArray(n); // right hand side
  sol = allocate1dRealArray(n);

  /*
  xCoord = allocate1dRealArray(n);
  yCoord = allocate1dRealArray(n);
  zCoord = allocate1dRealArray(n);
  */

  // since don't include the origin point: [0][0][0]
  // some special treatment is needed here

  for (m=nPatch; m>=0; m--) {
    for (k=pointNumberMeshP[m][2]-1; k>=0; k--) {
      for (j=pointNumberMeshP[m][1]-1; j>=0; j--) {
	for (i=pointNumberMeshP[m][0]-1; i>=0; i--) {
	  iC = index[m][i][j][k];
	  b[iC] = -h[m] * h[m] * fp[m][i][j][k]; 
	  indexM[iC] = m;
	  indexI[iC] = i;
	  indexJ[iC] = j;
	  indexK[iC] = k;
	} 
      } 
    }
  }



  // initialize matrix
  valueA = allocateValueMatrix(); 
  columnA = allocateColumnMatrix(); 
  for (i=0; i<n; i++) {
    for (j=0; j<nStencil; j++) {
      valueA[i][j] = 0.0;      
    }
  }   
  for (i=0; i<n; i++) {
    for (j=0; j<nStencil; j++) {
      columnA[i][j] = 0; //any value is ok      
    }
  }   

  //-------- all interior points for all meshes ---------
  // set interior points other than those of outer boundaries
  // i.e., except left, right, bottom, and top boundary surface.
  // i.e. those patch related points in coarse mesh needed to
  // to be modified later
  // Note: m<=nPatch is equivalent to m<nMeshes


  //////////////////////////////////////////////////////
  //
  // coarse mesh, m = 0
  //
  /////////////////////////////////////////////////////

  // there are five categories of pressure points:
  // 1. a corner point with p = const
  // 2. rest 7 corner points
  // 3. 8 edages
  // 4. 6 surfaces
  // 5. interior points
  // except 1, the rest are included in this for loops

  m=0; 
  for (k=0; k<=pointNumberMeshP[m][2]-1; k++){
    for (j=0; j<=pointNumberMeshP[m][1]-1; j++){
      for (i=0; i<=pointNumberMeshP[m][0]-1; i++){

	// center pressure points
	iC = index[m][i][j][k];
	valueA [iC][0] =  6.0;
	columnA[iC][0] = iC;
	
	// calculate valueA[iC][0]
	// unexisted pressure point
	// use zero flux
	if (i-1 < 0)                        valueA[iC][0] -= 1.0;

	if (i+1 > pointNumberMeshP[m][0]-1) valueA[iC][0] -= 1.0;

	if (j-1 < 0)                        valueA[iC][0] -= 1.0;

	if (j+1 > pointNumberMeshP[m][1]-1) valueA[iC][0] -= 1.0;

	if (k-1 < 0)                        valueA[iC][0] -= 1.0;

	if (k+1 > pointNumberMeshP[m][2]-1) valueA[iC][0] -= 1.0;

	// initialization
	for (iD=1; iD<nStencil; iD++){
	  valueA [iC][iD] =  valueA[iC][0];
	  columnA[iC][iD] = iC;
	} 

	// modify for
	// existed pressure points
	if (i-1 > -1) { // not good to use i-1 >= 0
	  iW = index[m][i-1][j][k];	
	  valueA [iC][1] = -1.0;
	  columnA[iC][1] = iW;
	}

	if (i+1 < pointNumberMeshP[m][0]) {
	  iE = index[m][i+1][j][k];	
	  valueA [iC][2] = -1.0;
	  columnA[iC][2] = iE;
	}

	if (j-1 > -1) { 
	  iS = index[m][i][j-1][k];	
	  valueA [iC][3] = -1.0;
	  columnA[iC][3] = iS;
	}

	if (j+1 < pointNumberMeshP[m][1]) { 
	  iN = index[m][i][j+1][k];	
	  valueA [iC][4] = -1.0;
	  columnA[iC][4] = iN;
	}

	if (k-1 > -1) { 
	  iB = index[m][i][j][k-1];	
	  valueA [iC][5] = -1.0;
	  columnA[iC][5] = iB;
	}

	if (k+1 < pointNumberMeshP[m][2]) { 
	  iT = index[m][i][j][k+1];	
	  valueA [iC][6] = -1.0;
	  columnA[iC][6] = iT;
	}
	
		
      }
    }
  }


  //-------- all interior points for all meshes ---------
  // set interior points other than those of outer boundaries
  // i.e., except left, right, bottom, and top boundary surface.
  // i.e. those patch related points in coarse mesh needed to
  // to be modified later
  // Note: m<=nPatch is equivalent to m<nMeshes

  for (m=1; m<=nPatch; m++) {
    for (k=1; k<=pointNumberMeshP[m][2]-2; k++){
      for (j=1; j<=pointNumberMeshP[m][1]-2; j++){
	for (i=1; i<=pointNumberMeshP[m][0]-2; i++){

	  iC = index[m][i][j][k];
	  valueA [iC][0] =  6.0;
	  columnA[iC][0] = iC;
	  
	  // initialization
	  for (iD=1; iD<nStencil; iD++){
	    valueA [iC][iD] =  valueA[iC][0];
	    columnA[iC][iD] = iC;
	  } 
	  
	  iW = index[m][i-1][j][k];
	  iE = index[m][i+1][j][k];
	  iS = index[m][i][j-1][k];
	  iN = index[m][i][j+1][k];
	  iB = index[m][i][j][k-1];
	  iT = index[m][i][j][k+1];
	  
	  valueA [iC][1] = -1.0;
	  columnA[iC][1] = iW;
	  valueA [iC][2] = -1.0;
	  columnA[iC][2] = iE;
	  valueA [iC][3] = -1.0;
	  columnA[iC][3] = iS;
	  valueA [iC][4] = -1.0;
	  columnA[iC][4] = iN;
	  valueA [iC][5] = -1.0;
	  columnA[iC][5] = iB;
	  valueA [iC][6] = -1.0;
	  columnA[iC][6] = iT;

	}
      }
    }
  }



  //////////////////////////////////////////////////////
  //
  //first, modify special points in coarse mesh, m = 0
  //
  /////////////////////////////////////////////////////

  //
  // outer boundaries
  //

  // west: i= 0
  // east: i = pointNumberMeshP[0][0] - 1

  // south: j = 0
  // north: j = pointNumberMeshP[0][1] - 1

  // bottom: k = 0
  // top: k = pointNumberMeshP[0][2] - 1

  // corners:

  // west south bottom corner
  m = 0;
  i = 0;
  j = 0;
  k = 0;

  /*
  i = pointNumberMeshP[0][0] - 1;
  j = pointNumberMeshP[0][1] - 1;
  k = pointNumberMeshP[0][2] - 1;
  */

  iC = index[m][i][j][k];
  //  iC = 0;
  
  b[iC] = 0.0; //set to zero

  valueA[iC][0] = 6.0;
  columnA[iC][0] = iC;

  for (iD=1; iD<nStencil; iD++){
    valueA [iC][iD] =  valueA[iC][0]; // note: repetition
    columnA[iC][iD] = iC;
  } 

 
  //
  // ghost points
  //points in coarse mesh and inside partilces
  //


  for (m=1; m<=nPatch; m++) {


    //west surface
    i = StartIndex[m][0];
    for (k=StartIndex[m][2]; k<=StartIndex[m][2]+cellNumberPerPatch-1; k++){
      for (j=StartIndex[m][1]; j<=StartIndex[m][1]+cellNumberPerPatch-1; j++){

	// west south bottom corner index
	iFine = 2;
	jFine = 2 + 2 * (j - StartIndex[m][1]);
	kFine = 2 + 2 * (k - StartIndex[m][2]);

	coarseGhostPoint(h, fp);
      }
    }



    //east surface
    i = StartIndex[m][0] + cellNumberPerPatch-1;
    for (k=StartIndex[m][2]; k<=StartIndex[m][2]+cellNumberPerPatch-1; k++){
      for (j=StartIndex[m][1]; j<=StartIndex[m][1]+cellNumberPerPatch-1; j++){

	// west south bottom corner index
	iFine = 2 + 2 * (cellNumberPerPatch - 1);
	jFine = 2 + 2 * (j - StartIndex[m][1]);
	kFine = 2 + 2 * (k - StartIndex[m][2]);
	
	coarseGhostPoint(h, fp);	
      }
    }


    //south surface
    j = StartIndex[m][1];
    for (k=StartIndex[m][2]; k<=StartIndex[m][2]+nCellPerPatch-1; k++){
      for (i=StartIndex[m][0]; i<=StartIndex[m][0]+nCellPerPatch-1; i++){
	
	// west south bottom corner index
	iFine = 2 + 2 * (i - StartIndex[m][0]);
	jFine = 2;
	kFine = 2 + 2 * (k - StartIndex[m][2]);
	
	coarseGhostPoint(h, fp);	
      }
    }     
   
    //north surface
    j = StartIndex[m][1] + cellNumberPerPatch - 1;
    for (k=StartIndex[m][2]; k<=StartIndex[m][2]+cellNumberPerPatch-1; k++){
      for (i=StartIndex[m][0]; i<=StartIndex[m][0]+cellNumberPerPatch-1; i++){
	
	// west south bottom corner index
	iFine = 2 + 2 * (i - StartIndex[m][0]);
	jFine = 2 + 2 * (cellNumberPerPatch - 1);
	kFine = 2 + 2 * (k - StartIndex[m][2]);

	coarseGhostPoint(h, fp);	
      }
    }

    //bottom surface
    k = StartIndex[m][2];
    for (j=StartIndex[m][1]; j<=StartIndex[m][1]+nCellPerPatch-1; j++){
      for (i=StartIndex[m][0]; i<=StartIndex[m][0]+nCellPerPatch-1; i++){
	
	// west south bottom corner index
	iFine = 2 + 2 * (i - StartIndex[m][0]);
	jFine = 2 + 2 * (j - StartIndex[m][1]);
	kFine = 2;
	
	coarseGhostPoint(h, fp);	
      }
    }     
   
    //top surface
    k = StartIndex[m][2] + cellNumberPerPatch - 1;
    for (j=StartIndex[m][1]; j<=StartIndex[m][1]+nCellPerPatch-1; j++){
      for (i=StartIndex[m][0]; i<=StartIndex[m][0]+nCellPerPatch-1; i++){
	
	// west south bottom corner index
	iFine = 2 + 2 * (i - StartIndex[m][0]);
	jFine = 2 + 2 * (j - StartIndex[m][1]);
	kFine = 2 + 2 * (cellNumberPerPatch - 1);

	coarseGhostPoint(h, fp);	
      }
    }
 

  
  //
  //next, construct fine mesh
  //

  // boundary points


    //////////////////////////////////////////////
    //
    //  x direction
    //
    /////////////////////////////////////////////

    //west:
    iFine = 0;
    //odd j, //odd k
    for (kFine=0; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (jFine=0; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){
      
	// west south bottom corner index
	i = StartIndex[m][0]-2;
	j = StartIndex[m][1]-2 + jFine / 2;
	k = StartIndex[m][2]-2 + kFine / 2;

	fineGhostPoint(0, // 0 for odd; 1 for even in x direction
		       0, // 0 for odd; 1 for even in y direction
		       0, // 0 for odd; 1 for even in z direction
		       h,
		       fp);	
      }
    }

    //west:
    iFine = 0;
    //odd j,even k
    for (kFine=1; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (jFine=0; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){
	
	// left bottom corner index
	i = StartIndex[m][0]-2;
	j = StartIndex[m][1]-2 + jFine / 2;
	k = StartIndex[m][2]-1 + (kFine-1) / 2;
	
	fineGhostPoint(0, // 0 for odd; 1 for even in x direction
		       0, // 0 for odd; 1 for even in y direction
		       1, // 0 for odd; 1 for even in z direction
		       h,
		       fp);	
      }
      
    }

    //west:
    iFine = 0;
    //even j, odd k
    for (kFine=0; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (jFine=1; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){
	//pay special attention to corner
	
	// left bottom corner index
	i = StartIndex[m][0]-2;
	j = StartIndex[m][1]-1 + (jFine-1) / 2;
	k = StartIndex[m][2]-2 + kFine / 2;

  	fineGhostPoint(0, // 0 for odd; 1 for even in x direction
		       1, // 0 for odd; 1 for even in y direction
		       0, // 0 for odd; 1 for even in z direction
		       h,
		       fp);	
      }
    }

    //west:
    iFine = 0;
    //even j, even k
    for (kFine=1; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (jFine=1; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){
	
	// left bottom corner index
	i = StartIndex[m][0]-2;
	j = StartIndex[m][1]-1 + (jFine-1) / 2;
	k = StartIndex[m][2]-1 + (kFine-1) / 2;

	fineGhostPoint(0, // 0 for odd; 1 for even in x direction
		       1, // 0 for odd; 1 for even in y direction
		       1, // 0 for odd; 1 for even in z direction
		       h,
		       fp);	
      }
      
    }


    //east:
    iFine = pointNumberMeshP[m][0]-1;

    // j odd, k odd
    for (kFine=0; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (jFine=0; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){

	// left bottom corner index
	i = StartIndex[m][0] + cellNumberPerPatch;
	j = StartIndex[m][1]-2 + jFine / 2;
	k = StartIndex[m][2]-2 + kFine / 2;

	fineGhostPoint(1, // 0 for odd; 1 for even in x direction
		       0, // 0 for odd; 1 for even in y direction
		       0, // 0 for odd; 1 for even in z direction
		       h,
		       fp);	
      }
    }

    //east:
    iFine = pointNumberMeshP[m][0]-1;
    // j odd, k even
    for (kFine=1; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (jFine=0; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){
	
	// left bottom corner index
	i = StartIndex[m][0] + cellNumberPerPatch;
	j = StartIndex[m][1]-2 + jFine / 2;
	k = StartIndex[m][2]-1 + (kFine-1) / 2;

 	fineGhostPoint(1, // 0 for odd; 1 for even in x direction
		       0, // 0 for odd; 1 for even in y direction
		       1, // 0 for odd; 1 for even in z direction
		       h,
		       fp);	
     }
    }


    //east:
    iFine = pointNumberMeshP[m][0]-1;

    // j even, k odd
    for (kFine=0; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (jFine=1; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){

	// left bottom corner index
	i = StartIndex[m][0] + cellNumberPerPatch;
	j = StartIndex[m][1]-1 + (jFine-1) / 2;
	k = StartIndex[m][2]-2 + kFine / 2;

	fineGhostPoint(1, // 0 for odd; 1 for even in x direction
		       1, // 0 for odd; 1 for even in y direction
		       0, // 0 for odd; 1 for even in z direction
		       h,
		       fp);	
      }
    }

    //east:
    iFine = pointNumberMeshP[m][0]-1;
    // j even, k even
    for (kFine=1; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (jFine=1; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){
	
	// left bottom corner index
	i = StartIndex[m][0] + cellNumberPerPatch;
	j = StartIndex[m][1]-1 + (jFine-1) / 2;
	k = StartIndex[m][2]-1 + (kFine-1) / 2;

	fineGhostPoint(1, // 0 for odd; 1 for even in x direction
		       1, // 0 for odd; 1 for even in y direction
		       1, // 0 for odd; 1 for even in z direction
		       h,
		       fp);	
      }
      
    }



    //////////////////////////////////////////////
    //
    //  y direction
    //
    /////////////////////////////////////////////

    //pay special attention to those shared by surfaces


    //south:
    jFine = 0;
    //odd i, //odd k
    for (kFine=0; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (iFine=0; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
     
	// west south bottom corner index
	i = StartIndex[m][0]-2 + iFine / 2;
	j = StartIndex[m][1]-2;
	k = StartIndex[m][2]-2 + kFine / 2;
	
	fineGhostPoint(0, // 0 for odd; 1 for even in x direction
		       0, // 0 for odd; 1 for even in y direction
		       0, // 0 for odd; 1 for even in z direction
		       h,
		       fp);	
      }
    }

    //south:
    jFine = 0;
    //odd i,even k
    for (kFine=1; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (iFine=0; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
	
	// left bottom corner index
	i = StartIndex[m][0]-2 + iFine / 2;
	j = StartIndex[m][1]-2;
	k = StartIndex[m][2]-1 + (kFine-1) / 2;
	
	fineGhostPoint(0, // 0 for odd; 1 for even in x direction
		       0, // 0 for odd; 1 for even in y direction
		       1, // 0 for odd; 1 for even in z direction
		       h,
		       fp);	
      }
    }

    //south:
    jFine = 0;
    //even i, odd k
    for (kFine=0; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (iFine=1; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
		
	// left bottom corner index
	i = StartIndex[m][0]-1 + (iFine-1) / 2;
	j = StartIndex[m][1]-2;
	k = StartIndex[m][2]-2 + kFine / 2;

 	fineGhostPoint(1, // 0 for odd; 1 for even in x direction
		       0, // 0 for odd; 1 for even in y direction
		       0, // 0 for odd; 1 for even in z direction
		       h,
		       fp);	 
      }
    }

    //south:
    jFine = 0;
    //even i, even k
    for (kFine=1; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (iFine=1; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
	
	// left bottom corner index
	i = StartIndex[m][0]-1 + (iFine-1) / 2;
	j = StartIndex[m][1]-2;
	k = StartIndex[m][2]-1 + (kFine-1) / 2;

	fineGhostPoint(1, // 0 for odd; 1 for even in x direction
		       0, // 0 for odd; 1 for even in y direction
		       1, // 0 for odd; 1 for even in z direction
		       h,
		       fp);
      }
    }


    //north:
    jFine = pointNumberMeshP[m][1]-1;

    // i odd, k odd
    for (kFine=0; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (iFine=0; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
     //pay special attention to those shared by surfaces

	// left bottom corner index
	i = StartIndex[m][0]-2 + iFine / 2;
	j = StartIndex[m][1] + cellNumberPerPatch;
	k = StartIndex[m][2]-2 + kFine / 2;

	fineGhostPoint(0, // 0 for odd; 1 for even in x direction
		       1, // 0 for odd; 1 for even in y direction
		       0, // 0 for odd; 1 for even in z direction
		       h,
		       fp); 
      }
    }

    //north: j even
    jFine = pointNumberMeshP[m][1]-1;
    // i odd, k even
    for (kFine=1; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (iFine=0; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
	
	// left bottom corner index
	i = StartIndex[m][0]-2 + iFine / 2;
	j = StartIndex[m][1] + cellNumberPerPatch;
	k = StartIndex[m][2]-1 + (kFine-1) / 2;

	fineGhostPoint(0, // 0 for odd; 1 for even in x direction
		       1, // 0 for odd; 1 for even in y direction
		       1, // 0 for odd; 1 for even in z direction
		       h,
		       fp); 
      }
    }


    //north:
    jFine = pointNumberMeshP[m][1]-1;

    // i even, k odd
    for (kFine=0; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (iFine=1; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
     //pay special attention to those shared by surfaces

	// left bottom corner index
	j = StartIndex[m][1] + cellNumberPerPatch;
	i = StartIndex[m][0]-1 + (iFine-1) / 2;
	k = StartIndex[m][2]-2 + kFine / 2;

	fineGhostPoint(1, // 0 for odd; 1 for even in x direction
		       1, // 0 for odd; 1 for even in y direction
		       0, // 0 for odd; 1 for even in z direction
		       h,
		       fp); 
      }
    }

    //north:
    jFine = pointNumberMeshP[m][1]-1;
    // i even, k even
    for (kFine=1; kFine<=pointNumberMeshP[m][2]-1; kFine=kFine+2){
      for (iFine=1; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
	
	// left bottom corner index
	i = StartIndex[m][0]-1 + (iFine-1) / 2;
	j = StartIndex[m][1] + cellNumberPerPatch;
	k = StartIndex[m][2]-1 + (kFine-1) / 2;

	fineGhostPoint(1, // 0 for odd; 1 for even in x direction
		       1, // 0 for odd; 1 for even in y direction
		       1, // 0 for odd; 1 for even in z direction
		       h,
		       fp); 
      }     
    }





    //////////////////////////////////////////////
    //
    //  z direction
    //
    /////////////////////////////////////////////

    //bottom:
   kFine = 0;
    //odd j, //odd k
   for (jFine=0; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){
     for (iFine=0; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
	
	// west south bottom corner index
	i = StartIndex[m][0]-2 + iFine / 2;
	j = StartIndex[m][1]-2 + jFine / 2;
	k = StartIndex[m][2]-2;

	fineGhostPoint(0, // 0 for odd; 1 for even in x direction
		       0, // 0 for odd; 1 for even in y direction
		       0, // 0 for odd; 1 for even in z direction
		       h,
		       fp); 
      }
    }

    //bottom:
    kFine = 0;
    //odd j,even i
    for (jFine=0; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){
      for (iFine=1;iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
	
	// left bottom corner index
	i = StartIndex[m][0]-1 + (iFine-1) / 2;
	j = StartIndex[m][1]-2 + jFine / 2;
	k = StartIndex[m][2]-2;
	
	fineGhostPoint(1, // 0 for odd; 1 for even in x direction
		       0, // 0 for odd; 1 for even in y direction
		       0, // 0 for odd; 1 for even in z direction
		       h,
		       fp); 
      }      
    }


    //bottom:
    kFine = 0;
    //even j, odd i
    for (iFine=0; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
      for (jFine=1; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){
	//pay special attention to corner
	
	// left bottom corner index
	i = StartIndex[m][0]-2 + iFine / 2;
	j = StartIndex[m][1]-1 + (jFine-1) / 2;
	k = StartIndex[m][2]-2;

	fineGhostPoint(0, // 0 for odd; 1 for even in x direction
		       1, // 0 for odd; 1 for even in y direction
		       0, // 0 for odd; 1 for even in z direction
		       h,
		       fp); 
      }
    }

    //bottom:
    kFine = 0;
    //even j, even i
    for (iFine=1; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
      for (jFine=1; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){
	
	// left bottom corner index
	i = StartIndex[m][0]-1 + (iFine-1) / 2;
	j = StartIndex[m][1]-1 + (jFine-1) / 2;
	k = StartIndex[m][2]-2;

	fineGhostPoint(1, // 0 for odd; 1 for even in x direction
		       1, // 0 for odd; 1 for even in y direction
		       0, // 0 for odd; 1 for even in z direction
		       h,
		       fp); 
      }      
    }


    //top:
    kFine = pointNumberMeshP[m][2]-1;

    // j odd, i odd
    for (iFine=0; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
      for (jFine=0; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){
	
	// left bottom corner index
	i = StartIndex[m][0]-2 + iFine / 2;
	j = StartIndex[m][1]-2 + jFine / 2;
	k = StartIndex[m][2] + cellNumberPerPatch;

	fineGhostPoint(0, // 0 for odd; 1 for even in x direction
		       0, // 0 for odd; 1 for even in y direction
		       1, // 0 for odd; 1 for even in z direction
		       h,
		       fp); 
      }
    }

    //top:
    kFine = pointNumberMeshP[m][2]-1;
    // j odd, i even
    for (iFine=1; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
      for (jFine=0; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){

	// left bottom corner index
	i = StartIndex[m][0]-1 + (iFine-1) / 2;
	j = StartIndex[m][1]-2 + jFine / 2;
	k = StartIndex[m][2] + cellNumberPerPatch;

	fineGhostPoint(1, // 0 for odd; 1 for even in x direction
		       0, // 0 for odd; 1 for even in y direction
		       1, // 0 for odd; 1 for even in z direction
		       h,
		       fp); 
      }
    }


    //top:
    kFine = pointNumberMeshP[m][2]-1;

    // j even, i odd
    for (iFine=0; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
      for (jFine=1; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){

	// left bottom corner index
	k = StartIndex[m][2] + cellNumberPerPatch;
	j = StartIndex[m][1]-1 + (jFine-1) / 2;
	i = StartIndex[m][0]-2 + iFine / 2;

	fineGhostPoint(0, // 0 for odd; 1 for even in x direction
		       1, // 0 for odd; 1 for even in y direction
		       1, // 0 for odd; 1 for even in z direction
		       h,
		       fp); 
      }
    }

    //top:
    kFine = pointNumberMeshP[m][2]-1;
    // j even, k even
    for (iFine=1; iFine<=pointNumberMeshP[m][0]-1; iFine=iFine+2){
      for (jFine=1; jFine<=pointNumberMeshP[m][1]-1; jFine=jFine+2){
	
	// left bottom corner index
	i = StartIndex[m][0]-1 + (iFine-1) / 2;
	j = StartIndex[m][1]-1 + (jFine-1) / 2;
	k = StartIndex[m][2] + cellNumberPerPatch;

	fineGhostPoint(1, // 0 for odd; 1 for even in x direction
		       1, // 0 for odd; 1 for even in y direction
		       1, // 0 for odd; 1 for even in z direction
		       h,
		       fp); 
      }     
    }

  // third

  // pressure points inside cages
  // modify the initial matrix for these points

    for (k=1; k<=pointNumberMeshP[m][2]-2; k++){
      for (j=1; j<=pointNumberMeshP[m][1]-2; j++){
	for (i=1; i<=pointNumberMeshP[m][0]-2; i++){
	  
	  if ( pCagePoint[m][i][j][k] == 1 ) {
	    
	    iC = index[m][i][j][k];
	    iW = index[m][i-1][j][k];
	    iE = index[m][i+1][j][k];
	    iS = index[m][i][j-1][k];
	    iN = index[m][i][j+1][k];
	    iB = index[m][i][j][k-1];
	    iT = index[m][i][j][k+1];
	    
	    valueA [iC][0] =  6.0;
	    columnA[iC][0] = iC;
	    
	    //get correct coefficient for the center point
	    //west
	    if (pUnknowns[m][i-1][j][k] == 0) valueA [iC][0] -=  1.0;
	    //east
	    if (pUnknowns[m][i+1][j][k] == 0) valueA [iC][0] -=  1.0;
	    //south
	    if (pUnknowns[m][i][j-1][k] == 0) valueA [iC][0] -=  1.0;
	    //north
	    if (pUnknowns[m][i][j+1][k] == 0) valueA [iC][0] -=  1.0;
	    //bottom
	    if (pUnknowns[m][i][j][k-1] == 0) valueA [iC][0] -=  1.0;
	    //top
	    if (pUnknowns[m][i][j][k+1] == 0) valueA [iC][0] -=  1.0;
	    
	    // initialization
	    for (iD=1; iD<nStencil; iD++){
	      valueA [iC][iD] =  valueA[iC][0];
	      columnA[iC][iD] = iC;
	    } 

	    
	    //get correct coefficient and also index for surrendering points
	    //west
	    if (pUnknowns[m][i-1][j][k] == 1) {
	      valueA [iC][1] = -1.0;
	      columnA[iC][1] = iW;	    
	    }
	    
	    //east
	    if (pUnknowns[m][i+1][j][k] == 1) {
	      valueA [iC][2] = -1.0;
	      columnA[iC][2] = iE;	    
	    }
	    
	    //south
	    if (pUnknowns[m][i][j-1][k] == 1) {
	      valueA [iC][3] = -1.0;
	      columnA[iC][3] = iS;	    
	    }
	    
	    //north
	    if (pUnknowns[m][i][j+1][k] == 1) {
	      valueA [iC][4] = -1.0;
	      columnA[iC][4] = iN;	    
	    }

	    //bottom
	    if (pUnknowns[m][i][j][k-1] == 1) {
	      valueA [iC][5] = -1.0;
	      columnA[iC][5] = iB;	    
	    }
	    
	    //north
	    if (pUnknowns[m][i][j][k+1] == 1) {
	      valueA [iC][6] = -1.0;
	      columnA[iC][6] = iT;	    
	    }

	    /*	    
	    valueA [iC][1] *= 4.0/valueA [iC][0] ;
	    valueA [iC][2] *= 4.0/valueA [iC][0] ;
	    valueA [iC][3] *= 4.0/valueA [iC][0] ;
	    valueA [iC][4] *= 4.0/valueA [iC][0] ;
	    b[iC] *= 4.0/valueA [iC][0] ;
	    
	    valueA [iC][0] *= 4.0/valueA [iC][0] ;*/
	  }
	}
      }
    }     
    
      
      
  } // end for m=1


  // ---------------------- PETSc starts ---------------------
  // PETSc solving A x = b
  PCType preCond = PCILU; // others: PCICC, PCJACOBI, PCILU
  KSPType KrylovMethod = KSPBCGS; // others: KSPCG,KSPCGS, KSPBICG, KSPBCGS
  
  // solve the linear system Ax=b, where x=sol and b=b
  pSolver.LinearSolver(n,
		       valueA,
		       columnA,
		       nStencil,
		       b,
		       sol,
		       preCond,
		       KrylovMethod,
		       tol,
		       zeroInitialGuess);
  
  // ---------------------- PETSc ends ---------------------

  for (ib=0; ib<n; ib++) {
    m = indexM[ib];
    i = indexI[ib];
    j = indexJ[ib];
    k = indexK[ib];
    
    p[m][i][j][k] = sol[ib];
  }

  /*
  maxError = fabs(exactP[0] -sol[0]);
  for (ib=0; ib<n; ib++) {
    //         cout <<ib << "  " <<exactP[ib] -sol[ib] << endl;
    //         cout <<exactP[ib] << endl;
      if (fabs( exactP[ib] -sol[ib] ) > maxError) 
	maxError = fabs(exactP[ib] -sol[ib]);
  }
  cout <<"max error is " <<maxError << endl;
  */

  
  /*    
  for (j=0; j<n; j++) {
  for (i=0; i<n; i++) {
  cout << A[i][j] << " ";
  }   
  cout << endl;
  } 

 
  for (i=0; i<n; i++) {
    cout << p[i]<< "  ";
  }  
  cout <<endl;


  for (i=0; i<n; i++) {
    cout << b[i] << "  ";
  }  
  cout <<endl;
 
 
    for (j=0; j<cellNumberCoarse[1]; j++) {
  for (i=0; i<cellNumberCoarse[0]; i++) {
      cout << originalIndexCoarse[j+i*cellNumberCoarse[0]] << "  ";
    }
  cout << endl;  
  }  
  cout << endl;  
 
    for (j=0; j<cellNumberCoarse[1]; j++) {
 for (i=0; i<cellNumberCoarse[0]; i++) {
      cout << insideOutsideFineMesh[j+i*cellNumberCoarse[0]] << "  ";
    }  
    cout << endl;  
  }
  cout << endl;  

    for (j=0; j<cellNumberCoarse[1]; j++) {
  for (i=0; i<cellNumberCoarse[0]; i++) {
      cout << OriginalToCoarse[j+i*cellNumberCoarse[0]] << "  ";
    }
    cout << endl;  
  }
  cout << endl;  

  for (i=0; i<cellNumberCoarse[0]*cellNumberCoarse[1]- nPatch*(cellNumberPerPatch-2)*(cellNumberPerPatch-2); i++) {
    cout << ToOriginalIndex[i] << "  ";
  }  
  cout << endl;  

*/ 

  deallocateMemory(pointNumberMeshP);

}

void pressure::coarseGhostPoint(real *h,
				real ****fp){
  
  // the center point   
  iC = index[0][i][j][k];
  b[iC] = - 1.0/4.0 * h[0] * h[0] * fp[0][i][j][k];
  // b[iC] = 0.0; // interpolation      
  
  // eight sorrounding points
  iWSB = index[m][iFine][jFine][kFine];
  iESB = index[m][iFine+1][jFine][kFine];
  iWNB = index[m][iFine][jFine+1][kFine];
  iENB = index[m][iFine+1][jFine+1][kFine];
  
  iWST = index[m][iFine][jFine][kFine+1];
  iEST = index[m][iFine+1][jFine][kFine+1];
  iWNT = index[m][iFine][jFine+1][kFine+1];
  iENT = index[m][iFine+1][jFine+1][kFine+1];
  
  // use the average of eight sorrounding points
  // to approximate the center point   
  
  valueA [iC][0] =  8.0;
  columnA[iC][0] = iC;
  
  valueA [iC][1] = -1.0;
  columnA[iC][1] = iWSB;
  valueA [iC][2] = -1.0;
  columnA[iC][2] = iESB;
  valueA [iC][3] = -1.0;
  columnA[iC][3] = iWNB;
  valueA [iC][4] = -1.0;
  columnA[iC][4] = iENB;
  
  valueA [iC][5] = -1.0;
  columnA[iC][5] = iWST;
  valueA [iC][6] = -1.0;
  columnA[iC][6] = iEST;
  valueA [iC][7] = -1.0;
  columnA[iC][7] = iWNT;
  valueA [iC][8] = -1.0;
  columnA[iC][8] = iENT;
  
}

void pressure::fineGhostPoint(real ix, // 0 for odd; 1 for even in x direction
			      real iy, // 0 for odd; 1 for even in y direction
			      real iz, // 0 for odd; 1 for even in z direction
			      real *h,
			      real ****fp){
  
  iC = index[m][iFine][jFine][kFine];
  b[iC] = -24.0 * h[m] * h[m] * fp[m][iFine][jFine][kFine];       
  //      b[iC] *= 0.0; // interpolation      
    
  valueA [iC][0] =  64.0;
  columnA[iC][0] = iC;
  
  valueA [iC][1] = - pow(3.0, fabs(ix) + fabs(iy) + fabs(iz));
  iWSB = index[0][i][j][k];
  columnA[iC][1] = iWSB;
  
  valueA [iC][2] = -pow(3.0, fabs(ix - 1) + fabs(iy) + fabs(iz));
  iESB = index[0][i+1][j][k];
  columnA[iC][2] = iESB;
  
  valueA [iC][3] = -pow(3.0, fabs(ix) + fabs(iy - 1) + fabs(iz));
  iWNB = index[0][i][j+1][k];
  columnA[iC][3] = iWNB;
  
  valueA [iC][4] = -pow(3.0, fabs(ix - 1) + fabs(iy - 1) + fabs(iz));
  iENB = index[0][i+1][j+1][k];
  columnA[iC][4] = iENB;
  
  valueA [iC][5] = -pow(3.0, fabs(ix) + fabs(iy) + fabs(iz - 1));
  iWST = index[0][i][j][k+1];
  columnA[iC][5] = iWST;
  
  valueA [iC][6] = -pow(3.0, fabs(ix - 1) + fabs(iy) + fabs(iz - 1));
  iEST = index[0][i+1][j][k+1];
  columnA[iC][6] = iEST;
  
  valueA [iC][7] = -pow(3.0, fabs(ix) + fabs(iy - 1) + fabs(iz - 1));
  iWNT = index[0][i][j+1][k+1];
  columnA[iC][7] = iWNT;
  
  valueA [iC][8] = -pow(3.0, fabs(ix - 1) + fabs(iy - 1) + fabs(iz - 1));
  iENT = index[0][i+1][j+1][k+1];
  columnA[iC][8] = iENT;
  
}

int * pressure::allocate1dIntArray(int size){
// pointer can be used only after initialized

  int *x = new int[size];
  return x; 
}
void pressure::delete1dArray(int *x){
  delete [] x;
}

real * pressure::allocate1dRealArray(int size){
  real *x = new real[size];
  return x;
}
void pressure::delete1dArray(real *x){
  delete [] x;
}


int **** pressure::allocate4dIntArray(int nMeshes, int **pointNumberMeshP){
  int ****x = new int ***[nMeshes];
  for (m=0; m<nMeshes; m++) {
    x[m] = new int **[pointNumberMeshP[m][0]];
    for (i=0; i<pointNumberMeshP[m][0]; i++) {
      x[m][i] = new int*[pointNumberMeshP[m][1]];
      for (j=0; j<pointNumberMeshP[m][1]; j++) {
	x[m][i][j] = new int[pointNumberMeshP[m][2]];
      }
    }
  }
  return x;
}
void pressure::delete4dArray(int ****x, int nMeshes, int **pointNumberMeshP){
  for (m=0; m<nMeshes; m++) {
    for (i=0; i<pointNumberMeshP[m][0]; i++) {
      for (j=0; j<pointNumberMeshP[m][1]; j++) {
	delete [] x[m][i][j];
      }
      delete [] x[m][i];
    }
    delete [] x[m];
  }
  delete [] x;
}

real ** pressure::allocateValueMatrix(){
  real **x = new real *[n];
  for (m=0; m<n; m++) {
    x[m] = new real[nStencil];
  }   
  return x;
}
void pressure::deleteMatrix(real **x){
  for (m=0; m<n; m++) {
    delete [] x[m];
  }   
  delete [] x;
}

int ** pressure::allocateColumnMatrix(){
  int **x = new int *[n];
  for (m=0; m<n; m++) {
    x[m] = new int[nStencil];
  }   
  return x;
}
void pressure::deleteMatrix(int **x){
  for (m=0; m<n; m++) {
    delete [] x[m];
  }   
  delete [] x;
}


void pressure::deallocateMemory(int ** pointNumberMeshP){
  deleteMatrix(valueA);
  deleteMatrix(columnA);

  delete1dArray(indexM);
  delete1dArray(indexI);
  delete1dArray(indexJ);
  delete1dArray(indexK);

  delete1dArray(b);
  delete1dArray(sol);
  /*
  delete1dArray(xCoord);
  delete1dArray(yCoord);
  delete1dArray(zCoord);
  */ 
  delete4dArray(index, nMeshes, pointNumberMeshP);

}

pressure::~pressure(){
}
