#ifndef PRESSURE_H
#define PRESSURE_H

/** \file pressure.cpp
    \brief pressure Poisson equation Solver 

    This class is the pressure Poisson equation Solver. It creats indices for all pressure points on the coarse and fine mesh, generates the matrix, and assigns the right hand side of the equation, then call  PetscLinearSolver to solve the linear system.
*/

#include <iostream>
#include <string>
#include <cmath>
#include "petscksp.h"
#include "PetscLinearSolver.h"
using namespace std;

#ifdef SINGLE
#define real float
#else
#define real double
#endif


/** 
    \brief pressure Poisson equation Solver 

    This class is the pressure Poisson equation Solver. It creats indices for all pressure points on the coarse and fine mesh, generates the matrix, and assigns the right hand side of the equation, then call  PetscLinearSolver to solve the linear system.

   Fields of pressure transform to 1d vector from the original two or three-dimensional domain. Solving pressure in vector form A x = b

*/

class pressure {

 public:

  /**
     constructor
   */
  pressure();

/** 
    \brief pressure Poisson equation Solver 

    This class is the pressure Poisson equation Solver. It creats indices for all pressure points on the coarse and fine mesh, generates the matrix, and assigns the right hand side of the equation, then call  PetscLinearSolver to solve the linear system.
*/
  void initialize(int numberOfPatches, /**< number of patches */
		  int cellNumberPerPatch, /**<  cell numbers per patch surrounding the sphere*/
		  int ** StartIndex, /**< used to put the refined mesh */
		  int ** pointNumberMeshP, /**< mesh points for p */
		  int dimension,/**<  dimension */
		  real ***xP, /**< coordinates of p  */
		  real *h,/**< space step  */
		  real ****fp, /**< extra force */
		  real ****exactPressure, /**< exact pressure */
		  int ****pUnknowns,/**< mesh points for p */
		  real ****p,/**<  pressure */
		  int ****pCagePoint,/**< cage points for p */
		  PetscLinearSolver &pSolver,/**< PETSC linear solver */
		  int &zeroInitialGuess/**< if use zero as the initial guess for pressure */
		  );
  /**
     destructor
   */
  ~pressure();
  
 private: 

  void coarseGhostPoint(real *h,
			real ****fp);
  void fineGhostPoint(real ix, // 0 for odd; 1 for even in x direction
		      real iy, // 0 for odd; 1 for even in y direction
		      real iz, // 0 for odd; 1 for even in z direction
		      real *h,
		      real ****fp);
 
  int * allocate1dIntArray(int size);
  int **** allocate4dIntArray(int nMeshes, int **pointNumberMesh);
  real * allocate1dRealArray(int size);
  real ** allocateValueMatrix();
  int ** allocateColumnMatrix();

  void deleteMatrix(real **x, int n);
  void delete4dArray(int ****x, int nMeshes, int **pointNumberMesh);
  void delete1dArray(real *x);
  void delete1dArray(int *x);
  void deleteMatrix(real **x);
  void deleteMatrix(int **x);

  void deallocateMemory(int ** pointNumberMesh);
 // total points in coarse mesh inside the patch not included
  int n; // total number of unknowns

  int nStencil; // total number of unknowns

  int dim;

  // all points including in the original coarse and mesh points.
  // including points considerred in the linear system and those unconsiderred
  int originalTotalPoints;

  // iRHS for index in matrix A = totalUnknowns x totalUnknowns
  // jRHS
  int i, j, k;

  int ib; //dummy index for b=RHS
  int iD; //dummy index for anything

  int iFine, jFine, kFine;
  int iWSB, iESB,iWNB,iENB; // four bottom corners
  int iWST, iEST,iWNT,iENT; // four top corners

  // m for mesh iteration: meshes for patchs and coarse grid
  int m;

  int nPatch; // each patch carries a square fine mesh
  int nMeshes;   // including the coarse mesh
  // the coarse mesh is the zeroth mesh and fine meshes are from 1st to the last

  // indices on the b for 
  // center, 
  // west, east, 
  // south, north,
  // bottom, and top
  int iC,iW,iE,iS,iN,iB,iT;  
 


  int nCellPerPatch; // cell number in the coarse mesh that a patch uses

  int * indexM;  // indexM[n]
  int * indexI;  // indexI[n]
  int * indexJ;  // indexJ[n]
  int * indexK;  // indexK[n]

  
  int  ****index;


  // solving linear system
  
  real **valueA;   // values of A matrix
  int **columnA;   // corrsponding column of A matrix

  real tol;  // tolerance
  real *b; // right hand side
  real *sol;  // solution

  /*
  real *xCoord; // x coordinate
  real *yCoord; // y coordinate
  real *zCoord; // y coordinate
  */

  // checking accuracy
  real maxError;



};

#endif
